/*---------------------------------------------------------------------------*
  Project:  sndio
  File:     sndio_VoiceTypes.h

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#ifndef SNDIO_VOICE_TYPES_H
#define SNDIO_VOICE_TYPES_H

#include <nn/types.h>
#include <nn/fs.h>


namespace sndio {


/*!
    @brief      ۑy[W
*/
typedef enum
{
    VOICE_PAGE_TWL_NAND = -1,       //!< TWLTEḧ
    VOICE_PAGE_CTR_NAND,            //!< {̕ۑ̈
    VOICE_PAGE_SD_01,               //!< SDJ[h 1y[W
    VOICE_PAGE_SD_02,               //!< SDJ[h 2y[W
    VOICE_PAGE_SD_03,               //!< SDJ[h 3y[W
    VOICE_PAGE_SD_04,               //!< SDJ[h 4y[W
    VOICE_PAGE_SD_05,               //!< SDJ[h 5y[W
    VOICE_PAGE_SD_06,               //!< SDJ[h 6y[W
    VOICE_PAGE_SD_07,               //!< SDJ[h 7y[W
    VOICE_PAGE_SD_08,               //!< SDJ[h 8y[W
    VOICE_PAGE_SD_09,               //!< SDJ[h 9y[W
    VOICE_PAGE_SD_10,               //!< SDJ[h10y[W
    VOICE_PAGE_NUM
} VoicePage;


/*!
    @brief      Py[W̉
*/
static const s32 VOICE_COLUMN_COUNT = 6;


/*!
    @brief      Py[W̉s
*/
static const s32 VOICE_ROW_COUNT = 3;


/*!
    @brief      Py[W̉ۑ
*/
static const u32 VOICE_NUM_IN_PAGE = VOICE_COLUMN_COUNT * VOICE_ROW_COUNT;


/*!
    @brief      TCY萔(50KByte)
*/
static const std::size_t VOICE_LIMIT_BYTES = 50 * 1024;


/*!
    @brief      J[̎
*/
typedef enum
{
    VOICE_COLOR_RED,            //!<  nn::util::Color8( 255,  49,   0, 255 )
    VOICE_COLOR_ORANGE,         //!<  nn::util::Color8( 255, 132,   0, 255 )
    VOICE_COLOR_YELLOW,         //!<  nn::util::Color8( 255, 189,  16, 255 )
    VOICE_COLOR_GREEN,          //!<  nn::util::Color8(  24, 222,  66, 255 )
    VOICE_COLOR_BLUE,           //!<  nn::util::Color8(  49, 132, 255, 255 )
    VOICE_COLOR_PURPLE,         //!<  nn::util::Color8( 132,  82, 255, 255 )
    VOICE_COLOR_NUM
} VoiceColor;


/*!
    @brief      ӂ܂
*/
typedef enum
{
    VOICE_FILL_25_LT,           //!< 25% (Ȃ)
    VOICE_FILL_50_LT,           //!< 50%
    VOICE_FILL_75_LT,           //!< 75%
    VOICE_FILL_100_LT,          //!< 100%
    VOICE_FILL_100              //!< 100% ()
} VoiceFill;


/*!
    @brief      ӂ̌`
*/
typedef enum
{
    VOICE_FORM_EMPTY,           //!< 
    VOICE_FORM_NORMAL,          //!< ʏ()
    VOICE_FORM_JAG,             //!< tB^(MUMU)
    VOICE_FORM_SQUARE,          //!< sb`EXs[h(lp)
    VOICE_FORM_NUM
} VoiceForm;


/*!
    @brief      Gg
*/
struct VoiceEntry
{
    //! pX
    //! ztpX FileStreamł͒ڈȂSoundReaderɂ͂̂܂ܓn
    wchar_t             entryName[ nn::fs::MAX_FILE_PATH_LENGTH + 1 ];

    VoicePage           page;               //!< y[W
    NN_PADDING3;
    s32                 index;              //!< CfbNX
    VoiceColor          color;              //!< J[
    VoiceFill           fill;               //!< ܂
    VoiceForm           form;               //!< `
    bool                invalid;            //!< s
    bool                readOnly;           //!< ǂݎp
    NN_PADDING3;

    //! GgpX
    const wchar_t* GetEntryPath() const
    {
        return entryName;
    }

    //! y[W
    s16 GetPage() const
    {
        return static_cast<s16>(page);
    }

    //! y[WɕۑĂʒu (0`17)
    s32 GetIndex() const
    {
        return static_cast<s32>(index);
    }

    //! J[
    s32 GetColor() const
    {
        return static_cast<s32>(color);
    }

    //! ܂
    s32 GetFill() const
    {
        return static_cast<s32>(fill);
    }

    //! ӂ`
    s32 GetForm() const
    {
        return static_cast<s32>(form);
    }

    //! f[^
    bool IsInvalid() const
    {
        return invalid;
    }

    //! ǂݎpf[^
    bool IsReadOnly() const
    {
        return readOnly;
    }
};


/*!
    @brief      ۑp[^
*/
struct VoiceParameter
{
    char                titleName[ 128 ];       //!< ^Cg(UTF8)
    u8                  titleNameLen;           //!< ^Cg(byte)
    char                albumName[ 128 ];       //!< Ao(UTF8)
    u8                  albumNameLen;           //!< Ao(byte)
    char                artistName[ 128 ];      //!< A[eBXg(UTF8)
    u8                  artistNameLen;          //!< A[eBXg(byte)
    char                sftw[6];                //!< ۑ\tgEFAID
    char                rcid[9];                //!< ^{ID
    char                edid[9];                //!< CTRAvŕҏW{ID
    NN_PADDING1;

    bool                permitUpload;           //!< C^[lbgAbv[h
    bool                permitEditDistribute;   //!< ό̉zz邩
    bool                permitFirstDistribute;  //!< \
    bool                permitSecondDistribute; //!< 񎟔zz\
    NN_PADDING4;

    //! 쐬
    nn::fnd::DateTimeParameters createDate;
};


}


#endif /* SNDIO_VOICE_TYPES_H */
