#ifndef TITLE_3DSSOUND_H_
#define TITLE_3DSSOUND_H_

#include <nn/cfg.h>
#include <nn/types.h>
#include <nn/fnd.h>


#define SNDIO_CONVERT_LENGTH 128


static inline void Format( wchar_t* pOut, const wchar_t* fmt, ... )
{
    std::va_list vlist;

    va_start(vlist, fmt);

    (void)std::vswprintf( pOut, SNDIO_CONVERT_LENGTH, fmt, vlist );

    va_end(vlist);
}


static inline void ConvertDateTime(
    wchar_t* pOut,
    const nn::cfg::CfgRegionCode& region,
    const nn::cfg::CfgLanguageCode& lang,
    const nn::fnd::DateTime& date )
{
    switch( region )
    {
    default:
    case nn::cfg::CFG_REGION_JAPAN:
        {
            switch( lang )
            {
            default:
            case nn::cfg::CFG_LANGUAGE_JAPANESE:
                // JP_Japanese    %d/%2d/%2d   %2d:%02d:%02d    [Replace:Replace_Year ]/[Replace:Replace_MonthSpace ]/[Replace:Replace_DaySpace ]   [Replace:Replace_HourSpace ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%d/%2d/%2d   %2d:%02d:%02d", date.GetYear(), date.GetMonth(), date.GetDay(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            }
        }
        break;

    case nn::cfg::CFG_REGION_AMERICA:
        {
            switch( lang )
            {
            default:
            case nn::cfg::CFG_LANGUAGE_ENGLISH:
                // US_English    %d/%d/%d   %2d:%02d:%02d    [Replace:Replace_MonthNoSpace ]/[Replace:Replace_DayNoSpace ]/[Replace:Replace_Year ]   [Replace:Replace_HourSpace ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%d/%d/%d   %2d:%02d:%02d", date.GetMonth(), date.GetDay(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_FRENCH:
                // US_French    %02d/%02d/%d   %02d:%02d:%02d    [Replace:Replace_Day ]/[Replace:Replace_Month ]/[Replace:Replace_Year ]   [Replace:Replace_Hour ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d/%02d/%d   %02d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_SPANISH:
                // US_Spanish    %02d/%02d/%d   %02d:%02d:%02d    [Replace:Replace_Day ]/[Replace:Replace_Month ]/[Replace:Replace_Year ]   [Replace:Replace_Hour ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d/%02d/%d   %02d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_PORTUGUESE:
                // US_Portuguese    %02d/%02d/%d   %02d:%02d:%02d    [Replace:Replace_Day ]/[Replace:Replace_Month ]/[Replace:Replace_Year ]   [Replace:Replace_Hour ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d/%02d/%d   %02d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            }
        }
        break;

    case nn::cfg::CFG_REGION_EUROPE:
        {
            switch( lang )
            {
            default:
            case nn::cfg::CFG_LANGUAGE_ENGLISH:
                // EU_English    %02d/%02d/%d   %02d:%02d:%02d    [Replace:Replace_Day ]/[Replace:Replace_Month ]/[Replace:Replace_Year ]   [Replace:Replace_Hour ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d/%02d/%d   %02d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_FRENCH:
                // EU_French    %02d/%02d/%d   %2d:%02d:%02d    [Replace:Replace_Day ]/[Replace:Replace_Month ]/[Replace:Replace_Year ]   [Replace:Replace_HourSpace ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d/%02d/%d   %2d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_GERMAN:
                // EU_German    %02d.%02d.%d   %2d:%02d:%02d    [Replace:Replace_Day ].[Replace:Replace_Month ].[Replace:Replace_Year ]   [Replace:Replace_HourSpace ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d.%02d.%d   %2d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_ITALIAN:
                // EU_Italian    %02d/%02d/%d   %02d:%02d:%02d    [Replace:Replace_Day ]/[Replace:Replace_Month ]/[Replace:Replace_Year ]   [Replace:Replace_Hour ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d/%02d/%d   %02d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_SPANISH:
                // EU_Spanish    %02d-%02d-%d   %02d:%02d:%02d    [Replace:Replace_Day ]-[Replace:Replace_Month ]-[Replace:Replace_Year ]   [Replace:Replace_Hour ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d-%02d-%d   %02d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_DUTCH:
                // EU_Dutch    %02d-%02d-%d   %2d:%02d:%02d    [Replace:Replace_Day ]-[Replace:Replace_Month ]-[Replace:Replace_Year ]   [Replace:Replace_HourSpace ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d-%02d-%d   %2d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_PORTUGUESE:
                // EU_Portuguese    %02d/%02d/%d   %02d:%02d:%02d    [Replace:Replace_Day ]/[Replace:Replace_Month ]/[Replace:Replace_Year ]   [Replace:Replace_Hour ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d/%02d/%d   %02d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_RUSSIAN:
                // EU_Russian    %02d.%02d.%d   %02d:%02d:%02d    [Replace:Replace_Day ].[Replace:Replace_Month ].[Replace:Replace_Year ]   [Replace:Replace_Hour ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d.%02d.%d   %02d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            }
        }
        break;

    case nn::cfg::CFG_REGION_CHINA:
        {
            switch( lang )
            {
            default:
            case nn::cfg::CFG_LANGUAGE_SIMP_CHINESE:
                // CN_Simp_Chinese    %d/%2d/%2d   %2d:%02d:%02d    [Replace:Replace_Year ]/[Replace:Replace_MonthSpace ]/[Replace:Replace_DaySpace ]   [Replace:Replace_HourSpace ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%d/%2d/%2d   %2d:%02d:%02d", date.GetYear(), date.GetMonth(), date.GetDay(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            }
        }
        break;

    case nn::cfg::CFG_REGION_KOREA:
        {
            switch( lang )
            {
            default:
            case nn::cfg::CFG_LANGUAGE_KOREAN:
                // KR_Hangeul    %d/%2d/%2d   %2d:%02d:%02d    [Replace:Replace_Year ]/[Replace:Replace_MonthSpace ]/[Replace:Replace_DaySpace ]   [Replace:Replace_HourSpace ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%d/%2d/%2d   %2d:%02d:%02d", date.GetYear(), date.GetMonth(), date.GetDay(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            }
        }
        break;

    case nn::cfg::CFG_REGION_TAIWAN:
        {
            switch( lang )
            {
            default:
            case nn::cfg::CFG_LANGUAGE_TRAD_CHINESE:
                // TW_Trad_Chinese    %d/%2d/%2d   %2d:%02d:%02d    [Replace:Replace_Year ]/[Replace:Replace_MonthSpace ]/[Replace:Replace_DaySpace ]   [Replace:Replace_HourSpace ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%d/%2d/%2d   %2d:%02d:%02d", date.GetYear(), date.GetMonth(), date.GetDay(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            case nn::cfg::CFG_LANGUAGE_ENGLISH:
                // TW_English    %02d/%02d/%d   %02d:%02d:%02d    [Replace:Replace_Day ]/[Replace:Replace_Month ]/[Replace:Replace_Year ]   [Replace:Replace_Hour ]:[Replace:Replace_Minute ]:[Replace:Replace_Second ]
                Format( pOut, L"%02d/%02d/%d   %02d:%02d:%02d", date.GetDay(), date.GetMonth(), date.GetYear(), date.GetHour(), date.GetMinute(), date.GetSecond() );
                break;
            }
        }
        break;
    }
}

#endif
