/**-----------------------------------------------------------------------------

	@file		sndio_CachedReader.cpp
	@author		tetsu
	@par		[]
				LbVǂݍ݃NX(J)

				Copyright (c)2010 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#include "sndio_CachedReader.h"


#if defined(APP_SNOTE) && defined(NN_BUILD_DEBUG)
#define PRINTRESULT( result )	NN_LOG( "%s(%d)\n", __FILE__, __LINE__); NN_DBG_TPRINT_RESULT_( result )
#else
#define PRINTRESULT( result )
#endif


namespace sndio { namespace detail {


////////////////////////////////////////////////////////////////////////////////
// LbVǂݍ݃NX
////////////////////////////////////////////////////////////////////////////////


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				RXgN^

----------------------------------------------------------------------------- */
CachedReader::CachedReader() :
m_allocator( NULL ),
m_reader(),
m_seekPosition( 0 ),
m_cache( NULL ),
m_cacheSize( c_defaultSize ),
m_cachePosition( 0 ),
m_fileSize( 0ll )
{
}


/**-----------------------------------------------------------------------------

	@brief		
	@return		Ȃ
	@par		[]
				fXgN^

----------------------------------------------------------------------------- */
CachedReader::~CachedReader()
{
	Close();
}


/**-----------------------------------------------------------------------------

	@brief		t@CI[v
	@param[in]	allocator		AP[^
	@param[in]	pathName		t@CpX
	@param[in]	cacheSize		LbVTCY
	@retval		true			I[v
	@retval		false			I[vs
	@par		[]
				t@CJ܂<br>
				w肳ꂽLbVTCỸobt@Ɋmۂ܂

----------------------------------------------------------------------------- */
bool CachedReader::Open(
	nn::fnd::IAllocator& allocator,
	const wchar_t* pathName,
	size_t cacheSize )
{
	m_allocator = &allocator;
	nn::Result result = m_reader.TryInitialize( pathName );

	if (result.IsSuccess())
	{
		if (m_reader.TryGetSize( &m_fileSize ).IsFailure())
			return false;

		m_cache = reinterpret_cast<u8*>( m_allocator->Allocate( cacheSize * 2, 32 ) );
		if (m_cache == NULL)
			return false;

		std::memset( m_cache, 0, cacheSize * 2 );
		m_cacheSize = cacheSize;
		m_cachePosition = 0;
		m_seekPosition = 0;

		s32 readCount = 0;
		result = m_reader.TryRead( &readCount, m_cache, m_cacheSize );
		if (result.IsFailure())
			return false;
	}
	else
	{
		PRINTRESULT( result );
	}

	return result.IsSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		ǂݍ
	@param[out]	buffer			obt@
	@param[in]	size			ǂݎTCY
	@return		ǂݍ߂TCY
	@par		[]
				t@Cǂݍݏs܂<br>
				ǂݎ͈͂LbVɎ܂ꍇALbVǂݍ܂܂

----------------------------------------------------------------------------- */
nn::Result CachedReader::TryRead( s32* pOut, void* buffer, size_t size )
{
	s32 readCount = 0;
	nn::Result result;

	NN_NULL_TASSERT_( pOut );
	NN_NULL_TASSERT_( m_cache );

	*pOut = 0;

	// ǂݎ낤ƂĂʒuƁALbVĂʒu̍o
	s32 current = m_seekPosition - m_cachePosition;

	if (0 <= current && current + size < m_cacheSize)
	{
		// LbV͈͂Ɏ܂Ă
		m_seekPosition += size;

		std::memcpy( buffer, m_cache + current, size );
		*pOut = size;
		return nn::ResultSuccess();
	}

	// ACg
	s32 aligndiff = (m_seekPosition & 0xf);

	// ͈͂Ȃꍇ
	if (size + aligndiff < m_cacheSize)
	{
		// LbVɎ܂

		// V[Nʒur[Ȃ4oCgACgɂ
		s32 propCachePosition = (m_seekPosition - aligndiff);

		// V[Nɐ烊[h
		result = m_reader.TrySeek( propCachePosition, nn::fs::POSITION_BASE_BEGIN );
		if (result.IsFailure())
		{
			return result;
		}

		s64 pos = 0;
		result = m_reader.TryGetPosition( &pos );
		if (result.IsFailure())
		{
			return result;
		}
		m_cachePosition = s32(pos);

		result = m_reader.TryRead( &readCount, m_cache, m_cacheSize );

		if (result.IsSuccess())
		{
			if (readCount < size)
			{
				// TCYǂݍ߂ōXV(I[)
				size = readCount;

				m_seekPosition = m_cachePosition + readCount + aligndiff;
			}
			else
			{
				m_seekPosition = m_cachePosition + size + aligndiff;
			}

			std::memcpy( buffer, m_cache + aligndiff, size );
		}
		else
		{
			return result;
		}
	}
	else
	{
		// ܂肫Ȃ
		// V[Nɐ烊[h
		result = m_reader.TrySeek( m_seekPosition, nn::fs::POSITION_BASE_BEGIN );
		if (result.IsFailure())
		{
			return result;
		}

		// obt@ɂ͕ʂɓǂ
		result = m_reader.TryRead( &readCount, buffer, size );
		if (result.IsFailure())
		{
			return result;
		}

		// ̈ʒuLbV
		s64 pos = 0;
		result = m_reader.TryGetPosition( &pos );
		if (result.IsFailure())
		{
			return result;
		}
		m_cachePosition = s32(pos);
		m_seekPosition = m_cachePosition;

		result = m_reader.TryRead( &readCount, m_cache, m_cacheSize );
		if (result.IsFailure())
		{
			return result;
		}
	}

	*pOut = size;
	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		ǂݍ
	@param[out]	buffer			obt@
	@param[in]	size			ǂݎTCY
	@return		ǂݍ߂TCY
	@par		[]
				t@Cǂݍݏs܂<br>
				ǂݎ͈͂LbVɎ܂ꍇALbVǂݍ܂܂

----------------------------------------------------------------------------- */
s32 CachedReader::Read( void* buffer, size_t size )
{
	s32 readCount = 0;

	if (TryRead( &readCount, buffer, size ).IsSuccess())
	{
		return readCount;
	}

	// ǂݍݎs
	return -1;
}


/**-----------------------------------------------------------------------------

	@brief		V[N
	@param[in]	position		V[Nʒu
	@return		Ȃ
	@par		[]
				V[Ns܂<br>
				ǂݎLbV\zȂ̂ŁAׂ܂<br>
				pɂɂ͌ĂяoȂŉ

----------------------------------------------------------------------------- */
nn::Result CachedReader::TrySeek( s64 position, nn::fs::PositionBase base )
{
	s32 readCount = 0;
	nn::Result result;
	NN_UNUSED_VAR( base );

	s64 fileSize = m_fileSize;

	if (m_cachePosition <= position && position < m_cachePosition + m_cacheSize )
	{
		// vV[NʒuLbV͈͂Ȃ玟̓ǂݎʒu𓮂
		m_seekPosition = position;
	}
	else if (m_seekPosition != position)
	{
		// vꂽV[Nʒu݈ʒuƋt̏ꍇ́AtʒuLbV
		if (position < m_seekPosition)
		{
			s32 seekPrev = m_seekPosition - m_cacheSize;

			if (seekPrev < 0)
			{
				seekPrev = 0;
			}

			m_cachePosition = seekPrev;
			m_seekPosition = position;
			result = m_reader.TrySeek( seekPrev, nn::fs::POSITION_BASE_BEGIN );
			if (result.IsFailure())
			{
				return result;
			}

			result = m_reader.TryRead( &readCount, m_cache, m_cacheSize );
			if (result.IsFailure())
			{
				return result;
			}
		}
		else
		{
			if (position < s32(fileSize))
			{
				result = m_reader.TrySeek( position, nn::fs::POSITION_BASE_BEGIN );
				if (result.IsFailure())
				{
					return result;
				}
				m_cachePosition = s32(m_reader.GetPosition());
				m_seekPosition = m_cachePosition;

				result = m_reader.TryRead( &readCount, m_cache, m_cacheSize );
				if (result.IsFailure())
				{
					return result;
				}
			}
			else
			{
				position = m_fileSize;
				result = m_reader.TrySeek( position, nn::fs::POSITION_BASE_BEGIN );
				if (result.IsFailure())
				{
					return result;
				}
				std::memset( m_cache, 0x00, m_cacheSize );
				m_cachePosition = position;
				m_seekPosition = position;
			}
		}
	}

	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		V[N
	@param[in]	position		V[Nʒu
	@return		Ȃ
	@par		[]
				V[Ns܂<br>
				ǂݎLbV\zȂ̂ŁAׂ܂<br>
				pɂɂ͌ĂяoȂŉ

----------------------------------------------------------------------------- */
void CachedReader::Seek( s64 position, nn::fs::PositionBase base )
{
	TrySeek( position, base );
}


/**-----------------------------------------------------------------------------

	@brief		V[Nʒu擾
	@param[out]	V[Nʒu
	@return		
	@par		[]
				݂̃t@CV[Nʒu擾܂

----------------------------------------------------------------------------- */
nn::Result CachedReader::TryGetPosition( s64* pOut ) const
{
	NN_NULL_TASSERT_( pOut );

	*pOut = static_cast<s64>(m_seekPosition);
	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		V[Nʒu擾
	@return		V[Nʒu
	@par		[]
				݂̃t@CV[Nʒu擾܂

----------------------------------------------------------------------------- */
s64 CachedReader::GetPosition() const
{
	s64 position;

	if (TryGetPosition( &position ).IsSuccess())
	{
		return position;
	}

	return 0;
}


/**-----------------------------------------------------------------------------

	@brief		|WVw
	@return		
	@par		[]
				gp֎~

----------------------------------------------------------------------------- */
nn::Result CachedReader::TrySetPosition( s64 )
{
	NN_TASSERT_( false );
	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		|WVw
	@return		Ȃ
	@par		[]
				gp֎~

----------------------------------------------------------------------------- */
void CachedReader::SetPosition( s64 )
{
	NN_TASSERT_( false );
}


/**-----------------------------------------------------------------------------

	@brief		t@CTCY擾
	@return		
	@par		[]
				݊JĂt@C̃TCY擾܂<br>
				JĂȂꍇ̓G[ɂȂ܂

----------------------------------------------------------------------------- */
nn::Result CachedReader::TryGetSize( s64* pOut ) const
{
	NN_NULL_TASSERT_( pOut );

	if (pOut)
	{
		*pOut = m_fileSize;
	}

	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		t@CTCY擾
	@return		t@CTCY
	@par		[]
				݊JĂt@C̃TCY擾܂<br>
				JĂȂꍇ̓G[ɂȂ܂

----------------------------------------------------------------------------- */
s64 CachedReader::GetSize() const
{
	s64 fileSize;

	if (TryGetSize( &fileSize ).IsSuccess())
	{
		return fileSize;
	}

	return 0;
}


/**-----------------------------------------------------------------------------

	@brief		t@CN[Y
	@return		Ȃ
	@par		[]
				t@C܂<br>
				̃LbVobt@J܂

----------------------------------------------------------------------------- */
void CachedReader::Close()
{
	m_reader.Finalize();

	if (m_allocator && m_cache )
	{
		m_allocator->Free( m_cache );
		m_cache = NULL;
	}

	m_allocator = NULL;
	m_cacheSize = 0;
	m_cachePosition = 0;
	m_seekPosition = 0;
}


}}
