/**-----------------------------------------------------------------------------

	@file		CachedWriter.cpp
	@author		tetsu
	@par		[]
				݂ɓt@Co̓NX(J)

				Copyright (c)2010 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#include "sndio_CachedWriter.h"


namespace sndio { namespace detail {


////////////////////////////////////////////////////////////////////////////////
// ݃NX
////////////////////////////////////////////////////////////////////////////////


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				RXgN^

----------------------------------------------------------------------------- */
CachedWriter::CachedWriter( nn::fnd::IAllocator& allocator ) :
m_allocator( allocator ),
m_writeSize( 0 )
{
	m_buffer = reinterpret_cast<u8*>( m_allocator.Allocate( 50 * 1024, 32 ) );
	m_bufferSize = 50 * 1024;

	std::memset( m_pathName, 0, sizeof(m_pathName) );
}


/**-----------------------------------------------------------------------------

	@brief		
	@return		Ȃ
	@par		[]
				fXgN^

----------------------------------------------------------------------------- */
CachedWriter::~CachedWriter()
{
	m_allocator.Free( m_buffer );
}


/**-----------------------------------------------------------------------------

	@brief		t@CI[v
	@param[in]	filePath		t@CpX
	@return		Ȃ
	@par		[]
				t@CI[vs

----------------------------------------------------------------------------- */
nn::Result CachedWriter::TryInitialize( const wchar_t* pathName )
{
	std::wcsncpy( m_pathName, pathName, nn::fs::MAX_FILE_PATH_LENGTH );
	m_writeSize = 0;
	m_seekPosition = 0ll;

	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		
	@param[out]	pOut			߂TCY
	@param[in]	buffer			݃obt@
	@param[in]	size			݃TCY
	@return		݌
	@par		[]
				݂s܂

----------------------------------------------------------------------------- */
nn::Result CachedWriter::TryWrite( s32* pOut, const void* buffer, size_t size, bool flush )
{
	NN_UNUSED_VAR( flush );

	if (m_bufferSize < m_writeSize + size)
	{
		return nn::fs::ResultOperationDenied();
	}

	std::memcpy( m_buffer + s32(m_seekPosition), buffer, size );

	if (pOut)
	{
		*pOut = size;
	}

	m_seekPosition += s32(size);

	if (m_writeSize < s32(m_seekPosition))
	{
		m_writeSize = s32(m_seekPosition);
	}

	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		
	@param[out]	buffer			obt@
	@param[in]	size			݃TCY
	@return		߂TCY
	@par		[]
				ݏs܂

----------------------------------------------------------------------------- */
s32 CachedWriter::Write( const void* buffer, size_t size, bool flush )
{
	s32 writeCount = 0;

	if (TryWrite( &writeCount, buffer, size, flush ).IsSuccess())
	{
		return writeCount;
	}

	return writeCount;
}


/**-----------------------------------------------------------------------------

	@brief		V[N
	@param[in]	position		V[Nʒu
	@return		Ȃ
	@par		[]
				V[Ns܂

----------------------------------------------------------------------------- */
nn::Result CachedWriter::TrySeek( s64 position, nn::fs::PositionBase base )
{
	switch (base)
	{
	case nn::fs::POSITION_BASE_BEGIN:
		m_seekPosition = position;
		break;

	case nn::fs::POSITION_BASE_CURRENT:
		m_seekPosition += position;
		break;

	case nn::fs::POSITION_BASE_END:
		m_seekPosition = s64( m_writeSize ) - position;
		break;
	}

	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		V[N
	@param[in]	position		V[Nʒu
	@return		Ȃ
	@par		[]
				V[Ns܂

----------------------------------------------------------------------------- */
void CachedWriter::Seek( s64 position, nn::fs::PositionBase base )
{
	if (TrySeek( position, base ).IsFailure())
	{
		// V[NɎs
		NN_TASSERT_( false );
	}
}


/**-----------------------------------------------------------------------------

	@brief		V[Nʒu擾
	@param[out]	V[Nʒu
	@return		
	@par		[]
				݂̃t@CV[Nʒu擾܂

----------------------------------------------------------------------------- */
nn::Result CachedWriter::TryGetPosition( s64* pOut ) const
{
	NN_NULL_TASSERT_( pOut );

	*pOut = static_cast<s64>(m_seekPosition);
	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		V[Nʒu擾
	@return		V[Nʒu
	@par		[]
				݂̃t@CV[Nʒu擾܂

----------------------------------------------------------------------------- */
s64 CachedWriter::GetPosition() const
{
	s64 position;

	if (TryGetPosition( &position ).IsSuccess())
	{
		return position;
	}

	return -1;
}


/**-----------------------------------------------------------------------------

	@brief		|WVw
	@return		
	@par		[]
				gp֎~

----------------------------------------------------------------------------- */
nn::Result CachedWriter::TrySetPosition( s64 position )
{
	return TrySeek( position, nn::fs::POSITION_BASE_BEGIN );
}


/**-----------------------------------------------------------------------------

	@brief		|WVw
	@return		Ȃ
	@par		[]
				gp֎~

----------------------------------------------------------------------------- */
void CachedWriter::SetPosition( s64 position )
{
	if (TrySetPosition( position ).IsFailure())
	{
		NN_TASSERT_( false );
	}
}


/**-----------------------------------------------------------------------------

	@brief		TCY擾
	@return		
	@par		[]
				݂̃Xg[TCY擾܂

----------------------------------------------------------------------------- */
nn::Result CachedWriter::TryGetSize( s64* pOut ) const
{
	NN_NULL_TASSERT_( pOut );

	if (pOut)
	{
		*pOut = s64(m_writeSize);
		return nn::ResultSuccess();
	}

	return nn::Result(
		nn::Result::LEVEL_PERMANENT,
		nn::Result::SUMMARY_INVALID_RESULT_VALUE,
		nn::Result::MODULE_INVALID_RESULT_VALUE,
		nn::Result::DESCRIPTION_INVALID_RESULT_VALUE
	);
}


/**-----------------------------------------------------------------------------

	@brief		t@CTCY擾
	@return		t@CTCY
	@par		[]
				݂̃Xg[TCY擾܂

----------------------------------------------------------------------------- */
s64 CachedWriter::GetSize() const
{
	s64 fileSize;

	if (TryGetSize( &fileSize ).IsSuccess())
	{
		return fileSize;
	}

	return -1;
}


/**-----------------------------------------------------------------------------

	@brief		t@CTCY̐ݒ

----------------------------------------------------------------------------- */
nn::Result CachedWriter::TrySetSize(s64 size)
{
	if (m_bufferSize < size)
	{
		return nn::fs::ResultInvalidArgument();
	}

	m_writeSize = size;
	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		t@CTCY̐ݒ

----------------------------------------------------------------------------- */
void CachedWriter::SetSize(s64 size)
{
	(void)TrySetSize( size );
}


/**-----------------------------------------------------------------------------

	@brief		ۂ̃t@Cɏo
	@return		
	@par		[]
				ۂ̃t@Cɏo܂<br/>
				̃\bhĂяo܂
				ۂɃt@Cɂ͏݂s܂

----------------------------------------------------------------------------- */
nn::Result CachedWriter::FlushWrite()
{
	nn::Result result;
	nn::fs::FileOutputStream output;

	// LbVCɏ
	result = nn::fs::TryCreateFile( m_pathName, m_writeSize );
	if (result.IsFailure())
	{
		NN_DBG_TPRINT_RESULT_( result );
		return result;
	}

	result = output.TryInitialize( m_pathName, false );
	if (result.IsFailure())
	{
		NN_DBG_TPRINT_RESULT_( result );
		return result;
	}

	s32 writeCount = 0;
	result = output.TryWrite( &writeCount, m_buffer, m_writeSize, true );
	if (result.IsFailure())
	{
		NN_DBG_TPRINT_RESULT_( result );
		return result;
	}

	result = output.TryFlush();
	if (result.IsFailure())
	{
		return result;
	}
	output.Finalize();

	return result;
}


}}
