#include <nn/util/util_Crc.h>
#include "sndio_Converter.h"
#include "sndio_Mount.h"
#include "sndio_Resampler.h"
#include "../sndio_Api.h"
#include "../sndio_VoiceWriter.h"

#if defined(APP_SNOTE)
#define SNDIO_CONVERT_TO_ADPCM
#endif

#if defined(SNDIO_CONVERT_TO_ADPCM)
#include "sndio_DecoderAACArm.h"
#endif


#define HW_CPU_CLOCK_ARM7	( 33513982 )
#define SND_TIMER_CLOCK		( HW_CPU_CLOCK_ARM7 / 2 )
#define INDEX_NUM			( 89 )


namespace
{
	// ADPCM̏\
	typedef struct _ADPCMstate
	{
		s16	prevsample;				//!< ԍŌɏo͂ADPCMli[Ă
		s8	previndex;				//!< XebvTCYe[uւ̃CfNX
		u8	bitmode;				//!< ʏpading bitmode==3Ȃ3bitMode(\)

	} ADPCMstate;
	NN_COMPILER_ASSERT( 4 == sizeof(ADPCMstate) );

#if defined(SNDIO_CONVERT_TO_ADPCM)
	static const int MIC_SAMPLING_RATE_16360 =  (HW_CPU_CLOCK_ARM7 / 16360);

	// ^̃TvO[g(16364Hz)
	static const int REC_SAMPLING_RATE = ( HW_CPU_CLOCK_ARM7 / MIC_SAMPLING_RATE_16360 );
#endif

	// ԌvZpTvO[g(16̔{)
	static const int REC_SAMPLING_RATE_CALC_TIME = 16368;

	// ^g`̍őb
	static const int REC_WAVE_SECONDS_MAX = 10;

	// ^g`̍őTv
	static const int REC_WAVE_SAMPLES_MAX = ( REC_SAMPLING_RATE_CALC_TIME * REC_WAVE_SECONDS_MAX );

	// ^g`̍őTCY(20450 * 16)
	static const int REC_WAVE_SIZE_MAX = ( REC_WAVE_SAMPLES_MAX * 2 );
	static const u32 SAVE_WAVE_SIZE = ( REC_WAVE_SIZE_MAX / 4 );
	static const u32 SAVE_SAMPLES_SIZE = ( SAVE_WAVE_SIZE + sizeof(ADPCMstate) );

	// TEhtH[}bg`
	typedef enum
	{
		SND_WAVE_FORMAT_PCM8,
		SND_WAVE_FORMAT_PCM16,
		SND_WAVE_FORMAT_ADPCM,
		SND_WAVE_FORMAT_PSG,
		SND_WAVE_FORMAT_NOISE = SND_WAVE_FORMAT_PSG
	} SNDWaveFormat;

	// ADPCMo̓f[^
	typedef struct _SNDWaveParam
	{
		u8				format;
		u8				loopflag;
		u16				rate;
		u16				timer;
		u16				loopstart;
		u32				looplen;
	} SNDWaveParam;

	// ADPCMۑf[^
#pragma pack(1)
	typedef struct _SndSaveWave
	{
		SNDWaveParam	param;
		u8				samples[ SAVE_SAMPLES_SIZE ];
		u8				info[ 1088 ];
		u16				crc;
	} SndSaveWave;
#pragma pack()
	NN_COMPILER_ASSERT( sizeof(SndSaveWave) == 82946 );
	// 81 * 1024 + 2(CRC)

	static const s8 IndexTable[16] =
	{
		-1, -1, -1, -1, 2, 4, 6, 8,
		-1, -1, -1, -1, 2, 4, 6, 8,
	};

	static const s16 StepsizeTable[INDEX_NUM] =
	{
		7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
		19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
		50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
		130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
		337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
		876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
		2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
		5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
		15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767
	};

#if defined(SNDIO_CONVERT_TO_ADPCM)
	inline s32 c16000to16360( s32 in )
	{
		return static_cast<s32>( (s64(in) * 16364ll) / 16000ll );
	}
#endif

	inline s32 c16360to16000( s32 in )
	{
		return static_cast<s32>( (s64(in) * 16000ll) / 16364ll );
	}
}


/**-----------------------------------------------------------------------------

	@brief		񈳏k16bitg`f[^1/4 (4bit) ̃TCYɈk܂

	@param[in]	indata			PCMf[^i̓f[^j
	@param[out]	outdata			IMA ADPCMGR[hꂽf[^io̓f[^j
	@param[in]	len				PCMf[^̃Tv
	@param[in]	state			ԕۑ̂߂̕ϐQ
	@return		ϊ̃f[^TCY(byte)

----------------------------------------------------------------------------- */
#if defined(SNDIO_CONVERT_TO_ADPCM)
static u32 EncodeADPCM( s16* indata, s8* outdata, u32 len, ADPCMstate* state )
{
	s16 	insample;			//  ADPCMϊΏۂƂȂP̃Tvl
	s32 	sign;				//  ADPCMo͒l̕"+, -"
	s32 	delta;				//  PTvADPCMϊ̂Rbit̒l
	s32 	diff;				//@f[^TvlƗ\lƂ̈Ⴂ
	s32 	step;				//  XebvTCY
	s32 	predsample;			//  Tv̗\l
	s32 	diffq;				//  ̃Tv̖\l
	s32 	index;				//  StepsizeTable ւ̃CfNX;
	s32 	outputbuffer = 0;	//  㕔Srbg̒lL[vĂ;
	s32 	bufferstep;			//  tOi㕔SrbgFOASrbgFPj
	u32		outdatasize=len/2;	//  ADPCMϊ̃f[^TCY

	//NN_TASSERT_(!(len%2));
	len += (len%2);

	// l̏
	predsample = state->prevsample;
	index = state->previndex;
	step = StepsizeTable[index];
	bufferstep = 0;

	//	PTviPUbitjƂADPCMiSbitj֕ϊBiP/SɈkj
	for ( ; len > 0 ; len-- )
	{
		insample = *indata++;

		//	͒lƗ\lƂ̌덷
		diff = insample - predsample;

		//	o͒l̂Sbitڂ"+, -"̕B덷vXȂ΂OA}CiXȂ΂P
		sign = (diff < 0) ? 8 : 0;		//	}CiX̂ƂA8 (10) = 1000 (2)
		if ( sign ) diff = (-diff);		//	ΒlƂ

		//
		//	 ADPCMϊvZ̊TvF
		//    delta = diff * 4 / step;			Fo͒lʂRbit߂
		//    diffq = ( delta + 0.5 ) * step / 4;	F̗\l߂
		//
		delta = 0;

		diffq = (step >> 3);
		if ( diff >= step )
		{
			delta = 4;
			diff -= step;
			diffq += step;
		}
		step >>= 1;
		if ( diff >= step  )
		{
			delta |= 2;
			diff -= step;
			diffq += step;
		}
		step >>= 1;
		if ( diff >= step )
		{
			delta |= 1;
			diffq += step;
		}

		//	̗\l߂
		if ( sign )
		{
			predsample -= diffq;
		}
		else
		{
			predsample += diffq;
		}

		//	̗\l̍őEŏ͈̔́iPUbitjɌŒ肷
		if ( predsample > 32767 )
		{
			predsample = 32767;
		}
		else if ( predsample < -32768 )
		{
			predsample = -32768;
		}

		//	̃XebvTCYXV
		delta |= sign;

		index += IndexTable[delta];

		if ( index < 0 )
		{
			index = 0;
		}
		if ( index > (INDEX_NUM-1) )
		{
			index = (INDEX_NUM-1);
		}
		step = StepsizeTable[index];

		//	߂liSbit Sbitgݍ킹ĂPbyteɁBj
		if ( !bufferstep )
		{
			outputbuffer = (delta) & 0x0f;
		}
		else
		{
			*outdata++ = (s8)((delta<<4 & 0xf0) | outputbuffer);
		}

		bufferstep = !bufferstep;
	}

	if ( bufferstep )
	{
		*outdata++ = (s8)outputbuffer;
	}

	//	ԍŌɏo͂ADPCMli[Ă
	state->prevsample = (s16)predsample;
	state->previndex = (s8)index;

	return outdatasize;
}
#endif


/**-----------------------------------------------------------------------------

	@brief		ADPCMꂽkf[^16bitPCMg`f[^ɃfR[h܂

	@param[in]	indata			fR[hADPCMf[^ւ̃|C^
	@param[out]	outdata			fR[hꂽf[^i[ꏊւ̃|C^
	@param[in]	len				fR[hADPCMf[^̃Tv
	@param[in]	state			ԕۑ̂߂̕ϐQ
	@return		ϊ̃f[^TCY(byte)

----------------------------------------------------------------------------- */
static u32 DecodeADPCM( s8* indata, s16* outdata, u32 len, ADPCMstate*state )
{
	s32 sign;				//	"+, -"
	s32 delta;				//	ADPCM̈(4bit)̃Tvl
	s32 step;				//	XebvTCY
	s32 predsample;			//	Tv̗\l
	s32 diffq;				//	̃Tv̖\l
	s32 index;				//	StepsizeTable ւ̃CfNX
	s32 inputbuffer;		//	Srbg̒lL[vĂ
	s32 bufferstep;			//	tOi㕔SrbgFOASrbgFPj
	u32 outdatasize = len * 2;		//	ϊ̃TCY

	NN_TASSERT_(!(len%2));

	//	l̏
	predsample = state->prevsample;
	index = state->previndex;
	step = StepsizeTable[index];
	bufferstep = 0;

	for ( ; len > 0 ; len-- )
	{
		//	4bit delta ֓
		if ( !bufferstep )
		{
			inputbuffer = *indata++;
			delta = inputbuffer & 0xf;
		}
		else
		{
			delta = (inputbuffer >> 4) & 0xf;
		}
		bufferstep = !bufferstep;

		//	̃CfbNX߂
		index += IndexTable[delta];
		if ( index < 0 )
		{
			index = 0;
		}
		if ( index > (INDEX_NUM-1) )
		{
			index = (INDEX_NUM-1);
		}

		//	\ľvZ
		sign = delta & 8;		//	SbitځE
		delta = delta & 7;		//	ʂRbitE\l

		//	vZ̊Tv
		//	diffq = ( delta + 0.5 ) * step / 4
		diffq = step >> 3;
		if ( delta & 4 )
		{
			diffq += step;
		}
		if ( delta & 2 )
		{
			diffq += step>>1;
		}
		if ( delta & 1 )
		{
			diffq += step>>2;
		}
		
		if ( sign )
		{
			predsample -= diffq;
		}
		else
		{
			predsample += diffq;
		}

		//	\l͈͓ɌŒ
		if ( predsample > 32767 )
		{
			predsample = 32767;
		}
		else if ( predsample < -32768 )
		{
			predsample = -32768;
		}
		
		//	XebvTCY̍XV
		step = StepsizeTable[index];

		//	PCMŕۑ
		*outdata++ = (s16)predsample;
	}
	
	state->prevsample = (s16)predsample;
	state->previndex = (s8)index;
	
	return outdatasize;
}


#if defined(SNDIO_CONVERT_TO_ADPCM)
/**-----------------------------------------------------------------------------

	@brief		Xe[g̏s܂

	@param[in]	state			[UmۂXe[g̃|C^
	@param[out]	sample0			GR[hoblg`̍ŏ̃Tv̒l
	@return		Ȃ

----------------------------------------------------------------------------- */
static void InitADPCM( ADPCMstate* state, s16 sample0 )
{
	s32 index_ferret = 0;
	s32 min_diff = 65536;
	s16 encsample[2];
	s8 en_code[2];
	s16 de_code[2];
	s8 index;
	s32 diff;

	encsample[0] = sample0;
	encsample[1] = 1;

	for ( index = 0; index < INDEX_NUM ; index++ )
	{
		ADPCMstate state0;
		state0.prevsample = sample0;

		state0.previndex = index;
		(void)EncodeADPCM(encsample, en_code, 2, &state0);

		state0.previndex = index;
		(void)DecodeADPCM(en_code, de_code, 2, &state0);

		diff = de_code[0] - sample0;

		if (diff < 0)
		{
			diff = -diff;
		}

		if ( diff < min_diff )
		{
			min_diff = diff;
			index_ferret = index;
		}
	}

	state->prevsample = 0;
	state->previndex = (s8)index_ferret;
}
#endif


/*!
	@brief		TEhIO֘A`
*/
namespace sndio { namespace detail {


using namespace sndio;


//! ϊƂ̃[U[
static wchar_t				_userName[ 11 ] = {0};

//! ϊƂ̃Av
static wchar_t				_appName[ SNDIO_CONVERT_LENGTH ] = {0};

//! ϊɎgptR[obN
static DateTimeCallback		_callback = NULL;


/**-----------------------------------------------------------------------------

	@brief		ϊɎg郆[U[̐ݒ
	@param[in]	userName		[U[(11AI[܂)

----------------------------------------------------------------------------- */
void SetUserName( const wchar_t* userName )
{
	std::memcpy( _userName, userName, sizeof(_userName) );
}


/**-----------------------------------------------------------------------------

	@brief		ϊɎgAvP[V̐ݒ
	@param[in]	appName			AvP[V(128AI[܂)

----------------------------------------------------------------------------- */
void SetAppTitle( const wchar_t* appName )
{
	std::memcpy( _appName, appName, sizeof(_appName) );
}


/**-----------------------------------------------------------------------------

	@brief		ϊɎgtϊR[obN̐ݒ
	@param[in]	pCallback		tϊR[obN

----------------------------------------------------------------------------- */
void SetDateTimeCallback( DateTimeCallback pCallback )
{
	_callback = pCallback;
}


/**-----------------------------------------------------------------------------

	@brief		m4aADPCM֕ϊ
	@param[in]	manager			ǗNX
	@param[in]	allocator		AP[^[
	@param[in]	voice			CTR̈批f[^
	@param[in]	twlVoice		TWL̈批f[^
	@retval		true 
	@retval		false s
	@par		[]
				CTR̈̃f[^(m4a)TWL̈̃f[^(ADPCM)
				ϊs܂

----------------------------------------------------------------------------- */
bool ConvertToADPCM(
	VoiceManager& manager,
	nn::fnd::IAllocator& allocator,
	const VoiceData* voice,
	const TwlVoiceData* twlVoice )
{
#if defined(SNDIO_CONVERT_TO_ADPCM)
	nn::Result result;

	detail::DecoderAACArm decoder;
	nn::fs::FileInputStream inStream;

	if (voice == NULL || twlVoice == NULL)
	{
		return false;
	}

	VoiceManager::VoicePath srcPath, dstPath;
	if (!manager.GetVoicePath( srcPath, voice->GetPage(), voice->GetIndex() ))
	{
		return false;
	}
	if (!manager.GetVoicePath( dstPath, twlVoice->GetPage(), twlVoice->GetIndex() ))
	{
		return false;
	}

	sndio::detail::ToReal( srcPath );
	sndio::detail::ToReal( dstPath );

	result = inStream.TryInitialize( srcPath );
	if (result.IsFailure())
	{
		return false;
	}

	ExtendTag extTag;
	if (decoder.Initialize( &allocator, &inStream, &extTag ) != sndio::ResultSuccess)
	{
		return false;
	}

	// st@C`FbNs
	if (decoder.GetTotalTime() < 128)
	{
		return false;
	}
	else if (10048 < decoder.GetTotalTime())
	{
		return false;
	}
	else if (decoder.GetSamplingRate() != 16000)
	{
		return false;
	}
	else if (decoder.GetChannelCount() != 1)
	{
		return false;
	}

	u32 totalSamples = (decoder.GetTotalTime() / 1000 + 1) * decoder.GetSamplingRate();
	totalSamples = nn::math::RoundUp(totalSamples, 1024);

	s16* pOutBuf = (s16*)allocator.Allocate( totalSamples * sizeof(s16), 32 );
	s16* pOutBufR = (s16*)allocator.Allocate( 1024 * sizeof(s16), 32 );

	std::memset( pOutBuf, 0, totalSamples * sizeof(s16) );

	s32 resultSamples = 0;
	s32 offsetSamples = 0;

	do
	{
		decoder.TryRead( &resultSamples, pOutBuf + offsetSamples, pOutBufR, 1024 );
		offsetSamples += resultSamples;

	} while (0 < resultSamples);

	offsetSamples += decoder.GetFrameSamples();
	NN_TASSERT_( offsetSamples <= totalSamples );
	totalSamples = offsetSamples;

	// _~[ŎgĂEobt@͉
	allocator.Free( pOutBufR );

	decoder.Finalize();
	inStream.Finalize();

	Resampler res;
	s32 inSamples = totalSamples;
	s32 outSamples = c16000to16360( inSamples );
	outSamples = outSamples + (outSamples & 1);
	NN_ASSERT( inSamples < outSamples );

	s16* pOutBuf2 = (s16*)allocator.Allocate( outSamples * sizeof(s16), 32 );
	res.Resample( pOutBuf2, outSamples, pOutBuf, inSamples );

	// ̃obt@͊J
	allocator.Free( pOutBuf );

	SndSaveWave* pWave = reinterpret_cast<SndSaveWave*>( allocator.Allocate(sizeof(SndSaveWave), 32) );
	std::memset( pWave, 0, sizeof(SndSaveWave) );

	// Tvf[^̐擪ADPCM̏܂߂
	ADPCMstate* pStartState = reinterpret_cast<ADPCMstate*>(pWave->samples);
	std::memset( pStartState, 0x00, sizeof(ADPCMstate) );

	// ADPCM̏
	InitADPCM( pStartState, pOutBuf2[0] );

	ADPCMstate tempState = *pStartState;
	s8* outdata = reinterpret_cast<s8*>( &pWave->samples[ sizeof(ADPCMstate) ] );

	// ADPCMփGR[h
	u32 arcSize = EncodeADPCM( pOutBuf2, outdata, outSamples, &tempState );
	allocator.Free( pOutBuf2 );

	NN_TASSERT_(arcSize != 0);

	pWave->param.format		= SND_WAVE_FORMAT_ADPCM;
	pWave->param.loopflag	= false;
	pWave->param.rate		= REC_SAMPLING_RATE;
	pWave->param.timer		= SND_TIMER_CLOCK / REC_SAMPLING_RATE;
	pWave->param.loopstart	= 0;
	pWave->param.looplen	= arcSize / 4;		// 1/2ƁAĐԂ{

	// t@C֏o
	nn::fs::FileOutputStream outStream;
	result = outStream.TryInitialize( dstPath, true );
	if (result.IsFailure())
	{
		NN_DBG_TPRINT_RESULT_( result );
		// t@CI[vɎs
		return false;
	}

	// 
	std::memset( pWave->info, 0, sizeof(pWave->info) );

	// CRCvZ(I[2oCgCRC͏Ȃ)
	pWave->crc = nn::util::Crc16::Calculate( pWave, sizeof(SndSaveWave) - 2 );

	s32 writeCount = 0;
	result = outStream.TryWrite( &writeCount, pWave, sizeof(SndSaveWave), true );

	allocator.Free( pWave );
	if (result.IsFailure())
	{
		NN_DBG_TPRINT_RESULT_( result );
		// ݂Ɏs
		return false;
	}
	outStream.Finalize();

#else
	NN_UNUSED_VAR( manager );
	NN_UNUSED_VAR( allocator );
	NN_UNUSED_VAR( voice );
	NN_UNUSED_VAR( twlVoice );
#endif

	return true;
}


/**-----------------------------------------------------------------------------

	@brief		TWL̈̃f[^(ADPCM)CTR̈̃f[^(m4a)֕ϊ
	@param[in]	manager			ǗNX
	@param[in]	allocator		AP[^[
	@param[in]	twlVoice		TWL̈批f[^
	@param[in]	voice			CTR̈批f[^
	@retval		true 
	@retval		false s
	@par		[]
				TWL̈̃f[^(ADPCM)CTR̈̃f[^(m4a)֕ϊ
				s܂

----------------------------------------------------------------------------- */
bool ConvertToM4A(
	VoiceManager& manager,
	nn::fnd::IAllocator& allocator,
	const TwlVoiceData* twlVoice,
	const VoiceData* voice )
{
	nn::Result result;

	if (voice == NULL || twlVoice == NULL )
	{
		return false;
	}

	VoiceManager::VoicePath srcPath, dstPath;
	if (!manager.GetVoicePath( srcPath, twlVoice->GetPage(), twlVoice->GetIndex() ))
	{
		return false;
	}
	if (!manager.GetVoicePath( dstPath, voice->GetPage(), voice->GetIndex() ))
	{
		return false;
	}
	sndio::detail::ToReal( srcPath );
	sndio::detail::ToReal( dstPath );

	// dat(ADPCM)t@Cǂݍ
	nn::fs::FileInputStream inStream;
	result = inStream.TryInitialize( srcPath );
	if (result.IsFailure())
	{
		NN_DBG_TPRINT_RESULT_( result );
		return false;
	}

	s32 readCount = 0;

	SndSaveWave* pWave = reinterpret_cast<SndSaveWave*>( allocator.Allocate(sizeof(SndSaveWave), 32) );

	result = inStream.TryRead( &readCount, pWave, sizeof(SndSaveWave) );
	if (result.IsFailure())
	{
		NN_DBG_TPRINT_RESULT_( result );
		allocator.Free( pWave );
		return false;
	}

	// ADPCMfR[h̃obt@TCY(bytes)̗̈m
	s16* pOutBuf = (s16*)allocator.Allocate( pWave->param.looplen * 16, 32 );
	std::memset( pOutBuf, 0, pWave->param.looplen * 16 );

	ADPCMstate state;
	state = *reinterpret_cast<ADPCMstate*>( pWave->samples );

	// ADPCMPCM֕ϊ
	u32 bufferBytes = DecodeADPCM(
		(s8*)&pWave->samples[ sizeof(ADPCMstate) ],
		pOutBuf,
		pWave->param.looplen * 8,	// fR[h̃Tv
		&state );

	Resampler res;
	s32 inSamples = s32(bufferBytes / sizeof(s16));
	s32 outSamples = c16360to16000( inSamples );
	outSamples = outSamples + (outSamples & 1);

	s16* pOutBuf2 = (s16*)allocator.Allocate( outSamples * sizeof(s16), 32 );

	NN_TASSERT_( outSamples < inSamples );
	res.Resample( pOutBuf2, outSamples, pOutBuf, inSamples );
	allocator.Free( pOutBuf );

	// oPCMM4At@C֕ϊď
	VoiceWriter enc( allocator );

	nn::fnd::DateTime dateTime = twlVoice->GetDateTime();

	if (_callback)
	{
		wchar_t titleName[ SNDIO_CONVERT_LENGTH ];
		(*_callback)( titleName, dateTime );
		enc.SetTitleName( titleName );
	}
	else
	{
		// R[obNw肳ĂȂ
		allocator.Free( pOutBuf2 );
		return false;
	}

	enc.SetAlbumName( _appName );
	enc.SetArtistName( _userName );
	enc.SetCreateDate( dateTime.GetParameters() );

	// g^Oݒ
	{
		// Ro[gɎg\tgEFAID͌Œ
		u8 sftw[5];
		sftw[0] = '0';
		sftw[1] = '0';
		sftw[2] = '2';
		sftw[3] = '0';
		sftw[4] = '5';

		u8 body[8];
		sndio::GetBodyId( body );
		enc.SetSFTW( sftw );
		enc.SetRCID( body );
		enc.SetEDID( body );
	}

	sndio::Result sresult = enc.StartStream( dstPath );
	if (sresult != sndio::ResultSuccess)
	{
		allocator.Free( pWave );
		allocator.Free( pOutBuf2 );
		return false;
	}

	s32 writeCount = 0;
	for (u32 offset = 0; offset < outSamples; offset += 1024)
	{
		sresult = enc.TryWrite( &writeCount, pOutBuf2 + offset, 1024, false );
		if (sresult != sndio::ResultSuccess)
		{
			NN_DBG_TPRINT_RESULT_( result );
			// G[
			break;
		}
	}

	sresult = enc.EndStream();
	if (sresult != sndio::ResultSuccess)
	{
		allocator.Free( pWave );
		allocator.Free( pOutBuf2 );
		return false;
	}

	allocator.Free( pWave );
	allocator.Free( pOutBuf2 );
	return true;
}


}}
