/**-----------------------------------------------------------------------------

	@file		sndio_Decoder.cpp
	@author		tetsu
	@par		[]
				fR[hsC^tF[XNX(J)

				Copyright (c)2010 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#include <time.h>
#include <nn.h>
#if defined(USE_MP3)
#include <tmc_id3parse.h>
#endif
#include "sndio_UtfConv.h"
#include "sndio_Decoder.h"


////////////////////////////////////////////////////////////////////////////////
// [J`
////////////////////////////////////////////////////////////////////////////////

namespace
{
	//! TvO[g
	static const s32 samplingTable[9] =
	{
		48000,
		44100,
		32000,
		24000,
		22050,
		16000,
		12000,
		11025,
		8000
	};

#if defined(USE_MP3)
	//! mp3 sampling rate index [ver][rate]
	static const s32 samplingRateTable[3][3] =
	{
		{ 1, 0, 2 },
		{ 4, 3, 5 },
		{ 7, 6, 8 }
	};

	//! mp3 bitrate [layer][rate]
	static const short mp3bitrate[5][16] =
	{
		// MPEG1 LayerT = 0
		{ 0, 32,  64,  96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0 },

		// MPEG1 LayerU = 1
		{ 0, 32,  48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, 0 },

		// MPEG1 LayerV = 2
		{ 0, 32,  40,  48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 0 },

		// MPEG2 LayerT = 3
		{ 0, 32,  48,  56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256, 0 },

		// MPEG2 LayerUV = 4
		{ 0,  8,  16,  24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 }
	};

	//! mp3 frame size [ver][freq][rate]
	static const short mp3SizeTab[3][3][16] =
	{
	{
		// 44.1 kHz
		{ 0, 104, 130, 156, 182, 208, 261, 313, 365, 417, 522,  626,  731,  835, 1044, 0 },
		// 48 kHz
		{ 0,  96, 120, 144, 168, 192, 240, 288, 336, 384, 480,  576,  672,  768,  960, 0 },
		// 32 kHz
		{ 0, 144, 180, 216, 252, 288, 360, 432, 504, 576, 720,  864, 1008, 1152, 1440, 0 }
	},
	{
		// 22.05 kHz
		{ 0,  26,  52,  78, 104, 130, 156, 182, 208, 261, 313,  365,  417,  470,  522, 0 },
		// 24 kHz
		{ 0,  24,  48,  72,  96, 120, 144, 168, 192, 240, 288,  336,  384,  432,  480, 0 },
		// 16 kHz
		{ 0,  36,  72, 108, 144, 180, 216, 252, 288, 360, 432,  504,  576,  648,  720, 0 }
	},
	{
		// 11.025 kHz
		{ 0,  52, 104, 156, 208, 261, 313, 365, 417, 522, 626,  731,  835,  940, 1044, 0 },
		// 12 kHz
		{ 0,  48,  96, 144, 192, 240, 288, 336, 384, 480, 576,  672,  768,  864,  960, 0 },
		// 8 kHz
		{ 0,  72, 144, 216, 288, 360, 432, 504, 576, 720, 864, 1008, 1152, 1296, 1440, 0 }
	}
	};
#endif
}


namespace sndio { namespace detail {


////////////////////////////////////////////////////////////////////////////////
// fR[_[[eBeB`
////////////////////////////////////////////////////////////////////////////////


/**-----------------------------------------------------------------------------

	@brief		TvO[g擾
	@param[in]	rateIdx			[gCfbNX
	@return		擾łTEh^CvԂ܂
	@par		[]
				[gCfbNXTvO[g擾

----------------------------------------------------------------------------- */
int GetSamplingRateIdx( int rateIdx )
{
	NN_MINMAX_TASSERT_( rateIdx, 0, 8 );

	if (rateIdx < 0) { rateIdx = 0; }
	if (8 < rateIdx) { rateIdx = 8; }

	return samplingTable[ rateIdx ];
}


/**-----------------------------------------------------------------------------

	@brief		mvhd(fBAi[){bNX쐬o

	@param[out]	dateTime		쐬
	@param[in]	mvhdBox			mvhd{bNX擪f[^(32Byte)
	@return		擾ɐtrue, 擾Ɏsfalse

----------------------------------------------------------------------------- */
bool M4ADemux_GetCreationTime( nn::fnd::DateTime& dateTime, const u8 mvhdBox[32] )
{
	const u8* tag = &mvhdBox[0];

	if (tag[0] == 'm' && tag[1] == 'v' && tag[2] == 'h' && tag[3] == 'd')
	{
		s64 creationTime = s64(
			(mvhdBox[ 8] << 24) |
			(mvhdBox[ 9] << 16) |
			(mvhdBox[10] <<  8) |
			(mvhdBox[11]) );

		// 1970/1/1̌oߕbvZ
		std::time_t t = static_cast<std::time_t>( creationTime - 0x7C25B080ll );

		const std::tm* tt = std::localtime( &t );

		dateTime = nn::fnd::DateTime(
			tt->tm_year + 1900,
			tt->tm_mon + 1,
			tt->tm_mday,
			tt->tm_hour,
			tt->tm_min,
			tt->tm_sec );

		return true;
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		AACf[^TCY̌vZ(ADTSp)
	@param[in]	in				AACf[^obt@̃AhX(t[擪)
	@return		AACf[^TCY(Pt[)
	@par		[]
				AACf[^̃TCYoCgŕԂ܂<br>
				AACf[^ADTS`łȂꍇA0Ԃ܂

----------------------------------------------------------------------------- */
int calc_aac_size( const u8* in )
{
	register u32 sync, len;
	register u32 h1, h2;

	h1  = in[0];	h1 <<=8;
	h1 |= in[1];	h1 <<=8;
	h1 |= in[2];	h1 <<=8;
	h1 |= in[3];

	h2  = in[4];	h2 <<=8;
	h2 |= in[5];	h2 <<=8;
	h2 |= in[6];	h2 <<=8;
	h2 |= in[7];

	sync = (h1 >> 20) & 0xfff;
	if (sync != 0xfff)
		return 0;

	len = ((((h1>>0)&0x003)<<11) | ((h2>>21)&0x7ff));
	return (int)len;
}


/**-----------------------------------------------------------------------------

	@brief		AACTvO[gCfbNX擾(ADTSp)
	@param[in]	in				AACf[^obt@̃AhX(t[擪)
	@return		AACTvO[gCfbNX
	@par		[]
				AACTvO[gCfbNX擾܂<br>
				AACf[^ADTS`łȂꍇA0Ԃ܂

----------------------------------------------------------------------------- */
int calc_aac_sampleidx( const u8* in )
{
	u32 sync;
	u32 h1, h2;

	h1  = in[0];	h1 <<=8;
	h1 |= in[1];	h1 <<=8;
	h1 |= in[2];	h1 <<=8;
	h1 |= in[3];

	h2  = in[4];	h2 <<=8;
	h2 |= in[5];	h2 <<=8;
	h2 |= in[6];	h2 <<=8;
	h2 |= in[7];

	sync = (h1 >> 20) & 0xfff;

	if (sync != 0xfff)
		return 0;

	return (int)((h1 >> 10) & 0x00f);
}


/**-----------------------------------------------------------------------------

	@brief		AAC`l擾(ADTSp)
	@param[in]	in				AACf[^obt@̃AhX(t[擪)
	@return		AAC`l
	@par		[]
				AAC`l擾܂<br>
				AACf[^ADTS`łȂꍇA0Ԃ܂

----------------------------------------------------------------------------- */
int calc_aac_channel( const u8* in )
{
	u32 sync;
	u32 h1, h2;

	h1  = in[0];	h1 <<=8;
	h1 |= in[1];	h1 <<=8;
	h1 |= in[2];	h1 <<=8;
	h1 |= in[3];

	h2  = in[4];	h2 <<=8;
	h2 |= in[5];	h2 <<=8;
	h2 |= in[6];	h2 <<=8;
	h2 |= in[7];

	sync = (h1 >> 20) & 0xfff;
	if (sync != 0xfff)
		return 0;

	return (int)((h1 >> 6) & 0x007);
}


#if defined(USE_MP3)
/**-----------------------------------------------------------------------------

	@brief		MP3f[^TCY̎Zo
	@param[in]	in				MP3f[^obt@̃AhX(t[擪)
	@return		MP3f[^TCY(Pt[)
	@par		[]
				MP3f[^̃TCYoCgŕԂ܂

----------------------------------------------------------------------------- */
int calc_mp3_size( const u8* in )
{
	unsigned long sync, id, rate, fs, padd, len;
	register unsigned long h1, ver;

	h1  = in[0]; h1 <<=8;
	h1 |= in[1]; h1 <<=8;
	h1 |= in[2]; h1 <<=8;
	h1 |= in[3];

	sync  = (h1 >> 20) & 0xfff;
	id    = (h1 >> 19) & 0x001;
	rate  = (h1 >> 12) & 0x00f;
	fs    = (h1 >> 10) & 0x003;
	padd  = (h1 >>  9) & 0x001;

	if ((sync==0xfff) && (id == 1))
	{
		ver = 0;	/** MPEG-1 */
	}
	else if ((sync==0xfff) && (id == 0))
	{
		ver = 1;	/** MPGE-2 LSF */
	}
	else
	{
		ver = 2;	/** MPEG-2.5 */
	}

	len = mp3SizeTab[ver][fs][rate];
	len += padd;

	if ((sync != 0xfff) && (sync != 0xffe))
	{
		return 0;
	}

	return (int)len;
}


/**-----------------------------------------------------------------------------

	@brief		MP3TvO[g擾
	@param[in]	in				MP3f[^obt@̃AhX(t[擪)
	@return		MP3TvO[gCfbNX
	@par		[]
				MP3TvO[gCfbNXԂ܂

----------------------------------------------------------------------------- */
int calc_mp3_sampleidx( const u8* in )
{
	unsigned long sync, id, fs;
	unsigned long h1, ver;

	h1  = in[0]; h1 <<=8;
	h1 |= in[1]; h1 <<=8;
	h1 |= in[2]; h1 <<=8;
	h1 |= in[3];

	sync  = (h1 >> 20) & 0xfff; // 0-11
	id    = (h1 >> 19) & 0x001; // 11-13
	fs    = (h1 >> 10) & 0x003; // 32 - 10 = 10

	if ((sync==0xfff) && (id == 1))
	{
		ver = 0;	/** MPEG-1 */
	}
	else if ((sync==0xfff) && (id == 0))
	{
		ver = 1;	/** MPGE-2 LSF */
	}
	else
	{
		ver = 2;	/** MPEG-2.5 */
	}

	if ((sync != 0xfff) && (sync != 0xffe))
	{
		return 0;
	}

	return samplingRateTable[ver][fs];
}


/**-----------------------------------------------------------------------------

	@brief		MP3`l擾
	@param[in]	in				MP3f[^obt@̃AhX(t[擪)
	@return		MP3`l
	@par		[]
				MP3`l擾܂

----------------------------------------------------------------------------- */
int calc_mp3_channel( const u8* in )
{
	unsigned long sync, mode;
	unsigned long h1;

	h1  = in[0]; h1 <<=8;
	h1 |= in[1]; h1 <<=8;
	h1 |= in[2]; h1 <<=8;
	h1 |= in[3];

	sync  = (h1 >> 20) & 0xfff; // 0-11
	mode  = (h1 >>  6) & 0x003; // channel mode

	if ((sync != 0xfff) && (sync != 0xffe))
	{
		return 0;
	}

	return ((mode & 0x003) == 0x003) ? 1 : 2;
}


/**-----------------------------------------------------------------------------

	@brief		MP3TChTCY擾
	@param[in]	in				MP3f[^obt@̃AhX(t[擪)
	@return		MP3TChTCY
	@par		[]
				MP3TChTCY擾擾܂

----------------------------------------------------------------------------- */
int calc_mp3_side( const u8* in )
{
	unsigned long sync, id, mode;
	unsigned long h1, ver;

	h1  = in[0]; h1 <<=8;
	h1 |= in[1]; h1 <<=8;
	h1 |= in[2]; h1 <<=8;
	h1 |= in[3];

	sync  = (h1 >> 20) & 0xfff; // 0-11
	id    = (h1 >> 19) & 0x001; // 11-13
	mode  = (h1 >>  6) & 0x003; // channel mode

	if ((sync != 0xfff) && (sync != 0xffe))
	{
		return 0;
	}

	if ((sync==0xfff) && (id == 1))
	{
		ver = 0;	/** MPEG-1 */
	}
	else if ((sync==0xfff) && (id == 0))
	{
		ver = 1;	/** MPGE-2 LSF */
	}
	else
	{
		ver = 1;	/** MPEG-2.5 */
	}

	static const int side[2][2] =
	{
		{ 32, 17 },	/** MPEG 1 */
		{ 17, 9 }	/** MPEG 2 LSF/2.5 */
	};
	return side[ ver ][ ((mode & 0x003) == 0x003) ? 1 : 0 ];
}


/**-----------------------------------------------------------------------------

	@brief		MP3rbg[g擾
	@param[in]	in				MP3f[^obt@̃AhX(t[擪)
	@return		MP3TChTCY
	@par		[]
				MP3rbg[g擾܂

----------------------------------------------------------------------------- */
int calc_mp3_bitrate( const u8* in )
{
	unsigned long sync, id, rate, layer;
	register unsigned long h1, idx;

	h1  = in[0]; h1 <<=8;
	h1 |= in[1]; h1 <<=8;
	h1 |= in[2]; h1 <<=8;
	h1 |= in[3];

	sync  = (h1 >> 20) & 0xfff; // 0-11
	id    = (h1 >> 19) & 0x001; // 11-13
	layer = (h1 >> 17) & 0x003; // 13-15
	rate  = (h1 >> 12) & 0x00f; // 20-22

	if ((sync != 0xfff) && (sync != 0xffe))
	{
		return 0;
	}

	if ((sync==0xfff) && (id == 1))
	{
		/* MPEG-1 */
		if (layer == 0x03)
		{
			// LayerT
			idx = 0;
		}
		else if (layer == 0x02)
		{
			// LayerU
			idx = 1;
		}
		else
		{
			// LayerV
			idx = 2;
		}
	}
	else
	{
		/* MPGE-2 LSF */
		/* MPEG-2.5 */

		if (layer == 0x03)
		{
			// LayerT
			idx = 3;
		}
		else
		{
			// LayerUV
			idx = 4;
		}
	}

	return mp3bitrate[idx][rate];
}


/**-----------------------------------------------------------------------------

	@brief		mp3wb_L
	@return		t[TCY
	@par		[]
				mp3wb_̔sALȂt[TCYԂ܂B<br/>
				Ȃ0Ԃ܂

----------------------------------------------------------------------------- */
static int CheckMp3Header( unsigned char* buf )
{
	int syncX   = ( ( int )buf[0] << 3 ) | ( ( int )buf[1] >> 5 );	// 11 bit  syncword
	int syncY   = ( ( int )buf[1] >> 4 ) &  1;	//  1
	int id      = ( ( int )buf[1] >> 3 ) &  1;	//  1      ID
	int layer   = ( ( int )buf[1] >> 1 ) &  3;	//  2      layer
//	int noCrc   = ((int)buf[1] >> 0) &  1;		//  1      protection_bit
	int brIndex = ( ( int )buf[2] >> 4 ) & 15;	//  4      bitrate_index
	int fsIndex = ( ( int )buf[2] >> 2 ) &  3;	//  2      sampling_frequency
	int padding = ( ( int )buf[2] >> 1 ) &  1;	//  1      padding_bit
//	int priv    = ((int)buf[2] >> 0) &  1;		//  1      private_bit
//	int mode    = ((int)buf[3] >> 6) &  3;		//  2      mode
//	int modeExt = ((int)buf[3] >> 4) &  3;		//  2      mode_extension
//	int copy    = ((int)buf[3] >> 3) &  1;		//  1      copyright
//	int orig    = ((int)buf[3] >> 2) &  1;		//  1      original_copy
	int emph    = ( ( int )buf[3] >> 0 ) &  3;	//  1      emphasis

	if (( syncX == 0x07FF ) &&
		( layer == 1 ) &&
		( brIndex != 0 ) && ( brIndex != 15 ) &&
		( fsIndex != 3 ) &&
		( emph != 2 ))
	{
		int size;

		if ( syncY == 0 )
		{
			size = mp3SizeTab[2][fsIndex][brIndex]; // MPEG2.5 (Quarter Fs)
		}
		else if ( id == 0 )
		{
			size = mp3SizeTab[1][fsIndex][brIndex]; // MPEG-2  (Half Fs)
		}
		else
		{
			size = mp3SizeTab[0][fsIndex][brIndex]; // MPEG-1  (Normal Fs)
		}

		if ( padding != 0 )
		{
//  if ((fsIndex == 0) && (padding == 1)) {
			size += 1;
		}

		return ( size ); // Return frame length in byte
	}
	else
	{
		return ( 0 );  // Return 0 for invalid Mp3 header
	}
}


/**-----------------------------------------------------------------------------

	@brief		L
	@param[in]	seekPoint		V[NJnʒu
	@param[in]	stream			f[^Xg[
	@return		true MP3wb_ƂėL
	@par		[]
				MP3wb_L肵܂

----------------------------------------------------------------------------- */
bool valid_mp3_header( u32 seekPoint, nn::fs::IInputStream* stream )
{
	NN_NULL_TASSERT_( stream );

	static const u32 checkNum = 4;
	static const u32 needMinSize = 26;

	nn::Result result;
	s32 readCount = 0;
	u32 frameSize = 0;
	u8 mp3header[checkNum];
	u8 sample[checkNum];
	u8 channel[checkNum];

	result = stream->TrySeek( seekPoint, nn::fs::POSITION_BASE_BEGIN );
	if (result.IsFailure())
	{
		return false;
	}

	for (int i = 0; i < checkNum; i++)
	{
		result = stream->TrySeek( stream->GetPosition() + frameSize, nn::fs::POSITION_BASE_BEGIN );
		if (result.IsFailure())
		{
			return false;
		}

		result = stream->TryRead( &readCount, &mp3header, sizeof(mp3header) );
		if (result.IsFailure() || readCount != sizeof(mp3header))
		{
			return false;
		}

		frameSize = calc_mp3_size( mp3header );
		if (frameSize < needMinSize)
			return false;

		frameSize -= sizeof(mp3header);
		channel[i] = calc_mp3_channel( mp3header );
		sample[i] = calc_mp3_sampleidx( mp3header );

		if (CheckMp3Header( mp3header ) <= 0)
			return false;
	}

	if ((sample[0] != sample[1]) |
		(sample[2] != sample[3]) |
		(sample[3] != sample[0]))
	{
		return false;
	}

	if ((channel[0] != channel[1]) |
		(channel[2] != channel[3]) |
		(channel[3] != channel[0]))
	{
		return false;
	}

	return true;
}


/**-----------------------------------------------------------------------------

	@brief		MP3̓t[T
	@param[in]	seekPoint		V[NJnʒu
	@param[in]	stream			f[^Xg[
	@return		t[ItZbg(ꍇ u32(-1) )
	@par		[]
				MP3̓t[t@C|WVԂ܂<br>
				ɂׂ͕܂

----------------------------------------------------------------------------- */
u32 calc_mp3_findsync( u32 seekPoint, nn::fs::IInputStream* stream )
{
	static u8 syncBuffer[ 1441 ];	// XbhZ[tł͂Ȃ
	nn::Result result;
	s32 readCount = 0;

	NN_NULL_TASSERT_( stream );

	if (stream == NULL)
		return u32(-1);

	u32 prevPosition = stream->GetPosition();

	result = stream->TrySeek( seekPoint, nn::fs::POSITION_BASE_BEGIN );
	if (result.IsFailure())
		return u32(-1);

	result = stream->TryRead( &readCount, &syncBuffer, 1441 );
	if (result.IsFailure())
		return u32(-1);

	u32 i;
	for (i = 0; i < readCount; i++)
	{
		if (syncBuffer[i] == 0xFF && (syncBuffer[i+1] & 0xE0) == 0xE0)
		{
			if (valid_mp3_header( i + seekPoint, stream ))
			{
				result = stream->TrySeek( prevPosition, nn::fs::POSITION_BASE_BEGIN );
				if (result.IsFailure())
					return u32(-1);

				return (seekPoint + i);
			}
		}
	}

	result = stream->TrySeek( prevPosition, nn::fs::POSITION_BASE_BEGIN );
	if (result.IsFailure())
		return u32(-1);

	return seekPoint;
}


/**-----------------------------------------------------------------------------

	@brief		ID3^OXLbv擾
	@param[in]	stream			f[^Xg[
	@param[out]	pHeaderOfs		wb_ItZbg(Ȃ0)
	@param[out]	pFooterSize		tb^TCY(Ȃ0)
	@return		Xg[TCY
	@par		[]
				ID3^O͂Awb_TCYƃtb^TCY擾܂<br>
				ꂼꑶ݂Ȃꍇ0Ԃ܂

----------------------------------------------------------------------------- */
s32 id3_tag_skip( nn::fs::IInputStream* stream, u32* pHeaderOfs, u32* pFooterSize )
{
	NN_NULL_TASSERT_( stream );
	NN_NULL_TASSERT_( pHeaderOfs );
	NN_NULL_TASSERT_( pFooterSize );

	if (pHeaderOfs)
	{
		*pHeaderOfs = 0;
	}

	if (pFooterSize)
	{
		*pFooterSize = 0;
	}

	//tb^̃t@CǍ݃TCY
	static const u32 FOOTER_SEARCH_SIZE = (1024*10);
	static const u32 id3v1size = 128;

	nn::Result		result;
	s32				readCount;

	s32				streamSize = 0;
	unsigned char	TagHeader[10] = {0};

	int					id3result;
	TMC_ID3GetTag		ID3Tag;				// ^Owb_\
	TMC_ID3GetTagInfo	ID3Info;			// ^O\

	streamSize = stream->GetSize();

	static u8 Id3buffer[FOOTER_SEARCH_SIZE] = {0};
	result = stream->TrySeek( stream->GetSize() - id3v1size, nn::fs::POSITION_BASE_BEGIN );
	if (result.IsFailure())
	{
		return -1;
	}

	result = stream->TryRead( &readCount, Id3buffer, id3v1size);
	if (result.IsFailure() || readCount != id3v1size)
	{
		return streamSize;
	}

	// ID3v1.x
	id3result = 0;

	id3result = TMC_ID3Parse_V1GetTagInfo( &ID3Tag, &ID3Info, (const unsigned char *)Id3buffer, 128 );

	if (id3result == TMC_ID3PARSE_SUCCESS)
	{
		*pFooterSize = ID3Tag.TagSize;
		streamSize -= ID3Tag.TagSize;
	}

	// ID3v2.x
	result = stream->TrySeek( 0, nn::fs::POSITION_BASE_BEGIN );
	if (result.IsFailure())
	{
		return -1;
	}

	// mp3t@C̐擪10oCg
	result = stream->TryRead( &readCount, TagHeader, 10);
	if (result.IsFailure() || readCount != 10)
	{
		return streamSize;
	}

	// ID3v2.xwb_
	id3result = TMC_ID3Parse_V2GetTagHeader(&ID3Tag, (const unsigned char*)TagHeader, 10);

	// ID3^O̎擾
	if (id3result == TMC_ID3PARSE_SUCCESS)
	{
		if (pHeaderOfs)
		{
			*pHeaderOfs += ID3Tag.TagSize + ID3Tag.FooterPresent * 10;
			streamSize -= ID3Tag.TagSize + ID3Tag.FooterPresent * 10;
		}
	}

	return streamSize;
}

#endif


}}
