/**-----------------------------------------------------------------------------

	@file		sndio_DecoderAACArm.cpp
	@author		tetsu
	@par		[]
				AACCPU(ARM11)ŃfR[hsNX(J)

				Copyright (c)2010 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#include "sndio_Decoder.h"
#include "sndio_DecoderAACArm.h"
#include "sndio_VoiceExtendTag.h"
#include "sndio_Util.h"

#include <nn/assert.h>
#include <tmc_aacdec.h>
#include <tmc_m4ademux.h>


////////////////////////////////////////////////////////////////////////////////
// [J`
////////////////////////////////////////////////////////////////////////////////

#if defined(APP_SNOTE) && defined(NN_BUILD_DEBUG)
#define _LOG( ... )		NN_LOG( __VA_ARGS__ )
#else
#define _LOG( ... )
#endif

namespace
{
	//! AACKvobt@TCY
	const u32 AAC_REQUIRE_BUFFER_SIZE = 1536;

	//! eudtaTCY
	const u32 c_udtaSize = 512 * 1024;			// 512KByte܂

	//! etrakTCY
	const u32 c_trakSize = 1024 * 1024 * 9;		// 9MByte܂
}


////////////////////////////////////////////////////////////////////////////////
// DecoderAACArm`
////////////////////////////////////////////////////////////////////////////////

namespace sndio { namespace detail {


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				RXgN^

----------------------------------------------------------------------------- */
DecoderAACArm::DecoderAACArm() :
m_allocator(NULL),
m_buffer(NULL),
m_samplingRate(44100),
m_channel(0),
m_totalTime(0),
m_seekTime(0),
m_frameCount(0),
m_handle(NULL),
m_trak(NULL),
m_trakSize(0),
m_heap(NULL),
m_aacdec(NULL),
m_param()
{
	std::memset( m_titleName, 0, sizeof(m_titleName) );
	std::memset( m_artistName, 0, sizeof(m_artistName) );
	std::memset( m_albumName, 0, sizeof(m_albumName) );
}


/**-----------------------------------------------------------------------------

	@brief		
	@return		Ȃ
	@par		[]
				fXgN^

----------------------------------------------------------------------------- */
DecoderAACArm::~DecoderAACArm()
{
	Finalize();
}


/**-----------------------------------------------------------------------------

	@brief		s܂
	@param[in]	allocator		AP[^|C^(foCX)
	@param[in]	filePath		t@CpX
	@return		ɐ true As false Ԃ܂B
	@par		[]
				AP[^̓foCX̊mۂłKv܂

----------------------------------------------------------------------------- */
Result DecoderAACArm::Initialize(
	nn::fnd::IAllocator* allocator,
	nn::fs::IInputStream* stream,
	ExtendTag* pExtTag )
{
	nn::Result result;
	s32 readCount = 0;

	m_seekTime = 0;

	if (m_allocator)
	{
		return ResultNotFinalize;
	}

	// AP[^ݒ
	m_allocator = allocator;

	NN_NULL_TASSERT_( stream );
	if (stream == NULL)
	{
		return ResultInvalidParameter;
	}
	m_stream  = stream;

	// moov
	int moovSize = 0;
	unsigned int moovOffset = 0;
	unsigned int readOffset = 0;

	do
	{
		unsigned int nextBox = 0;
		unsigned char tag[TMC_M4ADEMUX_BOX_TAG_SIZE] = {0};

		result = m_stream->TrySeek( readOffset, nn::fs::POSITION_BASE_BEGIN );
		if (result.IsFailure())
		{
			return GetFsWholeError( result );
		}

		result = m_stream->TryRead( &readCount, tag, TMC_M4ADEMUX_BOX_TAG_SIZE );
		if (result.IsFailure())
		{
			return GetFsWholeError( result );
		}

		readOffset += TMC_M4ADEMUX_BOX_TAG_SIZE;

		int status = -1;
		status = TMC_M4ADemux_SearchMoov(tag, &nextBox);

		if (TMC_M4ADEMUX_SEARCH_NEXT == status)
		{
			if (static_cast<int>(nextBox) < 0)
			{
				// ̒l͕s
				return ResultInvalidFile;
			}
			else
			{
				readOffset += nextBox;
			}
		}
		else if (status > 0)
		{
			moovSize = status;
			break;
		}
		else
		{
			// moovŃG[܂
			return ResultInvalidFile;
		}
	} while(1);

	// moovItZbgۑĂ
	moovOffset = readOffset;

	// udta
	u8* udta = NULL;
	u32 udtaSize = 0;

	readOffset = moovOffset;
	do
	{
		unsigned int nextBox = 0;
		unsigned char tag[TMC_M4ADEMUX_BOX_TAG_SIZE] = {0};

		result = m_stream->TrySeek( readOffset, nn::fs::POSITION_BASE_BEGIN );
		if (result.IsFailure())
		{
			return GetFsWholeError( result );
		}
		result = m_stream->TryRead( &readCount, tag, TMC_M4ADEMUX_BOX_TAG_SIZE );
		if (result.IsFailure())
		{
			return GetFsWholeError( result );
		}
		readOffset += TMC_M4ADEMUX_BOX_TAG_SIZE;

		int status = -1;
		status = TMC_M4ADemux_SearchUdta(tag, &nextBox);

		if (TMC_M4ADEMUX_SEARCH_NEXT == status)
		{
			if (static_cast<int>(nextBox) < 0)
			{
				// ̒l͕s
				return ResultInvalidFile;
			}
			else
			{
				readOffset += nextBox;
			}
		}
		else if (status > 0)
		{
			udtaSize = status;

			// ُȑ傫̃[U[f[^TCYȂXLbv
			if (c_udtaSize < udtaSize)
			{
				udtaSize = 0;
				break;
			}

			udta = (u8*)m_allocator->Allocate( udtaSize, 32 );
			result = m_stream->TryRead( &readCount, udta, udtaSize );
			if (result.IsFailure())
			{
				SafeFree( m_allocator, udta );
				return GetFsWholeError( result );
			}

			if (udta == NULL)
			{
				return ResultOutOfResource;
			}
			break;
		}
		else
		{
			// udtaŃG[܂
			return ResultInvalidFile;
		}
	} while (readOffset - moovOffset < moovSize);

	// ^O
	if (udta && 0 < udtaSize)
	{
		TMC_M4ATagInfo tInfo;
		TMC_M4ATagExtendInfo tagInfo[5];
		std::memset( tagInfo, 0x00, sizeof(tagInfo) );
		tagInfo[0].tagname = TAG_CODE;
		tagInfo[0].next = &tagInfo[1];
		tagInfo[1].tagname = TAG_SFTW;
		tagInfo[1].next = &tagInfo[2];
		tagInfo[2].tagname = TAG_RCID;
		tagInfo[2].next = &tagInfo[3];
		tagInfo[3].tagname = TAG_EDID;
		tagInfo[3].next = &tagInfo[4];
		tagInfo[4].tagname = TAG_PRMT;
		tagInfo[4].next = NULL;

		int status = TMC_M4ADemux_GetTagInfo(udta, udtaSize, &tInfo, tagInfo);

		std::memset( m_titleName, 0x00, sizeof(m_titleName) );
		if (status == TMC_M4ADEMUX_SUCCESS)
		{
			// g^O擾
			TMC_M4ATagExtendInfo* pTag = &tagInfo[0];

			while (pTag)
			{
				switch (pTag->tagname)
				{
				case TAG_CODE:
					if (sizeof(pExtTag->code) == pTag->value_len)
					{
						std::memcpy( pExtTag->code, pTag->value, sizeof(pExtTag->code) );
					}
					break;
				case TAG_SFTW:
					if (sizeof(pExtTag->sftw) == pTag->value_len)
					{
						std::memcpy( pExtTag->sftw, pTag->value, sizeof(pExtTag->sftw) );
					}
					break;
				case TAG_RCID:
					if (sizeof(pExtTag->rcid) == pTag->value_len)
					{
						std::memcpy( pExtTag->rcid, pTag->value, sizeof(pExtTag->rcid) );
					}
					break;
				case TAG_EDID:
					if (sizeof(pExtTag->edid) == pTag->value_len)
					{
						std::memcpy( pExtTag->edid, pTag->value, sizeof(pExtTag->edid) );
					}
					break;

				case TAG_PRMT:
					{
						char* pAddr = pTag->value;

						while (1)
						{
							pAddr = std::strtok(pAddr, ",");

							if (pAddr)
							{
								if (std::strcmp( pAddr, PRMT_UPLOAD ) == 0)
								{
									pExtTag->permitUpload = true;
								}
								else if (std::strcmp( pAddr, PRMT_EDIT_DISTRIBUTE ) == 0)
								{
									pExtTag->permitEditDistribute = true;
								}
								else if (std::strcmp( pAddr, PRMT_FIRST_DISTRIBUTE ) == 0)
								{
									pExtTag->permitFirstDistribute = true;
								}
								else if (std::strcmp( pAddr, PRMT_SECOND_DISTRIBUTE ) == 0)
								{
									pExtTag->permitSecondDistribute = true;
								}

								pAddr = NULL;
							}
							else
							{
								break;
							}
						}
					}
					break;
				}

				pTag = pTag->next;
			}

			// ^Cg擾
			{
				wchar_t* convHead = m_titleName;
				u32 len = tInfo.title_len;

				Convert2UTF16LE( (u8*)m_titleName, (const u8*)tInfo.title, &len );
				len = len >> 1;
				
				// BOM
				if (m_titleName[0] == 0xfeff)
				{
					convHead = &m_titleName[1];
					len--;
				}
				else
				{
					convHead = &m_titleName[0];
				}

				for (u32 i = 0; i < len; i++)
				{
					m_titleName[i] = *(convHead + i);
				}
				m_titleName[len] = L'\0';
			}
		}

		SafeFree( m_allocator, udta );
	}

	// trak
	readOffset = moovOffset;
	do
	{
		unsigned int nextBox = 0;
		unsigned char tag[8] = {0};

		result = m_stream->TrySeek( readOffset, nn::fs::POSITION_BASE_BEGIN );
		if (result.IsFailure())
		{
			return GetFsWholeError( result );
		}
		result = m_stream->TryRead( &readCount, tag, TMC_M4ADEMUX_BOX_TAG_SIZE );
		if (result.IsFailure())
		{
			return GetFsWholeError( result );
		}
		readOffset += TMC_M4ADEMUX_BOX_TAG_SIZE;

		int status = TMC_M4ADemux_SearchTrak(tag, &nextBox);

		if (TMC_M4ADEMUX_SEARCH_NEXT == status)
		{
			if (static_cast<int>(nextBox) < 0)
			{
				// ̒l͕s
				return ResultInvalidFile;
			}
			else
			{
				readOffset += nextBox;
			}
		}
		else if (status > 0)
		{
			m_trakSize = status;

			// ُȑ傫̃gbNTCY
			if (c_trakSize < m_trakSize)
			{
				return ResultInvalidFile;
			}

			m_trak = (u8*)m_allocator->Allocate( m_trakSize, 32 );
			if (m_trak == NULL)
			{
				return ResultOutOfResource;
			}

			result = m_stream->TryRead( &readCount, m_trak, m_trakSize );
			if (result.IsFailure())
			{
				SafeFree( m_allocator, m_trak );
				return GetFsWholeError( result );
			}

			status = TMC_M4ADemux_CheckMp4aTrak( m_trak, m_trakSize );

			if (status == TMC_M4ADEMUX_SUCCESS)
			{
				break;
			}
			else
			{
				SafeFree( m_allocator, m_trak );
				m_trakSize = 0;
			}
		}
		else
		{
			// trakŃG[܂
			return ResultInvalidFile;
		}
	} while (readOffset - moovOffset < moovSize);

	// t[f[^
	TMC_M4AContInfo cInfo;
	{
		int handleSize = TMC_M4ADemux_GetHandleSize();

		m_handle = m_allocator->Allocate( handleSize, 32 );

		if (m_handle == NULL)
		{
			return ResultOutOfResource;
		}

		int status = TMC_M4ADemux_Init(m_handle, handleSize, m_trak, m_trakSize);
		if (status != TMC_M4ADEMUX_SUCCESS)
		{
			SafeFree( m_allocator, m_handle );
			SafeFree( m_allocator, m_trak );
			return ResultInvalidDecoder;
		}

		status = TMC_M4ADemux_SetStartTime( m_handle, 0 );
		if (status != TMC_M4ADEMUX_SUCCESS)
		{
			SafeFree( m_allocator, m_handle );
			SafeFree( m_allocator, m_trak );
			return ResultInvalidDecoder;
		}

		unsigned int offset = 0, size = 0, dts = 0;
		status = TMC_M4ADemux_GetNextFrame(m_handle, &offset, &size, &dts, 0);
		if (status < TMC_M4ADEMUX_SUCCESS || AAC_REQUIRE_BUFFER_SIZE < size)
		{
			SafeFree( m_allocator, m_handle );
			SafeFree( m_allocator, m_trak );
			return ResultInvalidDecoder;
		}

		status = TMC_M4ADemux_GetInfo(m_handle, &cInfo);
		if (status != TMC_M4ADEMUX_SUCCESS)
		{
			SafeFree( m_allocator, m_handle );
			SafeFree( m_allocator, m_trak );
			return ResultInvalidDecoder;
		}

		// 鎞Ԃ𒴂Ă
		if (c_totalTimeMax <= cInfo.total_time)
		{
			SafeFree( m_allocator, m_handle );
			SafeFree( m_allocator, m_trak );
			return ResultInvalidFile;
		}

		// TvO[g
		m_samplingRate = GetSamplingRateIdx( cInfo.sampling_rate_idx );

		// `l
		m_channel = cInfo.ch;

		// Ŝ̎
		m_totalTime = s32(cInfo.total_time);
	}

	// AAC fR[_̏
	m_heap = (s32*)m_allocator->Allocate( TMC_AACDEC_HEAP_SIZE, 32 );

	m_param.samplingRate	= cInfo.sampling_rate_idx;
	m_param.crcFlag			= TMC_AACDEC_CRC_CALC_OFF;
	m_param.streamForm		= TMC_AACDEC_FORMAT_RAW;
	m_param.pHeap			= m_heap;

	m_aacdec = TMC_AACDec_Initialize(&m_param);
	m_buffer = reinterpret_cast<s16*>(
		m_allocator->Allocate( AAC_REQUIRE_BUFFER_SIZE, 32 )
	);

	if (m_buffer == NULL)
	{
		SafeFree( m_allocator, m_handle );
		SafeFree( m_allocator, m_trak );
		return ResultOutOfResource;
	}

	return ResultSuccess;
}


/**-----------------------------------------------------------------------------

	@brief		j܂
	@par		[]
				fXgN^̒ŌĂ΂܂<br>
				IɌĂł肠܂

----------------------------------------------------------------------------- */
void DecoderAACArm::Finalize()
{
	SafeFree( m_allocator, m_handle );
	SafeFree( m_allocator, m_trak );

	TMC_AACDec_Finalize( m_aacdec );

	SafeFree( m_allocator, m_heap );
	SafeFree( m_allocator, m_buffer );

	m_allocator = NULL;
	m_samplingRate = 0;
	m_channel = 0;
	m_totalTime = 0;
	m_seekTime = 0;
}


/**-----------------------------------------------------------------------------

	@brief		Pt[̃Tv擾܂B

----------------------------------------------------------------------------- */
u32 DecoderAACArm::GetFrameSamples() const
{
	return AAC_FRAME_SAMPLES;
}


/**-----------------------------------------------------------------------------

	@brief		Pt[̃obt@TCY(Byte)擾܂

----------------------------------------------------------------------------- */
u32 DecoderAACArm::GetFrameBufferSize() const
{
	return AAC_FRAME_SAMPLES * sizeof(s16);
}


/**-----------------------------------------------------------------------------

	@brief		TvO[g擾܂B

----------------------------------------------------------------------------- */
s32 DecoderAACArm::GetSamplingRate() const
{
	return m_samplingRate;
}


/**-----------------------------------------------------------------------------

	@brief		`l擾܂B

----------------------------------------------------------------------------- */
s32 DecoderAACArm::GetChannelCount() const
{
	return m_channel;
}


/**-----------------------------------------------------------------------------

	@brief		ĐԂ擾܂(msecP)

----------------------------------------------------------------------------- */
s32 DecoderAACArm::GetTotalTime() const
{
	return m_totalTime;
}


/**-----------------------------------------------------------------------------

	@brief		V[NԂ擾܂(msecP)

----------------------------------------------------------------------------- */
s32 DecoderAACArm::GetSeekTime() const
{
	return m_seekTime;
}


/**-----------------------------------------------------------------------------

	@brief		V[N\

----------------------------------------------------------------------------- */
bool DecoderAACArm::CanSeek() const
{
	return true;
}


/**-----------------------------------------------------------------------------

	@brief		tĐw

----------------------------------------------------------------------------- */
void DecoderAACArm::SetReverse( bool )
{
}


/**-----------------------------------------------------------------------------

	@brief		(msecP)ŃV[Nʒuݒ肵܂

----------------------------------------------------------------------------- */
s32 DecoderAACArm::SetSeekTime( s32 seekTime )
{
	s32 prevSeek = m_seekTime;

	int result = TMC_M4ADemux_SetStartTime( m_handle, seekTime );

	if (result == TMC_M4ADEMUX_SUCCESS)
	{
		m_seekTime = seekTime;
	}
	else
	{
		// s猳ɖ߂
		TMC_M4ADemux_SetStartTime( m_handle, prevSeek );

		m_seekTime = prevSeek;
	}

	// fR[_[Cuď
	TMC_AACDec_Finalize( m_aacdec );
	m_aacdec = TMC_AACDec_Initialize(&m_param);

	return m_seekTime;
}


/**-----------------------------------------------------------------------------

	@brief		TEh̔g`f[^(PCM16)擾܂B

----------------------------------------------------------------------------- */
Result DecoderAACArm::TryRead( s32* readCount, s16* pBufL, s16* pBufR, s32 samples )
{
	nn::Result result;
	s32 readBytes = 0;
	s32 current = 0;

	if (readCount == NULL)
	{
		return ResultInvalidParameter;
	}

	if (0 < (samples % AAC_FRAME_SAMPLES) || samples < AAC_FRAME_SAMPLES)
	{
		*readCount = current;
		return ResultInvalidParameter;
	}

	for (s32 i = 0; i <= samples - AAC_FRAME_SAMPLES; i+= AAC_FRAME_SAMPLES)
	{
		unsigned int offset = 0, size = 0, dts = 0;
		int status = TMC_M4ADemux_GetNextFrame(m_handle, &offset, &size, &dts, 0);

		if (status < 0 || AAC_REQUIRE_BUFFER_SIZE < size)
		{
			// ̃t[͓ǂݎȂ
			m_seekTime = m_totalTime;
			*readCount = current;
			return ResultSuccess;
		}

		if (status == TMC_M4ADEMUX_ENDSTREAM)
		{
			m_seekTime = dts;

			// ŏIt[擾(f[^Ȃ)
		}

		// V[NԂXVēǂݔ΂
		if (pBufL == NULL && pBufR == NULL)
		{
			m_seekTime = dts;
			*readCount = 0;
			return ResultSuccess;
		}

		// ǂݍ݈ʒu܂ŃV[N
		result = m_stream->TrySeek( offset, nn::fs::POSITION_BASE_BEGIN );
		if (result.IsFailure())
		{
			return GetFsWholeError( result );
		}

		std::memset( m_buffer, 0x00, AAC_REQUIRE_BUFFER_SIZE );

		result = m_stream->TryRead( &readBytes, m_buffer, size );
		if (result.IsFailure())
		{
			return GetFsWholeError( result );
		}

		if (0 < size && 0 < readBytes)
		{
			// V[NʒuXV
			m_seekTime = dts;

			// AACfR[h
			TMC_AACDecInfo info;
			TMC_AACDec_Decode( m_aacdec, (u8*)m_buffer, size, pBufL, pBufR, &info );

			// ʂ̊mF
			if (info.errorCode != 0)
			{
				*readCount = current;
				return ResultInvalidDecoder;
			}
		}
		else
		{
			*readCount = current;
			return ResultSuccess;
		}

		if (pBufL)
		{
			pBufL += AAC_FRAME_SAMPLES;
		}

		if (pBufR)
		{
			pBufR += AAC_FRAME_SAMPLES;
		}

		current += AAC_FRAME_SAMPLES;
	}

	*readCount = current;
	return ResultSuccess;
}


/**-----------------------------------------------------------------------------

	@brief		^Cg擾

----------------------------------------------------------------------------- */
const wchar_t* DecoderAACArm::GetTitleName() const
{
	return m_titleName;
}


}}
