/**-----------------------------------------------------------------------------

	@file		sndio_MP3Seek.cpp
	@author		tetsu
	@par		[]
				MP3̃V[NsNX(J)

				Copyright (c)2010 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#include <nn/math/math_Utility.h>
#include "sndio_Decoder.h"
#include "sndio_MP3Seek.h"


////////////////////////////////////////////////////////////////////////////////
// [J`
////////////////////////////////////////////////////////////////////////////////

#if defined(APP_SNOTE) && defined(NN_BUILD_DEBUG)
#define _LOG( ... )		NN_LOG( __VA_ARGS__ )
#else
#define _LOG( ... )
#endif

namespace
{
	//! XING
	const u32 c_headerXING = VOICE_MAKE_SIGWORD('X', 'i', 'n', 'g');

	//! VBRI
	const u32 c_headerVBRI = VOICE_MAKE_SIGWORD('V', 'B', 'R', 'I');

	static u32 SwapEndian( u32 val )
	{
		const u32 MASK = 0xFF00FF00;
		val = ( (val & MASK) >> 8 ) | ( (val << 8) & MASK );
		return (val >> 16) | (val << 16);
	}
	
	static u16 SwapEndian( u16 val )
	{
		return (u16)( (val >> 8) | (val << 8) );
	}
}


////////////////////////////////////////////////////////////////////////////////
// MP3Seek`
////////////////////////////////////////////////////////////////////////////////

namespace sndio { namespace detail {


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				RXgN^

----------------------------------------------------------------------------- */
MP3Seek::MP3Seek() :
m_existXing( false ),
m_existVBRI( false ),
m_samplingRate( 0 ),
m_totalTime( 0 ),
m_headerTag( 0 ),
m_headerFlags( 0 ),
m_numberOfFrames( 0 ),
m_totalStreamSize( 0 ),
m_tocSize( 0 ),
m_tocScale( 0 ),
m_vbrQuality( 0 )
{
	std::memset( m_toc, 0x00, sizeof(m_toc) );
}


/**-----------------------------------------------------------------------------

	@brief		
	@return		Ȃ
	@par		[]
				fXgN^

----------------------------------------------------------------------------- */
MP3Seek::~MP3Seek()
{
}


/**-----------------------------------------------------------------------------

	@brief		ĐԎ擾
	@return		Đ
	@par		[]
				ĐԂ擾܂(msecP)

----------------------------------------------------------------------------- */
s32 MP3Seek::GetTotalTime() const
{
	return m_totalTime;
}


/**-----------------------------------------------------------------------------

	@brief		t@CV[Nʒu擾
	@param[in]	seekTime		V[Nʒu
	@return		t@CV[Nʒu
	@par		[]
				(msecP)ŃV[Nʒuݒ肵At@CʒuԂ܂<br>
				t@Cʒu͓t[ł͂Ȃ̂̈ʒuԂ̂
				擾 calc_mp3_findsync œt[̈ʒu܂ňړ
				

----------------------------------------------------------------------------- */
s32 MP3Seek::SeekToTime( s32 seekTime )
{
	// V[Nʒů 0.0`1.0֕ϊ
	f32 per = f32(seekTime) / f32(GetTotalTime());
	per = nn::math::Clamp( per, 0.0f, 1.0f );

	s32 seekPoint = 0;

	if (m_existXing)
	{
		// Xing ʒu
		f32 fa, fb, fx;
		u32 pc;

		pc = static_cast<u32>(per * 100.0f);
		fa = static_cast<f32>(m_toc[pc]);

		if (pc > 99)
		{
			pc = 99;
		}

		if (pc < 99)
		{
			fb = (f32)m_toc[pc+1];
		}
		else
		{
			fb = 256.00f;
		}

		fx = fa + (fb - fa) * ((100.0f * per) - (f32)pc);

		seekPoint = (s32)((fx / 256.0f) * (f32)m_totalStreamSize);
	}
	else if (m_existVBRI)
	{
		// VBRI
		u32 i = 0, fraction = 0, SamplesPerFrame;
		f32 fLengthMS;
		f32 fLengthMSPerTOCEntry;
		f32 fAccumulatedTimeMS = 0.0f;

		if (m_samplingRate >= 32000)
		{
			SamplesPerFrame = 1152;
		}
		else
		{
			SamplesPerFrame = 576;
		}

		fLengthMS = ((f32)m_numberOfFrames * (f32)SamplesPerFrame)
			/ (f32)m_samplingRate * 1000.0f;

		fLengthMSPerTOCEntry = (f32)fLengthMS / (f32)(m_tocSize + 1);

		if (seekTime > fLengthMS)
			seekTime = (u32)fLengthMS;

		while (fAccumulatedTimeMS <= f32(seekTime))
		{
			seekPoint += m_toc[i];
			fAccumulatedTimeMS += fLengthMSPerTOCEntry;
			i++;
		}

		fraction = ((int)((((fAccumulatedTimeMS - seekTime) / fLengthMSPerTOCEntry)
			+ (1.0f / (2.0f * (f32)m_tocSize))) * (f32)m_tocSize));

		seekPoint -= (int)((f32)m_toc[i-1] * (f32)(fraction)
			/ (f32)m_tocSize);
	}
	else
	{
		seekPoint = s32( f32(m_totalStreamSize) * per );
	}

	return seekPoint;
}


/**-----------------------------------------------------------------------------

	@brief		V[N񏉊
	@param[in]	stream			f[^Xg[
	@param[in]	streamSize		Xg[TCY(^O𔲂mȂ)
	@retval		true 
	@retval		false s(vIȃG[)
	@par		[]
				(msecP)ŃV[Nʒuݒ肵܂

----------------------------------------------------------------------------- */
bool MP3Seek::Initialize( nn::fs::IInputStream* stream, u32 streamSize )
{
	NN_NULL_TASSERT_( stream );

	nn::Result result;
	s32 readCount = 0;

	if (stream == NULL)
	{
		return false;
	}

	// MP3wb_ǂݏo
	u8 mp3header[4] = {0};
	if (stream->Read( mp3header, 4 ) != 4)
	{
		return false;
	}

	m_bitrate = calc_mp3_bitrate( mp3header );
	u32 frameSize = calc_mp3_size( mp3header );
	m_samplingRate = GetSamplingRateIdx( calc_mp3_sampleidx( mp3header ) );

	// XING/VBRIwb_邩Ȃʒu܂Ői߂
	result = stream->TrySeek(
		stream->GetPosition() + calc_mp3_side( mp3header ),
		nn::fs::POSITION_BASE_BEGIN
	);
	if (result.IsFailure())
	{
		return false;
	}

	// wb_VOl`o
	result = stream->TryRead( &readCount, &m_headerTag, sizeof(m_headerTag) );
	if (result.IsFailure())
	{
		return false;
	}

	if (m_headerTag == c_headerXING)
	{
		// XINGwb_
		_LOG( "[MP3] Find Xing Header\n" );
		if (ReadXING( stream ))
		{
			m_existXing = true;
		}

	}
	else if (m_headerTag == c_headerVBRI)
	{
		// VBRIwb_
		_LOG( "[MP3] Find VBRI Header\n" );
		if (ReadVBRI( stream ))
		{
			m_existVBRI = true;
		}
	}
	else
	{
		// "INFO" LAME-CBRƂĔ肷̂ŉȂ
	}

	if (!m_existXing && !m_existVBRI)
	{
		_LOG( "[MP3] CBR-MODE\n" );

		// CBR
		m_totalStreamSize = streamSize;
		m_numberOfFrames = m_totalStreamSize / frameSize;
		m_totalTime = m_totalStreamSize / m_bitrate * 8;
	}

	_LOG( "[MP3] bitrate (%d)\n", m_bitrate );
	_LOG( "[MP3] samplerate (%d)\n", m_samplingRate );
	_LOG( "[MP3] numberOfFrames (%d)\n", m_numberOfFrames );
	_LOG( "[MP3] totalStreamSize (%d)\n", m_totalStreamSize );
	_LOG( "[MP3] TOC Size (%d) TOC[%d][%d][%d] ` [%d]\n",
		m_tocSize, m_toc[0], m_toc[1], m_toc[2], m_toc[m_tocSize - 1] );

	return true;
}


/**-----------------------------------------------------------------------------

	@brief		XING
	@param[in]	stream			f[^Xg[
	@retval		true (TOCgpăV[Nł)
	@par		[]
				XINGœǂݍ݉͂s܂

----------------------------------------------------------------------------- */
bool MP3Seek::ReadXING( nn::fs::IInputStream* stream )
{
	NN_NULL_TASSERT_( stream );
	nn::Result result;
	s32 readCount = 0;

	// tO
	result = stream->TryRead( &readCount, &m_headerFlags, sizeof(m_headerFlags) );
	if (result.IsFailure())
	{
		return false;
	}
	m_headerFlags = SwapEndian( m_headerFlags );

	// t[
	if (m_headerFlags & FLAG_IS_FRAMES)
	{
		result = stream->TryRead( &readCount, &m_numberOfFrames, sizeof(m_numberOfFrames) );
		if (result.IsFailure())
		{
			return false;
		}
		m_numberOfFrames = SwapEndian( m_numberOfFrames );

		// g[^^C̃`FbN
		m_totalTime = s32(
			f32(m_numberOfFrames * GetSamplePerFrame()) / f32(m_samplingRate) * 1000.0f
		);
		if (360000000 <= m_totalTime)
		{
			// Ԃs
			return false;
		}
	}

	// Xg[TCY
	if (m_headerFlags & FLAG_IS_BYTES)
	{
		result = stream->TryRead( &readCount, &m_totalStreamSize, sizeof(m_totalStreamSize) );
		if (result.IsFailure())
		{
			return false;
		}
		m_totalStreamSize = SwapEndian( m_totalStreamSize );

		if (m_totalStreamSize <= 0 || 2147483648ul <= m_totalStreamSize)
		{
			// Xg[TCYs
			return false;
		}
	}

	// TOC
	if (m_headerFlags & FLAG_IS_TOC)
	{
		m_tocSize = 100;
		result = stream->TryRead( &readCount, m_toc, m_tocSize );
		if (result.IsFailure())
		{
			return false;
		}
	}

	// i
	if (m_headerFlags & FLAG_IS_VBR_SCALE)
	{
		u32 quality = 0;
		result = stream->TryRead( &readCount, &quality, sizeof(quality) );
		if (result.IsFailure())
		{
			return false;
		}

		quality = SwapEndian( quality );
		m_vbrQuality = static_cast<u32>( quality );
	}

	return true;
}


/**-----------------------------------------------------------------------------

	@brief		VBRI
	@param[in]	stream			f[^Xg[
	@retval		true (TOCgpăV[Nł)
	@par		[]
				VBRIœǂݍ݉͂s܂

----------------------------------------------------------------------------- */
bool MP3Seek::ReadVBRI( nn::fs::IInputStream* stream )
{
	NN_NULL_TASSERT_( stream );

	// o[W
	u16 version;
	stream->Read( &version, sizeof(version) );
	version = SwapEndian( version );

	// fBC
	u16 delay;
	stream->Read( &delay, sizeof(delay) );
	delay = SwapEndian( delay );

	// i
	stream->Read( &m_vbrQuality, sizeof(m_vbrQuality) );
	m_vbrQuality = SwapEndian( m_vbrQuality );

	// Xg[TCY
	stream->Read( &m_totalStreamSize, sizeof(m_totalStreamSize) );
	m_totalStreamSize = SwapEndian( m_totalStreamSize );
	
	// Xg[TCY`FbN
	if (m_totalStreamSize <= 0 || 2147483648ul <= m_totalStreamSize)
	{
		// Xg[TCYs
		return false;
	}

	// t[
	stream->Read( &m_numberOfFrames, sizeof(m_numberOfFrames) );
	m_numberOfFrames = SwapEndian( m_numberOfFrames );

	// g[^^C̃`FbN
	m_totalTime = s32(
		f32(m_numberOfFrames * GetSamplePerFrame()) / f32(m_samplingRate) * 1000.0f
	);
	if (360000000 <= m_totalTime)
	{
		// Ԃs
		return false;
	}

	// TOC
	stream->Read( &m_tocSize, sizeof(m_tocSize) );
	m_tocSize = SwapEndian( m_tocSize ) + 1;
	if (m_tocSize <= 0 || 65536 <= m_tocSize)
	{
		return false;
	}

	// TOCXP[
	stream->Read( &m_tocScale, sizeof(m_tocScale) );
	m_tocScale = SwapEndian( m_tocScale ) + 1;

	// PGg̃oCg
	u16 bytesPerEntry;
	stream->Read( &bytesPerEntry, sizeof(bytesPerEntry) );
	bytesPerEntry = SwapEndian( bytesPerEntry );

	// PGg̃t[
	u16 framesPerEntry;
	stream->Read( &framesPerEntry, sizeof(framesPerEntry) );
	framesPerEntry = SwapEndian( framesPerEntry );

	// TOCǂݍ
	stream->Read( m_toc, m_tocSize );
	return true;
}


}}
