/**-----------------------------------------------------------------------------

	@file		sndio_Mount.cpp
	@author		tetsu
	@par		[]
				}Eg`

				Copyright (c)2010 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#include <nn/fs.h>
#include <nn/fs/CTR/fs_ArchiveTypesForSystem.h>
#include <nn/fs/fs_ApiSharedExtSaveData.h>
#include <nn/fs/CTR/MPCore/fs_FileSystemBasePrivate.h>
#include "sndio_Mount.h"


namespace sndio { namespace detail {


static bool _sd_mount_flag = false;
static bool _ctr_nand_mount_flag = false;
static bool _twl_nand_mount_flag = false;


/*!
	@brief		SDJ[hւ̃}Egs܂B

				sndionCuSDJ[hꍇɕKĂяoĂB<br/>

	@return		nn::ResultŌʂԂ܂(W[nn::fs)
*/
nn::Result MountSD()
{
	if (nn::fs::IsSdmcInserted() && !_sd_mount_flag)
	{
		_sd_mount_flag = true;
		return nn::fs::MountSdmc( GetArchiveNameA( ARCHIVE_TYPE_SD ) );
	}

	// SDJ[hĂȂG[Ԃ
	return nn::fs::ResultMediaNotFound();
}


/*!
	@brief		SDJ[h̃A}Egs܂B

	@return		nn::ResultŌʂԂ܂(W[nn::fs)
*/
nn::Result UnmountSD()
{
	if (_sd_mount_flag)
	{
		_sd_mount_flag = false;
		return nn::fs::Unmount( GetArchiveNameA( ARCHIVE_TYPE_SD ) );
	}

	return nn::fs::ResultNotInitialized();
}


/*!
	@brief		CTR NAND̈̃}Egs܂B

				sndionCuCTR NAND̈ꍇɕKĂяoĂB<br/>

	@return		nn::ResultŌʂԂ܂(W[nn::fs)
*/
nn::Result MountCtrNand()
{
	if (!_ctr_nand_mount_flag)
	{
		_ctr_nand_mount_flag = true;
#if defined(SND_IO_EMULATE_CTR_NAND_ON_SD)
		return nn::fs::MountSdmc( GetArchiveNameA( ARCHIVE_TYPE_CTR_NAND ) );
#else
		return nn::fs::MountSharedExtSaveData( GetArchiveNameA( ARCHIVE_TYPE_CTR_NAND ), 0xf0000002 );
#endif
	}

	return nn::ResultSuccess();
}


/*!
	@brief		CTR NAND̈A}Eg܂B

	@return		nn::ResultŌʂԂ܂(W[nn::fs)
*/
nn::Result UnmountCtrNand()
{
	if (_ctr_nand_mount_flag)
	{
		_ctr_nand_mount_flag = false;
		return nn::fs::Unmount( GetArchiveNameA( ARCHIVE_TYPE_CTR_NAND ) );
	}

	return nn::fs::ResultNotInitialized();
}


/*!
	@brief		TWL NAND̈}Egs܂B

				sndionCuTWL NAND̈ꍇɕKĂяoĂB<br/>

	@return		nn::ResultŌʂԂ܂(W[nn::fs)
*/
nn::Result MountTwlNand()
{
	if (!_twl_nand_mount_flag)
	{
		_twl_nand_mount_flag = true;

		// TWLTEḧ̃}Eg
		return nn::fs::MountSpecialArchive(
			GetArchiveNameA( ARCHIVE_TYPE_TWL_NAND ),
			nn::fs::CTR::ARCHIVE_TYPE_TWL_SOUND );
	}

	return nn::ResultSuccess();
}


/*!
	@brief		TWL NAND̈A}Eg܂B

	@return		nn::ResultŌʂԂ܂(W[nn::fs)
*/
nn::Result UnmountTwlNand()
{
	if (_twl_nand_mount_flag)
	{
		_twl_nand_mount_flag = false;
		return nn::fs::Unmount( GetArchiveNameA( ARCHIVE_TYPE_TWL_NAND ) );
	}

	return nn::fs::ResultNotInitialized();
}


/*!
	@brief		ۂɎgpXϊ܂

	@return		ϊꂽpXւ̃|C^
*/
const wchar_t* ToReal( wchar_t* path )
{
	if (IsVirtual(path))
	{
		std::memcpy( path, SNDIO_SD_ARCHIVE_NAMEW, detail::SD_ARCHIVE_PREFIX_SIZE );
	}

	return path;
}


/*!
	@brief		zIɎgpXϊ܂

	@return		ϊꂽpXւ̃|C^
*/
const wchar_t* ToVirtual( wchar_t* path )
{
	std::memcpy( path, SNDIO_SD_VIRTUAL_ARCHIVE_NAMEW, detail::SD_ARCHIVE_PREFIX_SIZE );
	return path;
}


/*!
	@brief		zIȃpXǂԂ܂

	@return		true zpX
*/
bool IsVirtual( const wchar_t* path )
{
	NN_COMPILER_ASSERT( detail::SD_ARCHIVE_PREFIX_SIZE == 14 );
	return !!(std::memcmp( path, SNDIO_SD_VIRTUAL_ARCHIVE_NAMEW, detail::SD_ARCHIVE_PREFIX_SIZE ) == 0);
}


}}
