/**-----------------------------------------------------------------------------

	@file		sndio_UftConv.h
	@author		tetsu
	@par		[]
				ϊ֘A

				Copyright (c)2011 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#include <wchar.h>
#include <nn/types.h>
#include <nn/assert.h>
#include "sndio_UtfConv.h"


namespace sndio { namespace detail { namespace utf {
#define ENC_ASSERT NN_ASSERT

// TQ[gyAɑΉꍇɂTRUEɐݒ肷B
#define ENC_SURROGATE_ENABLE    0

static const u16 BOM = 0xFEFF; // Byte Order Mark
#define BOM_REVERSE 0xFFFE


typedef s32 (*ENCReadCharFunc)( const u8 *src, u32 src_len, u32 *utf32_code );
typedef s32 (*ENCWriteCharFunc)( u8* dst, u32 dst_len, u32 code );

static s32 ReadUTF8   ( const u8 *src, u32 src_len, u32 *utf32_code );
static s32 ReadUTF16BE( const u8 *src, u32 src_len, u32 *utf32_code );
static s32 ReadUTF16LE( const u8 *src, u32 src_len, u32 *utf32_code );
static s32 WriteUTF8   ( u8* dst, u32 dst_len, u32 code );
static s32 WriteUTF16BE( u8* dst, u32 dst_len, u32 code );
static s32 WriteUTF16LE( u8* dst, u32 dst_len, u32 code );

static struct
{
    ENCReadCharFunc     readFunc;
    ENCWriteCharFunc    writeFunc;
} sConvFunc[] = 
{
    { ReadUTF8,    WriteUTF8    },
    { ReadUTF16BE, WriteUTF16BE },
    { ReadUTF16LE, WriteUTF16LE },
};

// #define STRLEN_4BYTE_SKIP

int GetStringLength(const char *str)
{
#if defined(STRLEN_4BYTE_SKIP)
	int     n;
	unsigned long  *p32, y;

	for( n=0; ((unsigned long)(str + n) & 3) != 0; n++ )
	{
		if(!str[n])
			return n;
	}

	p32 = (unsigned long*)(str + n);
	for( ; ; n+=4 )
	{
		y = (*p32 & 0x7F7F7F7F) + 0x7F7F7F7F;
		y = ~(y | *p32 | 0x7F7F7F7F);

		if( y != 0 )
			break;

		p32++;
	}

#if 1
	while (str[n])
	{
		n++;
	}
	return n;
#else
	if( y & 0x000000FF )
		return n;
	else if( y & 0x0000FF00 )
		return n + 1;
	else if( y & 0x00FF0000 )
		return n + 2;
	else
		return n + 3;
	// ɓB邱Ƃ͂肦Ȃ
	return -1;
#endif

#else    // original
	int     n = 0;
	while (str[n])
	{
		n++;
	}
	return n;
#endif
}


int CopyLStringZeroFill(char *destp, const char *srcp, int n)
{
	int     i;
	const char *s = srcp;

	NN_TASSERT_(destp && srcp && n >= 0);

	for (i = 0; i < n - 1; i++)
	{
		destp[i] = (char)*s;
		if (*s)
		{
			s++;
		}
	}
	if( i < n )
	{
		destp[i] = '\0';
	}

	return GetStringLength(srcp);
}

/*---------------------------------------------------------------------------*
  Name:         ContLeadingOne

  Description:  oCgf[^̍ŏʃrbgł1̐𐔂܂B
                CLZ߂̃AZuōłꍇɂ́Å֐
                CLZ (~value)ŒuƂ悢ł傤B

  Arguments:    value       JEgΏۂ̃oCgf[^

  Returns:      ʉrbg1Ă邩߂܂B
 *---------------------------------------------------------------------------*/
static inline u8
CountLeadingOne( u8 value )
{
    u8 cnt = 0;
    
    // 8bitȂ̂ŁA2T͂Pȃ[vJEg
    while ( value & 0x80 )
    {
        ++cnt;
        value <<= 1;
    }
    return cnt;
}

/*---------------------------------------------------------------------------*
  Name:         ReadUTF8

  Description:  obt@UTF8̕ǂݍ݂܂B

  Arguments:    src        \[Xf[^
                src_len    \[X̎c̒
                utf32_code UTF32̕R[h

  Returns:      ǂݍ񂾃oCgԂ܂B
                zO̕R[h̏ꍇɂ̓G[Ƃĕ̃G[R[hԂ܂B
 *---------------------------------------------------------------------------*/
static s32
ReadUTF8(const u8 *src, u32 src_len, u32 *utf32_code)
{
    u32   code = 0;
    u8    leading_bits;
    u8    len;
    
    ENC_ASSERT( src != NULL );
    ENC_ASSERT( utf32_code != NULL  );
    if ( src_len == 0 ) { return ENC_ERR_INVALID_FORMAT; }

    leading_bits = CountLeadingOne( *src );
    
	if (leading_bits == 1) { return ENC_ERR_INVALID_FORMAT; }
    if (leading_bits >= 5) { return ENC_ERR_INVALID_FORMAT; } // 5`6Byte̕\͎󂯕tȂ
    
    len = leading_bits;
    if ( len == 0 ) { len = 1; }
    
    if ( len > src_len ) { return ENC_ERR_NOT_LOADED; }
    
    code = *src & ((1 << (8 - leading_bits)) - 1);
    
    while ( leading_bits > 1 )
    {
        ++src;
        --leading_bits;
        if ( (*src & 0xC0) != 0x80 ) { return ENC_ERR_INVALID_FORMAT; }
        
        code <<= 6;
        code += (*src & 0x3F);
    }
    *utf32_code = code;
    return len;
}


/*---------------------------------------------------------------------------*
  Name:         WriteUTF8

  Description:  UTF8̕obt@ɏo܂B
                dstNULLw肵ꍇɂ́ATCŶݕԂ܂B

  Arguments:    dst     o͐̃obt@
                dst_len o͐̃obt@TCY
                code    UTF32̕R[h

  Returns:      񂾃oCgԂ܂B
                ݐ̃obt@TCYȂꍇɂ́A
                ̃G[R[hԂ܂B
 *---------------------------------------------------------------------------*/
static s32
WriteUTF8(u8* dst, u32 dst_len, u32 code)
{
    s32 len, i;
    
    if ( code < 0x00000080 ) { len = 1; }
    else if ( code < 0x00000800 ) { len = 2; }
    else if ( code < 0x00010000 ) { len = 3; }
    else if ( code < 0x02000000 ) { len = 4; }
    else { return ENC_ERR_INVALID_FORMAT; }
    
    if ( ! dst ) { return len; }
    if ( dst_len == 0 ) { return ENC_ERR_NO_BUF_LEFT; }
    
    switch ( len )
    {
    case 1: { dst[ 0 ] = (u8)code;                  } break;
    case 2: { dst[ 0 ] = (u8)((code >> 6) + 0xC0);  } break;
    case 3: { dst[ 0 ] = (u8)((code >> 12) + 0xE0); } break;
    case 4: { dst[ 0 ] = (u8)((code >> 18) + 0xF0); } break;
    }
    
    if ( len > s32(dst_len) ) { return ENC_ERR_NO_BUF_LEFT; }
    
    for ( i = 1; i < len; ++i )
    {
        dst[ i ] = (u8)((code >> (6 * (len - i - 1)) & 0x3F) + 0x80);
    }
    return len;
}


/*---------------------------------------------------------------------------*
  Name:         ReadUTF16LE

  Description:  obt@UTF16LE̕ǂݍ݂܂B

  Arguments:    src        \[Xf[^
                src_len    \[X̎c̒
                utf32_code UTF32̕R[h

  Returns:      ǂݍ񂾃oCgԂ܂B
                zO̕R[h̏ꍇɂ̓G[Ƃĕ̃G[R[hԂ܂B
 *---------------------------------------------------------------------------*/
static s32
ReadUTF16LE(const u8 *src, u32 src_len, u32 *utf32_code)
{
    u32 code;
    ENC_ASSERT( src != NULL );
    ENC_ASSERT( utf32_code != NULL  );
    if ( src_len < 2 ) { return ENC_ERR_NOT_LOADED; }
    
    code = *src + (*(src + 1) << 8);
    
#if ENC_SURROGATE_ENABLE
    // TQ[gyAւ̑Ή
    if ( 0xD800 <= code && code < 0xDC00 )
    {
        u32 code_lo;
        if ( src_len < 4 ) { return ENC_ERR_NOT_LOADED; }
        code_lo = *(src + 2) + (*(src + 3) << 8);
        
        if ( code_lo < 0xDC00 || 0xE000 <= code_lo ) { return ENC_ERR_INVALID_FORMAT; }
        
        code = 0x10000 + ((code - 0xD800) << 10) + (code_lo - 0xDC00);
        *utf32_code = code;
        return 4;
    }
#endif
    
    *utf32_code = code;
    return 2;
}

/*---------------------------------------------------------------------------*
  Name:         WriteUTF16LE

  Description:  UTF16LE̕obt@ɏo܂B
                dstNULLw肵ꍇɂ́ATCŶݕԂ܂B

  Arguments:    dst     o͐̃obt@
                dst_len o͐̃obt@TCY
                code    UTF32̕R[h

  Returns:      񂾃oCgԂ܂B
                ݐ̃obt@TCYȂꍇɂ́A
                ̃G[R[hԂ܂B
 *---------------------------------------------------------------------------*/
static s32
WriteUTF16LE(u8* dst, u32 dst_len, u32 code)
{
#if ENC_SURROGATE_ENABLE
    // TQ[gyAւ̑Ή
    if ( code >= 0x10000 )
    {
        u32 code_lo, code_hi;
        if ( ! dst ) { return 4; }
        if ( dst_len < 4 ) { return ENC_ERR_NO_BUF_LEFT; }
        code_hi = ( ((code - 0x10000) >> 10) & 0x3FF ) + 0xD800;
        code_lo = ( (code - 0x10000) & 0x3FF ) + 0xDC00;
        
        dst[ 0 ] = code_hi & 0xFF;
        dst[ 1 ] = (code_hi >> 8) & 0xFF;
        dst[ 2 ] = code_lo & 0xFF;
        dst[ 3 ] = (code_lo >> 8) & 0xFF;
        return 4;
    }
#endif
    
    if ( ! dst ) { return 2; }
    if ( dst_len < 2 ) { return ENC_ERR_NO_BUF_LEFT; }
    
    dst[ 0 ] = code & 0xFF;
    dst[ 1 ] = (code >> 8) & 0xFF;
    
    return 2;
}

/*---------------------------------------------------------------------------*
  Name:         ReadUTF16BE

  Description:  obt@UTF16BE̕ǂݍ݂܂B

  Arguments:    src        \[Xf[^
                src_len    \[X̎c̒
                utf32_code UTF32̕R[h

  Returns:      ǂݍ񂾃oCgԂ܂B
                zO̕R[h̏ꍇɂ̓G[Ƃĕ̃G[R[hԂ܂B
 *---------------------------------------------------------------------------*/
static s32
ReadUTF16BE(const u8 *src, u32 src_len, u32 *utf32_code)
{
    u32 code;
    ENC_ASSERT( src != NULL );
    ENC_ASSERT( utf32_code != NULL  );
    if ( src_len < 2 ) { return ENC_ERR_NOT_LOADED; }
    
    code = (*src << 8) + *(src + 1);
    
#if ENC_SURROGATE_ENABLE
    // TQ[gyAւ̑Ή
    if ( 0xD800 <= code && code < 0xDC00 )
    {
        u32 code_lo;
        if ( src_len < 4 ) { return ENC_ERR_NOT_LOADED; }
        code_lo = (*(src + 2) << 8) + *(src + 3);
        
        if ( code_lo < 0xDC00 || 0xE000 <= code_lo ) { return ENC_ERR_INVALID_FORMAT; }
        
        code = 0x10000 + ((code - 0xD800) << 10) + (code_lo - 0xDC00);
        *utf32_code = code;
        return 4;
    }
#endif
    
    *utf32_code = code;
    // Ƃ肠TQ[gyAɂ͔Ή
    return 2;
}

/*---------------------------------------------------------------------------*
  Name:         WriteUTF16BE

  Description:  UTF16BE̕obt@ɏo܂B
                dstNULLw肵ꍇɂ́ATCŶݕԂ܂B

  Arguments:    dst     o͐̃obt@
                dst_len o͐̃obt@TCY
                code    UTF32̕R[h

  Returns:      񂾃oCgԂ܂B
                ݐ̃obt@TCYȂꍇɂ́A
                ̃G[R[hԂ܂B
 *---------------------------------------------------------------------------*/
static s32
WriteUTF16BE(u8* dst, u32 dst_len, u32 code)
{
#if ENC_SURROGATE_ENABLE
    // TQ[gyAւ̑Ή
    if ( code >= 0x10000 )
    {
        u32 code_lo, code_hi;
        if ( ! dst ) { return 4; }
        if ( dst_len < 4 ) { return ENC_ERR_NO_BUF_LEFT; }
        code_hi = ( ((code - 0x10000) >> 10) & 0x3FF ) + 0xD800;
        code_lo = ( (code - 0x10000) & 0x3FF ) + 0xDC00;
        
        dst[ 0 ] = (code_hi >> 8)& 0xFF;
        dst[ 1 ] = code_hi & 0xFF;
        dst[ 2 ] = (code_lo >> 8) & 0xFF;
        dst[ 3 ] = code_lo & 0xFF;
        return 4;
    }
#endif
    
    if ( ! dst ) { return 2; }
    if ( dst_len < 2 ) { return ENC_ERR_NO_BUF_LEFT; }
    
    dst[ 0 ] = (code >> 8) & 0xFF;
    dst[ 1 ] = code & 0xFF;
    
    return 2;
}


/*---------------------------------------------------------------------------*
  Name:         ENC_ConvUTF

  Description:  R[h̃Ro[gȂ܂B
                I[L͕t܂B

  Arguments:    dst        o͐obt@
                dst_len    o͐obt@̃obt@TCY
                dst_encode o̓GR[h
                src        ̓obt@
                src_len    ̓obt@̃obt@TCY
                src_encode ̓GR[h

  Returns:      o͂oCgԂ܂B
                sꍇɂ́AG[R[hԂ܂B
 *---------------------------------------------------------------------------*/
s32
ENC_ConvUTF(u8 *dst, s32 dst_len, ENCCode dst_encode, const u8 *src, s32 src_len, ENCCode src_encode)
{
    s32 out_len = 0;
    
    ENCReadCharFunc  read_func;
    ENCWriteCharFunc write_func;
    
    ENC_ASSERT( src != NULL );
    ENC_ASSERT( src_encode < ENC_CODE_MAX );
    ENC_ASSERT( dst_encode < ENC_CODE_MAX );
    ENC_ASSERT( src_len >= 0 );
    
    read_func  = sConvFunc[ src_encode ].readFunc;
    write_func = sConvFunc[ dst_encode ].writeFunc;
    
    while ( src_len > 0 )
    {
        u32 code;
        s32 len = read_func( src, (u32)src_len, &code );
        
        if ( len < 0 )
		{
			return len;
		} // ǂݍ݃G[
        
        src     += len;
        src_len -= len;
        
        // oCgI[_[}[N͖B
        if ( code == BOM )
		{
			continue;
		}
        if ( code == BOM_REVERSE )
		{
			return ENC_ERR_INVALID_FORMAT;
		}
        
        len = write_func( dst, dst_len, code );
        
        if ( len < 0 )
		{
			return len;
		} // o̓G[
        
        if ( dst )
        {
            dst     += len;
            dst_len -= len;
        }
        out_len += len;
    }
    
    return out_len;
}

/*---------------------------------------------------------------------------*
  Name:         ENC_SetUnicodeBOM

  Description:  oCgI[_[}[No͂܂B

  Arguments:    dst        o͐obt@
                dst_len    o͐obt@̃obt@TCY
                dst_encode o̓GR[h

  Returns:      o͂oCgԂ܂B
                sꍇɂ́AG[R[hԂ܂B
 *---------------------------------------------------------------------------*/
s32 
ENC_SetUnicodeBOM(u8 *dst, s32 dst_len, ENCCode dst_encode)
{
    ENCWriteCharFunc write_func;
    s32 len;
    
    ENC_ASSERT( dst_encode < ENC_CODE_MAX );
    
    write_func = sConvFunc[ dst_encode ].writeFunc;
    
    len = write_func( dst, dst_len, BOM );
    
    return len;
}


}}}
