/**-----------------------------------------------------------------------------

	@file		sndio_Util.h
	@author		tetsu
	@par		[]
				[eBeB֘A(J)

				Copyright (c)2010 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#include <wchar.h>
#include <nn/types.h>
#include <nn/fs.h>
#include <nn/dbg.h>
#include "sndio_UtfConv.h"
#include "sndio_Util.h"


namespace sndio { namespace detail {


////////////////////////////////////////////////////////////////////////////////
// G[֘A`
////////////////////////////////////////////////////////////////////////////////

/**-----------------------------------------------------------------------------

	@brief		FSñUgCupUgɕϊ

----------------------------------------------------------------------------- */
Result GetFsWholeError( nn::Result result )
{
	if (result.IsSuccess())
	{
		return ResultSuccess;
	}
	else if (result <= nn::fs::ResultNotFound())
	{
		if (result <= nn::fs::ResultMediaNotFound())
		{
			// fBAȂA͔Fł܂B
			// ResultNotFound ɑG[łB
			// fBA
			return ResultMediaNotFound;
		}

		// t@CEA[JCu܂B
		return ResultFileNotFound;
	}
	else if (result <= nn::fs::ResultAlreadyExists())
	{
		// t@CEA[JCu͊ɑ݂Ă܂B
		return ResultAlreadyExists;
	}
	else if (result <= nn::fs::ResultNotEnoughSpace())
	{
		// 󂫗eʂ܂B
		return ResultNotEnoughSpace;
	}
	else if (result <= nn::fs::ResultArchiveInvalidated())
	{
		// A[JCuɂȂ܂B
		return ResultArchiveInvalidated;
	}
	else if (result <= nn::fs::ResultOperationDenied())
	{
		if (result <= nn::fs::ResultWriteProtected())
		{
			// ݂֎~ɂĂ܂B
			// ResultOperationDenied ɑG[łB
			return ResultWriteProtected;
		}
		else if (result <= nn::fs::ResultMediaAccessError())
		{
			// fBAւ̃ANZXɃG[܂B
			// ResultOperationDenied ɑG[łB
			return ResultMediaAccessError;
		}

		// 삪ۂ܂B
		return ResultOperationDenied;
	}
	else if (result <= nn::fs::ResultBadFormat())
	{
		// sȃtH[}bgłB
		return ResultBadFormat;
	}

	// sȃG[iłŏoĂ͂ȂG[
	return ResultError;
}


////////////////////////////////////////////////////////////////////////////////
// ϊ
////////////////////////////////////////////////////////////////////////////////

/**-----------------------------------------------------------------------------

	@brief		R[hϊ
	@return		Ȃ
	@par		[]
				UTF8UTF16ϊ

----------------------------------------------------------------------------- */
void ConvertUTF82UTF16LE( u8* pDst, const u8* pSrc, u32* pLength )
{
	static const u32 c_strlen = 256;
	const u8* srcTmp = pSrc;
	u8 buffer[c_strlen];

	// oƏo悪ꏏ
	if (pSrc == pDst)
	{
		std::memset( buffer, 0, sizeof(buffer) );
		std::memcpy( buffer, pSrc, *pLength);
		srcTmp = buffer;
	}

	utf::CopyLStringZeroFill((char*)pDst, (const char*)srcTmp, c_strlen);

	int bom_len = utf::ENC_SetUnicodeBOM(
		(u8*)pDst,
		c_strlen,
		utf::ENC_CODE_UTF16LE
	);
	NN_TASSERT_(bom_len >= 0);

	int remain = c_strlen - bom_len - 2;
	int out_len = utf::ENC_ConvUTF(
		pDst + bom_len,
		remain,
		utf::ENC_CODE_UTF16LE,
		srcTmp,
		(s32)*pLength, utf::ENC_CODE_UTF8
	);

	// -1 ́Aobt@ȂȂ̂ŋ
	if (out_len < 0)
	{
		out_len = remain;
	}

	*pLength = static_cast<u32>(out_len + bom_len);

	pDst[*pLength] = 0;
	pDst[*pLength + 1] = 0;
}


/**-----------------------------------------------------------------------------

	@brief		R[hϊ
	@return		Ȃ
	@par		[]
				UTF16BEUTF16LEϊ

----------------------------------------------------------------------------- */
void ConvertBE2LE( u8* pDst, const u8* pSrc, u32* pLength )
{
	if (*pLength < 2)
		return;

	if (pSrc[0] == 0xFE && pSrc[1] == 0xFF)
	{
		pSrc += sizeof(wchar_t);

		for (u32 i = 0; i * 2  < *pLength; i++)
		{
			u8 tmp = pSrc[i * 2];
			pDst[i * 2] = pSrc[i * 2 + 1];
			pDst[i * 2 + 1] = tmp;
		}
	}
	else
	{
		pSrc += sizeof(wchar_t);
		std::memcpy( pDst, pSrc, *pLength );
	}
}


/**-----------------------------------------------------------------------------

	@brief		R[hϊ
	@return		Ȃ
	@par		[]
				UTF16֕ϊ

----------------------------------------------------------------------------- */
void Convert2UTF16LE( u8* pDst, const u8* pSrc, u32* pLength )
{
	bool bUTF8 = false;
	
	if (*pLength <= 0)
		return;

	if (*pLength < 2)
		bUTF8 = true;
	
	if (!bUTF8)
	{
		// BOM ݂邩
		if (pSrc[0] != 0xfe && pSrc[1] != 0xfe)
		{
			bUTF8 = true;
		}
	}

	if (bUTF8)
	{
		ConvertUTF82UTF16LE( pDst, pSrc, pLength );
	}
	else
	{
		ConvertBE2LE( pDst, pSrc, pLength );
	}
}


}}
