/**-----------------------------------------------------------------------------

	@file		VoiceManager.cpp
	@author		tetsu
	@par		[]
				^{CX̃f[^Ǘ`(J)

				Copyright (c)2010 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#include <ctype.h>
#include <algorithm>
#include <nn/dbg.h>
#include "sndio_Util.h"
#include "sndio_Mount.h"
#include "sndio_VoiceManager.h"


#if defined(APP_SNOTE) && defined(NN_BUILD_DEBUG)
#define PRINTRESULT( result )	\
	NN_LOG( "%s(%d)\n", __FILE__, __LINE__); NN_DBG_TPRINT_RESULT_( result )
#else
#define PRINTRESULT( result )
#endif

#define SD_ROOT_PATH			L"/Nintendo 3DS/Private/00020500/voice"
#define SD_DIR_PATH( page )		SNDIO_SD_VIRTUAL_ARCHIVE_NAMEW SD_ROOT_PATH page


////////////////////////////////////////////////////////////////////////////////
// [J`
////////////////////////////////////////////////////////////////////////////////

namespace
{
	// {CXf[^ۑ
	const wchar_t c_dirCtrNand[] = SNDIO_CTR_NAND_ARCHIVE_NAMEW L"/";
	const wchar_t c_dirTwlNand[] = SNDIO_TWL_NAND_ARCHIVE_NAMEW L"/";
	const wchar_t c_dirSDRoot[]  = SNDIO_SD_ARCHIVE_NAMEW SD_ROOT_PATH;

	// {CXf[^SDۑey[W(zpX)
	const wchar_t* c_filePathSD[] =
	{
		SD_DIR_PATH( L"/01" ),
		SD_DIR_PATH( L"/02" ),
		SD_DIR_PATH( L"/03" ),
		SD_DIR_PATH( L"/04" ),
		SD_DIR_PATH( L"/05" ),
		SD_DIR_PATH( L"/06" ),
		SD_DIR_PATH( L"/07" ),
		SD_DIR_PATH( L"/08" ),
		SD_DIR_PATH( L"/09" ),
		SD_DIR_PATH( L"/10" ),
	};

	// SDe|pX(zpX)
	const wchar_t c_filePathSDTemp[] = SNDIO_SD_VIRTUAL_ARCHIVE_NAMEW SD_ROOT_PATH L"/Temp";

	// ۑf[^̐擪ɂ
	const wchar_t c_dataNameHeader[] = L"V";
	const wchar_t c_dataNameHeaderTWL[] = L"voice";

	// gq
	const wchar_t c_dataExt[] = L"m4a";
	const wchar_t c_dataExtTWL[] = L"dat";

	// Rs[pobt@(100KBŒ)
	static const size_t c_fileSizeMax = 100 * 1024;
	static u8 workBuffer[ c_fileSizeMax ];
}


/**-----------------------------------------------------------------------------

	@brief		l֕ϊ
	@param[in]	digit			l
	@param[in]	offset			ϊItZbg
	@param[in]	size			ϊTCY
	@return		ϊl
	@par		[]
				w肳ꂽ͈̔͂𐔒lɕϊ܂<br/>
				4ȓŎgpł܂

----------------------------------------------------------------------------- */
static int ValueOf( const wchar_t* digit, int offset, int size )
{
	const int c_digitNum = 4;
	wchar_t work[ c_digitNum ];

	if (c_digitNum <= size)
		return 0;

	if (std::wcslen( digit ) <= offset)
		return 0;

	std::memcpy( work, &digit[ offset ], size * sizeof(wchar_t) );
	work[ size ] = L'\0';

	for (int i = 0; i < size; i++)
	{
		if (work[i] < L'0' || L'9' < work[i])
			return -1;
	}

	return std::wcstoul( work, NULL, 10 );
}


/**-----------------------------------------------------------------------------

	@brief		fBNg݃`FbN
	@param[in]	dirPath			fBNgpX
	@return		true fBNg݂
	@par		[]
				fBNg݂邩`FbN܂

----------------------------------------------------------------------------- */
static bool IsExistDirectory( const wchar_t* dirPath )
{
	nn::fs::Directory dir;
	nn::Result result = dir.TryInitialize(dirPath);

	if (result.IsSuccess())
	{
		dir.Finalize();
	}

	return result.IsSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		fBNg쐬
	@param[in]	dirPath			쐬fBNg
	@return		true 쐬
	@par		[]
				w肳ꂽfBNg쐬܂

----------------------------------------------------------------------------- */
static nn::Result CreateDirectory( const wchar_t* dirPath )
{
	static wchar_t dirWork[ nn::fs::MAX_FILE_PATH_LENGTH ];
	static wchar_t createPath[ nn::fs::MAX_FILE_PATH_LENGTH ];

	dirWork[0] = L'\0';
	createPath[0] = L'\0';

	wchar_t* pAddr = dirWork;
	wchar_t* pNextToken = NULL;

	std::wcsncpy( dirWork, dirPath, nn::fs::MAX_FILE_PATH_LENGTH );

	do
	{
		pAddr = std::wcstok(pAddr, L"/", &pNextToken);

		if (pAddr)
		{
			int len = std::wcslen(pAddr);
			std::wcsncat( createPath, pAddr, len );

			if (pAddr[len - 1] != L':')
			{
				if (!IsExistDirectory( createPath ))
				{
					nn::Result result = nn::fs::TryCreateDirectory(createPath);
					if (result.IsFailure())
					{
						return result;
					}
				}
			}

			std::wcsncat( createPath, L"/", nn::fs::MAX_FILE_PATH_LENGTH );

			pAddr = pNextToken;
		}
	} while (pAddr != NULL);

	return nn::ResultSuccess();
}


/**-----------------------------------------------------------------------------

	@brief		t@CRs[
	@param[in]	pathSrc			t@CpX
	@param[in]	pathDst			t@CpX
	@return		true ւ
	@par		[]
				t@C̃Rs[s܂<br/>
				fBNgA[JCuׂt@CRs[s܂

----------------------------------------------------------------------------- */
static bool CopyFile( const wchar_t* pathSrc, const wchar_t* pathDst )
{
	// IuWFNg
	static nn::os::LightSemaphore sem( 1, 1 );

	sem.Acquire();

	nn::Result result;

	// ǂݍ݌
	nn::fs::FileInputStream fsSrc;
	result = fsSrc.TryInitialize( pathSrc );
	if (result.IsFailure())
	{
		sem.Release();
		PRINTRESULT( result );
		return false;
	}

	// t@CTCY
	s64 fileSize = 0;
	result = fsSrc.TryGetSize( &fileSize );

	// obt@TCY𒴂Ăꍇ̓AT[gŗƂ
	NN_TASSERT_( fileSize <= c_fileSizeMax );

	if (result.IsFailure() || c_fileSizeMax < fileSize)
	{
		sem.Release();
		PRINTRESULT( result );
		return false;
	}

	s32 readCount = 0;
	result = fsSrc.TryRead( &readCount, workBuffer, fileSize );
	if (result.IsFailure() || readCount != fileSize)
	{
		sem.Release();
		PRINTRESULT( result );
		return false;
	}
	fsSrc.Finalize();

	// ݐ
	nn::fs::FileOutputStream fsDst;

	result = nn::fs::TryCreateFile( pathDst, fileSize );
	if (result.IsFailure())
	{
		sem.Release();
		PRINTRESULT( result );
		return false;
	}

	result = fsDst.TryInitialize( pathDst, false );
	if (result.IsFailure())
	{
		sem.Release();
		PRINTRESULT( result );
		return false;
	}

	s32 writeCount = 0;
	result = fsDst.TryWrite( &writeCount, workBuffer, fileSize, true );
	if (result.IsFailure() || writeCount != fileSize)
	{
		sem.Release();
		PRINTRESULT( result );
		return false;
	}

	sem.Release();
	return true;
}


/**-----------------------------------------------------------------------------

	@brief		t@Cړ
	@param[in]	path1			t@CpX
	@param[in]	path2			t@CpX
	@return		true ړ
	@par		[]
				t@C̈ړs܂<br/>
				fBNgA[JCuׂt@C̈ړs܂

----------------------------------------------------------------------------- */
static bool MoveFile( const wchar_t* path1, const wchar_t* path2,
					  const wchar_t* tmpPath )
{
	nn::Result result;

	if (tmpPath)
	{
		nn::fs::TryDeleteFile( tmpPath );

		if (CopyFile( path1, tmpPath ))
		{
			result = nn::fs::TryDeleteFile( path2 );
			if (result.IsSuccess() || result <= nn::fs::ResultNotFound())
			{
				if (nn::fs::TryRenameFile( tmpPath, path2 ).IsSuccess())
				{
					if (nn::fs::TryDeleteFile( path1 ).IsSuccess())
					{
						return true;
					}
				}

				// ŎsƂ̓f[^
			}
		}

		nn::fs::TryDeleteFile( tmpPath );
	}
	else
	{
		result = nn::fs::TryDeleteFile( path2 );

		if (result.IsSuccess() ||  result <= nn::fs::ResultNotFound())
		{
			if (CopyFile( path1, path2 ))
			{
				if (nn::fs::TryDeleteFile( path1 ).IsSuccess())
				{
					return true;
				}
			}
		}

		// ŎsƂ̓f[^
	}

	PRINTRESULT( result );
	return false;
}


/**-----------------------------------------------------------------------------

	@brief		t@Cړ(t)
	@param[in]	path1			t@CpX
	@param[in]	path2			t@CpX
	@return		true ړ
	@par		[]
				t@C̈ړs܂<br/>
				ƂāAfBNgA[JCuׂt@C̈ړ
				s܂

----------------------------------------------------------------------------- */
static bool MoveFileFirst( const wchar_t* path1, const wchar_t* path2 )
{
	nn::fs::TryDeleteFile( path2 );

	nn::Result result = nn::fs::TryRenameFile( path1, path2 );
	if (result.IsSuccess())
	{
		return true;
	}
	else
	{
		PRINTRESULT( result );
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		t@Cւ
	@param[in]	path1			t@CpX
	@param[in]	pathDst1		t@Cl[pX
	@param[in]	path2			t@CpX
	@param[in]	pathDst2		t@Cl[pX
	@param[in]	temp			ޔt@C
	@return		true ւ
	@par		[]
				t@C̓ւs܂<br/>
				tempɂ͈ꎞޔp̃t@Cw肵܂<br/>
				fBNgA[JCuׂt@C̓ւs܂

----------------------------------------------------------------------------- */
static bool SwapFile(
	const wchar_t* path1, const wchar_t* pathDst1,
	const wchar_t* path2, const wchar_t* pathDst2,
	const wchar_t* temp )
{
	nn::fs::TryDeleteFile( temp );

	if (MoveFile( path1, temp, NULL ))
	{
		if (MoveFile( path2, pathDst2, NULL ))
		{
			if (MoveFile( temp, pathDst1, NULL ))
			{
				return true;
			}
		}
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		t@Cւ(t)
	@param[in]	path1			t@CpX
	@param[in]	pathDst1		t@Cl[pX
	@param[in]	path2			t@CpX
	@param[in]	pathDst2		t@Cl[pX
	@param[in]	temp			ޔt@C
	@return		true ւ
	@par		[]
				t@C̓ւs܂<br/>
				tempɂ͈ꎞޔp̃t@Cw肵܂<br/>
				ƂāAfBNgA[JCuׂt@C̓ւ
				ł܂

----------------------------------------------------------------------------- */
static bool SwapFileFirst(
	const wchar_t* path1, const wchar_t* pathDst1,
	const wchar_t* path2, const wchar_t* pathDst2,
	const wchar_t* temp )
{
	nn::fs::TryDeleteFile( temp );

	nn::Result result = nn::fs::TryRenameFile( path1, temp );

	if (result.IsSuccess())
	{
		result = nn::fs::TryRenameFile( path2, pathDst2 );
		if (result.IsSuccess())
		{
			result = nn::fs::TryRenameFile( temp, pathDst1 );
			if (result.IsSuccess())
			{
				return true;
			}
			else
			{
				PRINTRESULT( result );
			}
		}
		else
		{
			PRINTRESULT( result );
		}
	}
	else
	{
		PRINTRESULT( result );
	}

	return false;
}


////////////////////////////////////////////////////////////////////////////////
// ǗNX
////////////////////////////////////////////////////////////////////////////////

namespace sndio { namespace detail {


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				f[^̃RXgN^

----------------------------------------------------------------------------- */
VoiceData::VoiceData(): 
m_use( 0 ),
m_valid( 0 ),
m_readOnly( 0 ),
m_page( 0 ),
m_index( 0 ),
m_color( 0 ),
m_fill( 0 ),
m_form( 0 )
{
}


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				f[^ŏ鉹f[^̃RXgN^

----------------------------------------------------------------------------- */
VoiceData::VoiceData( VoicePage page, int index, int color, int fill, int form ) :
m_use( 1 ),
m_valid( 1 ),
m_readOnly( 0 ),
m_page( page ),
m_index( static_cast<u8>(index) ),
m_color( static_cast<u8>(color) ),
m_fill( static_cast<u8>(fill) ),
m_form( static_cast<u8>(form) )
{
}


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				Ggŏ鉹f[^̃RXgN^

----------------------------------------------------------------------------- */
VoiceData::VoiceData( const wchar_t* entryName, VoicePage page ) :
m_use( 1 ),
m_valid( 1 ),
m_readOnly( 0 ),
m_page( page )
{
	m_index	= ValueOf( entryName, 1, 2 );
	m_color	= ValueOf( entryName, 3, 1 );
	m_fill	= ValueOf( entryName, 4, 1 );
	m_form	= ValueOf( entryName, 5, 1 );
}


/**-----------------------------------------------------------------------------

	@brief		gpĂ邩

----------------------------------------------------------------------------- */
bool VoiceData::IsUse() const
{
	return !!(0 < m_use);
}


/**-----------------------------------------------------------------------------

	@brief		Lȉ

----------------------------------------------------------------------------- */
bool VoiceData::IsValidVoice() const
{
	return !!(0 < m_valid);
}


/**-----------------------------------------------------------------------------

	@brief		ǂݎp

----------------------------------------------------------------------------- */
bool VoiceData::IsReadOnly() const
{
	return !!(0 < m_readOnly);
}


/**-----------------------------------------------------------------------------

	@brief		ۑĂy[W擾

----------------------------------------------------------------------------- */
VoicePage VoiceData::GetPage() const
{
	return static_cast<sndio::VoicePage>( m_page );
}


/**-----------------------------------------------------------------------------

	@brief		y[WɕۑĂʒu擾
	@return		0`VOICE_NUM_IN_PAGE

----------------------------------------------------------------------------- */
int VoiceData::GetIndex() const
{
	return static_cast<int>( m_index );
}


/**-----------------------------------------------------------------------------

	@brief		J[擾

----------------------------------------------------------------------------- */
int VoiceData::GetColor() const
{
	return static_cast<int>( m_color );
}


/**-----------------------------------------------------------------------------

	@brief		܂擾

----------------------------------------------------------------------------- */
int VoiceData::GetFill() const
{
	return static_cast<int>( m_fill );
}


/**-----------------------------------------------------------------------------

	@brief		ӂ̌`擾

----------------------------------------------------------------------------- */
int VoiceData::GetForm() const
{
	return static_cast<int>( m_form );
}


/**-----------------------------------------------------------------------------

	@brief		gp邩

----------------------------------------------------------------------------- */
void VoiceData::SetUse( bool use )
{
	m_use = use ? 1 : 0;
}


/**-----------------------------------------------------------------------------

	@brief		Lȉ

----------------------------------------------------------------------------- */
void VoiceData::SetValidVoice( bool valid )
{
	m_valid = valid ? 1 : 0;
}


/**-----------------------------------------------------------------------------

	@brief		ǂݎp

----------------------------------------------------------------------------- */
void VoiceData::SetReadOnly( bool readOnly )
{
	m_readOnly = readOnly ? 1 : 0;
}


/**-----------------------------------------------------------------------------

	@brief		y[Wݒ

----------------------------------------------------------------------------- */
void VoiceData::SetPage( VoicePage page )
{
	m_page = static_cast<u8>(page);
}


/**-----------------------------------------------------------------------------

	@brief		CfbNXݒ(0`17)

----------------------------------------------------------------------------- */
void VoiceData::SetIndex( int index )
{
	m_index = static_cast<u8>(index);
}


/**-----------------------------------------------------------------------------

	@brief		y[WƃCfbNX̓ւ

----------------------------------------------------------------------------- */
void VoiceData::Swap( VoiceData* pFrom )
{
	std::swap( this->m_index, pFrom->m_index );
	std::swap( this->m_page, pFrom->m_page );
}


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				f[^ŏ鉹f[^̃RXgN^

----------------------------------------------------------------------------- */
TwlVoiceData::TwlVoiceData( int index, int color, int fill, int form,
	const nn::fnd::DateTime& dateTime )
{
	m_page = VOICE_PAGE_TWL_NAND;

	m_index = index;
	m_color = color;
	m_fill = fill;
	m_form = form;
	m_dateTime = dateTime;

	m_copy = 0;
	m_version = 0;
	m_use = 1;
}


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				Ggŏ鉹f[^̃RXgN^

----------------------------------------------------------------------------- */
TwlVoiceData::TwlVoiceData( const wchar_t* entryName )
{
	m_page = VOICE_PAGE_TWL_NAND;

	// CfbNX
	m_index = ValueOf( entryName, 5, 2 );
	NN_MINMAX_ASSERT( m_index, 0, sndio::VOICE_NUM_IN_PAGE - 1 );

	// t
	s32 year	= ValueOf( entryName,  7, 2 );
	s32 month	= ValueOf( entryName,  9, 2 );
	s32 day		= ValueOf( entryName, 11, 2 );
	s32 hour	= ValueOf( entryName, 13, 2 );
	s32 minute	= ValueOf( entryName, 15, 2 );
	s32 second	= ValueOf( entryName, 17, 2 );
	NN_MINMAX_ASSERT( year,		0, 100 );
	NN_MINMAX_ASSERT( month,	1, 13 );
	NN_MINMAX_ASSERT( day,		1, 32 );
	NN_MINMAX_ASSERT( hour,		0, 24 );
	NN_MINMAX_ASSERT( minute,	0, 60 );
	NN_MINMAX_ASSERT( second,	0, 60 );
	m_dateTime = nn::fnd::DateTime( year + 2000, month, day, hour, minute, second );

	// ̑̏
	m_color		= ValueOf( entryName, 19, 3 );
	m_fill		= ValueOf( entryName, 22, 1 );
	m_form		= ValueOf( entryName, 23, 1 );
	m_copy		= ValueOf( entryName, 24, 2 );
	m_version	= ValueOf( entryName, 26, 2 );
	m_use		= 1;
}


/**-----------------------------------------------------------------------------

	@brief		t擾

----------------------------------------------------------------------------- */
const nn::fnd::DateTime& TwlVoiceData::GetDateTime() const
{
	return m_dateTime;
}


/**-----------------------------------------------------------------------------

	@brief		Rs[l擾

----------------------------------------------------------------------------- */
int TwlVoiceData::GetCopy() const
{
	return static_cast<int>( m_copy );
}


/**-----------------------------------------------------------------------------

	@brief		o[W擾

----------------------------------------------------------------------------- */
int TwlVoiceData::GetVersion() const
{
	return static_cast<int>( m_version );
}


////////////////////////////////////////////////////////////////////////////////
// ǗNX
////////////////////////////////////////////////////////////////////////////////


/**-----------------------------------------------------------------------------

	@brief		CTRL̈pX擾

----------------------------------------------------------------------------- */
const wchar_t* VoiceManager::GetNandPath()
{
	return c_dirCtrNand;
}


/**-----------------------------------------------------------------------------

	@brief		{CXpX擾

----------------------------------------------------------------------------- */
const wchar_t* VoiceManager::GetTwlPath()
{
	return c_dirTwlNand;
}


/**-----------------------------------------------------------------------------

	@brief		SD{CXpX擾(zpX)
	@param[in]	wy[W
	@return		ۑpX

----------------------------------------------------------------------------- */
const wchar_t* VoiceManager::GetDirectoryPath( VoicePage page )
{
	switch (page)
	{
	case VOICE_PAGE_TWL_NAND:
		return c_dirTwlNand;

	case VOICE_PAGE_CTR_NAND:
		return c_dirCtrNand;

	case VOICE_PAGE_SD_01:
	case VOICE_PAGE_SD_02:
	case VOICE_PAGE_SD_03:
	case VOICE_PAGE_SD_04:
	case VOICE_PAGE_SD_05:
	case VOICE_PAGE_SD_06:
	case VOICE_PAGE_SD_07:
	case VOICE_PAGE_SD_08:
	case VOICE_PAGE_SD_09:
	case VOICE_PAGE_SD_10:
		return c_filePathSD[ static_cast<int>(page - VOICE_PAGE_SD_01) ];
	}

	return NULL;
}


/**-----------------------------------------------------------------------------

	@brief		SDe|pX擾(zpX)
	@return		e|pX

----------------------------------------------------------------------------- */
const wchar_t* VoiceManager::GetDirectoryTempPath()
{
	return c_filePathSDTemp;
}


/**-----------------------------------------------------------------------------

	@brief		{CXÓ`FbN
	@param[in]	entryName		t@CGg
	@retval		true 
	@retval		false s
	@par		[]
				t@C̃Gg{CX̑Ó`FbN܂

----------------------------------------------------------------------------- */
bool VoiceManager::ValidateVoice( const wchar_t* entryName )
{
	if (std::wcslen( entryName ) == 10)
	{
		if (std::toupper(entryName[0]) != c_dataNameHeader[0])
			return false;

		int value = -1;

		// CfbNX`FbN
		value = ValueOf( entryName, 1, 2 );
		if (value < 0 || VOICE_NUM_IN_PAGE <= value)
			return false;

		// J[
		value = ValueOf( entryName, 3, 1 );
		if (value < 0 || VOICE_COLOR_NUM <= value)
			return false;

		// ܂
		value = ValueOf( entryName, 4, 1 );
		if (value < 0 || VOICE_FILL_100 < value)
			return false;

		// `
		value = ValueOf( entryName, 5, 1 );
		if (value < VOICE_FORM_NORMAL || VOICE_FORM_NUM <= value)
			return false;

		if ((entryName[7] == L'm' || entryName[7] == L'M') &&
			(entryName[8] == L'4') &&
			(entryName[9] == L'a' || entryName[9] == L'A'))
		{
			return true;
		}
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		KV[{CXÓ`FbN
	@param[in]	entryName		t@CGg
	@retval		true 
	@retval		false s
	@par		[]
				t@C̃Gg烌KV[{CX̑Ó`FbN܂

----------------------------------------------------------------------------- */
bool VoiceManager::ValidateTwlVoice( const wchar_t* entryName )
{
	if (std::tolower(entryName[0]) == c_dataNameHeaderTWL[0] &&
		std::tolower(entryName[1]) == c_dataNameHeaderTWL[1] &&
		std::tolower(entryName[2]) == c_dataNameHeaderTWL[2] &&
		std::tolower(entryName[3]) == c_dataNameHeaderTWL[3] &&
		std::tolower(entryName[4]) == c_dataNameHeaderTWL[4])
	{
		int value = -1;

		// CfbNX`FbN
		value = ValueOf( entryName, 5, 2 );
		if (value < 0 || VOICE_NUM_IN_PAGE <= value)
			return false;

		// J[
		value = ValueOf( entryName, 19, 3 );
		if (value < 0 || VOICE_COLOR_NUM <= value)
			return false;

		// ܂
		value = ValueOf( entryName, 22, 1 );
		if (value < 0 || VOICE_FILL_100 < value)
			return false;

		// `
		value = ValueOf( entryName, 23, 1 );
		if (value < VOICE_FORM_NORMAL || VOICE_FORM_NUM <= value)
			return false;

		if (std::wcslen( entryName ) == 32)
		{
			return true;
		}
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		\z
	@return		Ȃ
	@par		[]
				RXgN^

----------------------------------------------------------------------------- */
VoiceManager::VoiceManager() :
m_initialized( false ),
m_useCtrNand( false ),
m_useSD( false ),
m_twlSupport( false ),
m_twlConsistency( true ),
m_cs( nn::WithInitialize() )
{
	std::memset( m_voice, 0, sizeof(m_voice) );
	std::memset( m_twlVoice, 0, sizeof(m_twlVoice) );

	for (int i = 0; i < VOICE_PAGE_NUM; i++)
	{
		for (int j = 0; j < VOICE_NUM_IN_PAGE; j++)
		{
			new( reinterpret_cast<void*>(&m_voice[i][j]) ) VoiceData();
		}
	}

	for (int i = 0; i < VOICE_NUM_IN_PAGE; i++)
	{
		new( reinterpret_cast<void*>(&m_twlVoice[i]) ) TwlVoiceData();
	}
}


/**-----------------------------------------------------------------------------

	@brief		
	@return		Ȃ
	@par		[]
				fXgN^

----------------------------------------------------------------------------- */
VoiceManager::~VoiceManager()
{
	Finalize();
}


/**-----------------------------------------------------------------------------

	@brief		
	@param[in]	useSD			SD̈̃{CXǂݍނ (ftHgtrue)
	@param[in]	useCtrNand		CTR̈̃{CXǂݍނ (ftHgtrue)
	@param[in]	useTwlNand		TWL̈̃{CXǂݍނ (ftHgfalse)
	@return		Result
	@par		[]
				s܂<br/>
				{CXf[^̃fBNgTŝłȂ
				sԂ܂<br/>
				AvłTWLA[JCüANZX邱Ƃ͖̂
				useTwlNand = false ɂďĂ

----------------------------------------------------------------------------- */
Result VoiceManager::Initialize( bool useSD, bool useCtrNand, bool useTwlNand )
{
	sndio::Result sresult( ResultSuccess );

	Finalize();

	nn::os::CriticalSection::ScopedLock lock( m_cs );

	if (useTwlNand)
	{
		m_twlSupport = true;

		// TWL̈gƂ́ACTR NAND̈̏Kv
		NN_TASSERT_( useCtrNand );

		// {CXǂݍ
		ReadTwlNandVoice();
	}
	else
	{
		m_twlSupport = false;
	}

	// CTRL̈̃{CXǂݍ({CXǂݍ܂ꂽɍs)
	if (useCtrNand)
	{
		m_useCtrNand = true;
		ReadCtrNandVoice();
	}

	// ƂĂ邩`FbN(̎_CTRTWL̐̓`FbNłĂ)
	if (IsTwlSupport() && !IsTwlNeedRestore())
	{
		for (int i = 0; i < VOICE_NUM_IN_PAGE; i++)
		{
			if (Contains( VOICE_PAGE_TWL_NAND, i ) && !Contains( VOICE_PAGE_CTR_NAND, i ))
			{
				// ƂĂȂ
				m_twlConsistency = false;
				break;
			}
		}
	}

	// SD̈̃{CXǂݍ
	if (useSD)
	{
		m_useSD = true;
		sresult = ReadSDVoice();

		if (sresult != ResultSuccess)
		{
			return sresult;
		}
	}

	m_initialized = true;
	return ResultSuccess;
}


/**-----------------------------------------------------------------------------

	@brief		I
	@return		Ȃ
	@par		[]
				IsASf[^J܂<br/>
				̃\bh̓fXgN^ŌĂ΂̂
				IɌĂԕKv͂܂

----------------------------------------------------------------------------- */
void VoiceManager::Finalize()
{
	nn::os::CriticalSection::ScopedLock lock( m_cs );

	for (int i = 0; i < VOICE_PAGE_NUM; i++)
	{
		for (int j = 0; j < VOICE_NUM_IN_PAGE; j++)
		{
			m_voice[i][j].SetUse( false );
		}
	}

	for (int i = 0; i < VOICE_NUM_IN_PAGE; i++)
	{
		m_twlVoice[i].SetUse( false );
	}

	m_initialized = false;
	m_useCtrNand = true;
	m_useSD = true;
	m_twlSupport = false;
	m_twlConsistency = true;
	
}


/**-----------------------------------------------------------------------------

	@brief		sĂ邩ǂ
	@return		true ς

----------------------------------------------------------------------------- */
bool VoiceManager::IsInitialized() const
{
	return m_initialized;
}


/**-----------------------------------------------------------------------------

	@brief		{CXT|[g
	@return		true {CX̏s
	@par		[]
				{CX(DSiTEh)̏sǂ擾܂<br/>
				truȅꍇANAND̈𑀍(폜AւARs[)
				sƂATWLTEhA[JCü̃f[^삵܂<br/>
				ǉ̂ AddTwlVoice gpĂ

----------------------------------------------------------------------------- */
bool VoiceManager::IsTwlSupport() const
{
	return m_twlSupport;
}


/**-----------------------------------------------------------------------------

	@brief		TWL̈ƐƂĂȂ̂ŁAKv
	@retval		true Kv
	@retval		false ͕sv
	@par		[]
				TWLƋLgZ[uf[^̐ƂĂȂꍇtrue
				Ԃ܂B<br/>
				CTRTWL  TWLCTR ̗ł̐`FbNs܂

----------------------------------------------------------------------------- */
bool VoiceManager::IsTwlNeedRestore() const
{
	return !m_twlConsistency;
}


/**-----------------------------------------------------------------------------

	@brief		f[^ł݂邩mF܂
				CTR̈SDJ[ḧ旼T܂
	@retval		true ݂
	@retval		false ݂Ȃ

----------------------------------------------------------------------------- */
bool VoiceManager::IsExistsVoice() const
{
	for (int i = 0; i < VOICE_NUM_IN_PAGE; i++)
	{
		if (Contains( VOICE_PAGE_CTR_NAND, i ))	return true;
		if (Contains( VOICE_PAGE_SD_01, i ))	return true;
		if (Contains( VOICE_PAGE_SD_02, i ))	return true;
		if (Contains( VOICE_PAGE_SD_03, i ))	return true;
		if (Contains( VOICE_PAGE_SD_04, i ))	return true;
		if (Contains( VOICE_PAGE_SD_05, i ))	return true;
		if (Contains( VOICE_PAGE_SD_06, i ))	return true;
		if (Contains( VOICE_PAGE_SD_07, i ))	return true;
		if (Contains( VOICE_PAGE_SD_08, i ))	return true;
		if (Contains( VOICE_PAGE_SD_09, i ))	return true;
		if (Contains( VOICE_PAGE_SD_10, i ))	return true;
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		w̃y[W肵܂
	@retval		true 
	@retval		false L

----------------------------------------------------------------------------- */
bool VoiceManager::IsInvalidPage( VoicePage page )
{
	return !m_validatePage[ page ];
}


/**-----------------------------------------------------------------------------

	@brief		ǂԂ܂
	@retval		true 
	@retval		false ҋ@

----------------------------------------------------------------------------- */
bool VoiceManager::IsBusy() const
{
	return m_cs.IsLocked();
}


/**-----------------------------------------------------------------------------

	@brief		č\z
	@par		[]
				̍č\zs܂B<br/>
				Ďŝł̃\bh͎Ԃ܂B

----------------------------------------------------------------------------- */
Result VoiceManager::Restore()
{
	bool useSD = m_useSD;
	bool useCtrNand = m_useCtrNand;
	bool twlSupport = m_twlSupport;

	// ďs
	Finalize();
	return Initialize( useSD, useCtrNand, twlSupport );
}


/**-----------------------------------------------------------------------------

	@brief		SD؂藣
	@par		[]
				y[WPʂ̐؂藣s܂

----------------------------------------------------------------------------- */
void VoiceManager::PurgeSD()
{
	nn::os::CriticalSection::ScopedLock lock( m_cs );

	for (int i = VOICE_PAGE_SD_01; i < VOICE_PAGE_NUM; i++)
	{
		for (int j = 0; j < VOICE_NUM_IN_PAGE; j++)
		{
			m_voice[i][j].SetUse( false );
		}
	}

	m_useSD = false;
}


/**-----------------------------------------------------------------------------

	@brief		SDXgA
	@par		[]
				y[WPʂ̃XgAs܂

----------------------------------------------------------------------------- */
Result VoiceManager::RestoreSD()
{
	nn::os::CriticalSection::ScopedLock lock( m_cs );

	m_useSD = true;
	return ReadSDVoice();
}


/**-----------------------------------------------------------------------------

	@brief		{CX̎擾
	@param[in]	page			Ώۃy[Wԍ
	@param[in]	index			ΏۃCfbNXԍ(0`17)
	@return		{CXf[^ւ̃|C^
	@par		[]
				w肳ꂽ{CX̃f[^ւ̃|C^擾܂<br/>
				CfbNX̎w͉L̕тŎw肵܂<br/>
				 -----------------
				| 0  1  2  3  4  5|
				| 6  7  8  9 10 11|
				|12 13 14 15 16 17|
				 -----------------

----------------------------------------------------------------------------- */
VoiceData* VoiceManager::GetVoiceData( VoicePage page, int index )
{
	NN_MINMAX_TASSERT_( index, 0, VOICE_NUM_IN_PAGE - 1 );

	if (page == VOICE_PAGE_TWL_NAND)
	{
		return static_cast<VoiceData*>( &m_twlVoice[ index ] );
	}
	else
	{
		NN_MINMAX_TASSERT_( page, 0, VOICE_PAGE_NUM - 1 );
		return &m_voice[ static_cast<int>(page) ][ index ];
	}
}


/**-----------------------------------------------------------------------------

	@brief		{CX̎擾
	@param[in]	page			Ώۃy[Wԍ
	@param[in]	index			ΏۃCfbNXԍ(0`17)
	@return		{CXf[^ւ̃|C^(const)
	@par		[]
				w肳ꂽ{CX̃f[^ւ̃|C^擾܂

----------------------------------------------------------------------------- */
const VoiceData* VoiceManager::GetVoiceData( VoicePage page, int index ) const
{
	return const_cast<VoiceManager*>(this)->GetVoiceData( page, index );
}


/**-----------------------------------------------------------------------------

	@brief		{CX̎擾
	@param[in]	index			ΏۃCfbNXԍ(0`17)
	@return		{CXf[^ւ̃|C^
	@par		[]
				w肳ꂽ{CX̃f[^ւ̃|C^擾܂

----------------------------------------------------------------------------- */
TwlVoiceData* VoiceManager::GetTwlVoiceData( int index )
{
	NN_MINMAX_TASSERT_( index, 0, VOICE_NUM_IN_PAGE - 1 );
	return &m_twlVoice[ index ];
}


/**-----------------------------------------------------------------------------

	@brief		{CX̎擾
	@param[in]	index			ΏۃCfbNXԍ(0`17)
	@return		{CXf[^ւ̃|C^(const)
	@par		[]
				w肳ꂽ{CX̃f[^ւ̃|C^擾܂

----------------------------------------------------------------------------- */
const TwlVoiceData* VoiceManager::GetTwlVoiceData( int index ) const
{
	return const_cast<VoiceManager*>(this)->GetTwlVoiceData( index );
}


/**-----------------------------------------------------------------------------

	@brief		{CX̒ǉ
	@param[in]	page			Ώۃy[Wԍ
	@param[in]	index			ΏۃCfbNXԍ(0`17)
	@param[in]	color			J[
	@param[in]	fill			܂肮
	@param[in]	form			ӂ`
	@retval		true ǉ
	@retval		false ǉs
	@par		[]
				w肳ꂽCfbNX̃{CX̒ǉs܂<br/>
				ɓo^ĂsȃCfbNX̏ꍇ͎s܂

----------------------------------------------------------------------------- */
bool VoiceManager::AddVoice( VoicePage page, int index, int color, int fill, int form )
{
	if (Contains( page, index ))
		return false;

	VoiceData voice( page, index, color, fill, form );

	if (SetVoiceData( page, index, voice ))
	{
		return true;
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		{CX̒ǉ
	@param[in]	index			ΏۃCfbNXԍ(0`17)
	@param[in]	color			J[
	@param[in]	fill			܂肮
	@param[in]	form			ӂ`
	@param[in]	dateTime		XV
	@retval		true ǉ
	@retval		false ǉs
	@par		[]
				w肳ꂽCfbNX̋{CX̒ǉs܂<br/>
				ɓo^ĂsȃCfbNX̏ꍇ͎s܂

----------------------------------------------------------------------------- */
bool VoiceManager::AddTwlVoice(
	int index, int color, int fill, int form,
	const nn::fnd::DateTime& dateTime )
{
	if (IsTwlSupport() == false)
		return false;

	if (Contains( VOICE_PAGE_TWL_NAND, index ))
		return false;

	TwlVoiceData voice( index, color, fill, form, dateTime );

	if (index < 0 || VOICE_NUM_IN_PAGE <= index)
		return false;

	m_twlVoice[ index ] = voice;

	return true;
}


/**-----------------------------------------------------------------------------

	@brief		{CX̕ۑpX擾
	@param[out]	path			擾pX
	@param[in]	page			Ώۃy[Wԍ
	@param[in]	index			ΏۃCfbNXԍ(0`17)
	@retval		true 擾
	@retval		false 擾s
	@par		[]
				w肳ꂽ{CX̃tpX擾܂

----------------------------------------------------------------------------- */
bool VoiceManager::GetVoicePath( VoicePath& path, VoicePage page, int index ) const
{
	std::memset( path, 0, sizeof(path) );

	if (Contains( page, index ) == false)
		return false;

	// KV[ۑpX
	if (page == VOICE_PAGE_TWL_NAND)
	{
		if (const TwlVoiceData* pTwl = GetTwlVoiceData( index ))
		{
			const nn::fnd::DateTime& dateTime = pTwl->GetDateTime();
			Format(
				path,
				L"%ls%ls%02d%02d%02d%02d%02d%02d%02d%03d%01d%01d%02d%02d.%ls",
				c_dirTwlNand,
				const_cast<wchar_t*>(c_dataNameHeaderTWL),
				pTwl->GetIndex(),
				(dateTime.GetYear() % 100),
				dateTime.GetMonth(),
				dateTime.GetDay(),
				dateTime.GetHour(),
				dateTime.GetMinute(),
				dateTime.GetSecond(),
				pTwl->GetColor(),
				pTwl->GetFill(),
				pTwl->GetForm(),
				pTwl->GetCopy(),
				pTwl->GetVersion(),
				const_cast<wchar_t*>(c_dataExtTWL)
			);
		}
	}
	else
	{
		if (const VoiceData* pVoice = GetVoiceData( page, index ))
		{
			switch (pVoice->GetPage())
			{
			case VOICE_PAGE_CTR_NAND:

				// NAND̕ۑpX
				Format(
					path,
					L"%ls%ls%02d%01d%01d%01d.%ls",
					c_dirCtrNand,
					const_cast<wchar_t*>(c_dataNameHeader),
					pVoice->GetIndex(),
					pVoice->GetColor(),
					pVoice->GetFill(),
					pVoice->GetForm(),
					const_cast<wchar_t*>(c_dataExt)
					);
				break;

			case VOICE_PAGE_SD_01:
			case VOICE_PAGE_SD_02:
			case VOICE_PAGE_SD_03:
			case VOICE_PAGE_SD_04:
			case VOICE_PAGE_SD_05:
			case VOICE_PAGE_SD_06:
			case VOICE_PAGE_SD_07:
			case VOICE_PAGE_SD_08:
			case VOICE_PAGE_SD_09:
			case VOICE_PAGE_SD_10:

				// SD̕ۑpX𐶐
				Format(
					path,
					L"%ls/%ls%02d%01d%01d%01d.%ls",
					c_filePathSD[ static_cast<int>(pVoice->GetPage() - VOICE_PAGE_SD_01) ],
					const_cast<wchar_t*>(c_dataNameHeader),
					pVoice->GetIndex(),
					pVoice->GetColor(),
					pVoice->GetFill(),
					pVoice->GetForm(),
					const_cast<wchar_t*>(c_dataExt)
					);
				break;
			}
		}
	}

	return true;
}


/**-----------------------------------------------------------------------------

	@brief		{CX̃e|ۑpX擾
	@param[out]	path			擾pX
	@param[in]	page			Ώۃy[Wԍ
	@retval		true 擾
	@retval		false 擾s
	@par		[]
				w肳ꂽ{CX̃e|pɎgpX擾܂<br/>
				擾ꂽpX̓y[WɂЂƂȂ̂ŔrȈKvł

----------------------------------------------------------------------------- */
bool VoiceManager::GetVoiceTempPath( VoicePath& path, VoicePage page ) const
{
	// ւ̃e|pX
	static const wchar_t c_temp[] = L"__temp.m4a";
	std::memset( path, 0, sizeof(path) );

	switch (page)
	{
	// KV[e|pX
	case VOICE_PAGE_TWL_NAND:
		Format( path, L"%ls__temp.dat", c_dirTwlNand );
		break;

	// NAND̃e|pX
	case VOICE_PAGE_CTR_NAND:
		Format( path, L"%ls%ls", c_dirCtrNand, c_temp );
		break;

	// SD̃e|pX𐶐
	case VOICE_PAGE_SD_01:
	case VOICE_PAGE_SD_02:
	case VOICE_PAGE_SD_03:
	case VOICE_PAGE_SD_04:
	case VOICE_PAGE_SD_05:
	case VOICE_PAGE_SD_06:
	case VOICE_PAGE_SD_07:
	case VOICE_PAGE_SD_08:
	case VOICE_PAGE_SD_09:
	case VOICE_PAGE_SD_10:
		Format(
			path,
			L"%ls/%ls",
			c_filePathSD[ static_cast<int>(page - VOICE_PAGE_SD_01) ],
			c_temp
		);
		break;

	default:
		NN_TASSERT_( false );
		return false;
	}

	return true;
}


/**-----------------------------------------------------------------------------

	@brief		݃`FbN
	@param[in]	page			Ώۃy[Wԍ
	@param[in]	index			ΏۃCfbNXԍ(0`17)
	@retval		true ݂Ă
	@retval		false ݂ĂȂ
	@par		[]
				w肳ꂽ{CXf[^Ɋ܂܂Ă邩擾܂

----------------------------------------------------------------------------- */
bool VoiceManager::Contains( VoicePage page, int index ) const
{
	switch (page)
	{
	case VOICE_PAGE_TWL_NAND:
		return m_twlVoice[ index ].IsUse();

	default:
		NN_MINMAX_TASSERT_( page, 0, VOICE_PAGE_NUM - 1 );
		NN_MINMAX_TASSERT_( index, 0, VOICE_NUM_IN_PAGE - 1 );
		return m_voice[ page ][ index ].IsUse();
	}
}


/**-----------------------------------------------------------------------------

	@brief		폜
	@param[in]	page			폜Ώۃy[Wԍ
	@param[in]	index			폜ΏۃCfbNXԍ(0`17)
	@param[in]	withFile		t@Cs(ftHgtrue)
	@retval		true 
	@retval		false s
	@par		[]
				w肳ꂽ{CX̍폜s܂

----------------------------------------------------------------------------- */
bool VoiceManager::Delete( VoicePage page, int index, bool withFile )
{
	nn::os::CriticalSection::ScopedLock lock( m_cs );

	if (withFile)
	{
		VoicePath path;
		if (GetVoicePath( path, page, index ))
		{
			nn::Result result = nn::fs::TryDeleteFile( ToReal(path) );
			if (result.IsFailure())
			{
				PRINTRESULT( result );
			}
		}
	}

	switch (page)
	{
	case VOICE_PAGE_TWL_NAND:
		m_twlVoice[ index ].SetUse( false );
		break;

	default:
		m_voice[ page ][ index ].SetUse( false );
		break;
	}

	return true;
}


/**-----------------------------------------------------------------------------

	@brief		ւ
	@param[in]	page1			y[Wԍ
	@param[in]	index1			CfbNXԍ(0`17)
	@param[in]	page2			y[Wԍ
	@param[in]	index2			CfbNXԍ(0`17)
	@param[in]	withFile		t@Cs(ftHgtrue)
	@retval		true 
	@retval		false s
	@par		[]
				w肳ꂽ{CX̌s܂

----------------------------------------------------------------------------- */
bool VoiceManager::Swap( VoicePage page1, int index1, VoicePage page2, int index2, bool withFile )
{
	nn::os::CriticalSection::ScopedLock lock( m_cs );

	VoicePath path1;
	VoicePath path2;

	bool result1 = GetVoicePath( path1, page1, index1 );
	bool result2 = GetVoicePath( path2, page2, index2 );

	ToReal( path1 );
	ToReal( path2 );

	if (result1 && result2)
	{
		VoicePath pathDst1;
		VoicePath pathDst2;
		VoicePath pathTemp;

		// CfbNXƃy[Wւ
		GetVoiceData( page1, index1 )->Swap( GetVoiceData( page2, index2 ) );
		GetVoicePath( pathDst1, page1, index1 );
		GetVoicePath( pathDst2, page2, index2 );
		GetVoiceTempPath( pathTemp, page1 );

		ToReal( pathDst1 );
		ToReal( pathDst2 );
		ToReal( pathTemp );

		bool result = false;
		if (LookStrideArchive( page1, page2 ))
		{
			if (withFile)
			{
				result = SwapFile( path1, pathDst1, path2, pathDst2, pathTemp );
			}
			else
			{
				result = true;
			}
		}
		else
		{
			if (withFile)
			{
				result = SwapFileFirst( path1, pathDst1, path2, pathDst2, pathTemp );
			}
			else
			{
				result = true;
			}
		}

		if (result)
		{
			if (page1 == VOICE_PAGE_TWL_NAND && page2 == VOICE_PAGE_TWL_NAND)
			{
				std::swap( m_twlVoice[ index1 ], m_twlVoice[index2 ] );
			}
			else
			{
				std::swap( m_voice[ page1 ][ index1 ], m_voice[ page2 ][ index2 ] );
			}
		}
		else
		{
			// ߂
			GetVoiceData( page1, index1 )->Swap( GetVoiceData( page2, index2 ) );
		}

		return result;
	}
	else if (result1 && !result2)
	{
		// Rs[悪
		VoiceData* pVoice = GetVoiceData( page1, index1 );
		NN_NULL_TASSERT_( pVoice );

		if (pVoice)
		{
			pVoice->SetPage( page2 );
			pVoice->SetIndex( index2 );

			if (page1 == VOICE_PAGE_TWL_NAND && page2 == VOICE_PAGE_TWL_NAND)
			{
				std::swap( m_twlVoice[ index1 ], m_twlVoice[index2 ] );
			}
			else
			{
				std::swap( m_voice[ page1 ][ index1 ], m_voice[ page2 ][ index2 ] );
			}
		}

		// ւ̃pX擾
		if (withFile)
		{
			if (GetVoicePath( path2, page2, index2 ))
			{
				ToReal( path2 );

				if (LookStrideArchive( page1, page2 ))
				{
					VoicePath pathTemp;
					GetVoiceTempPath( pathTemp, page2 );
					ToReal( pathTemp );

					if (MoveFile( path1, path2, pathTemp ))
					{
						return true;
					}
					else
					{
						pVoice->SetPage( page1 );
						pVoice->SetIndex( index1 );
					}
				}
				else
				{
					if (MoveFileFirst( path1, path2 ))
					{
						return true;
					}
					else
					{
						pVoice->SetPage( page1 );
						pVoice->SetIndex( index1 );
					}
				}
			}
		}
		else
		{
			return true;
		}
	}
	else if (!result1 && result2)
	{
		// Rs[
		VoiceData* pVoice = GetVoiceData( page2, index2 );
		NN_NULL_TASSERT_( pVoice );

		if (pVoice)
		{
			pVoice->SetPage( page1 );
			pVoice->SetIndex( index1 );

			if (page1 == VOICE_PAGE_TWL_NAND && page2 == VOICE_PAGE_TWL_NAND)
			{
				std::swap( m_twlVoice[ index1 ], m_twlVoice[index2 ] );
			}
			else
			{
				std::swap( m_voice[ page1 ][ index1 ], m_voice[ page2 ][ index2 ] );
			}
		}

		// ւ̃pX擾
		if (withFile)
		{
			if (GetVoicePath( path1, page1, index1 ))
			{
				ToReal( path1 );

				if (LookStrideArchive( page1, page2 ))
				{
					VoicePath pathTemp;
					GetVoiceTempPath( pathTemp, page1 );
					ToReal( pathTemp );

					if (MoveFile( path2, path1, pathTemp ))
					{
						return true;
					}
					else
					{
						pVoice->SetPage( page2 );
						pVoice->SetIndex( index2 );
					}
				}
				else
				{
					if (MoveFileFirst( path2, path1 ))
					{
						return true;
					}
					else
					{
						pVoice->SetPage( page2 );
						pVoice->SetIndex( index2 );
					}
				}
			}
		}
		else
		{
			return true;
		}
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		Rs[
	@param[in]	pageSrc			Rs[y[Wԍ
	@param[in]	indexSrc		Rs[CfbNXԍ(0`17)
	@param[in]	pageDst			Rs[y[Wԍ
	@param[in]	indexDst		Rs[CfbNXԍ(0`17)
	@param[in]	withFile		t@Cs(ftHgtrue)
	@retval		true 
	@retval		false s
	@par		[]
				{CXw肳ꂽy[WACfbNXփRs[܂

----------------------------------------------------------------------------- */
bool VoiceManager::Copy( VoicePage pageSrc, int indexSrc, VoicePage pageDst, int indexDst, bool withFile )
{
	nn::os::CriticalSection::ScopedLock lock( m_cs );

	// Rs[邩
	if (!Contains( pageSrc, indexSrc ))
		return false;

	// Rs[悪Ȃ
	if ( Contains( pageDst, indexDst ))
		return false;

	if (pageDst == VOICE_PAGE_TWL_NAND)
	{
		TwlVoiceData* pSrcVoice = GetTwlVoiceData( indexSrc );

		TwlVoiceData voice(
			indexDst,
			pSrcVoice->GetColor(),
			pSrcVoice->GetFill(),
			pSrcVoice->GetForm(),
			nn::fnd::DateTime::GetNow()
		);

		m_twlVoice[ indexDst ] = voice;

		if (withFile)
		{
			// f[^̃Rs[
			VoicePath srcPath;
			VoicePath dstPath;

			if (GetVoicePath( srcPath, pageSrc, indexSrc ))
			{
				if (GetVoicePath( dstPath, pageDst, indexDst ))
				{
					// e|
					VoicePath tmpPath;
					GetVoiceTempPath( tmpPath, pageDst );
					ToReal(tmpPath);

					if (CopyFile( ToReal(srcPath), tmpPath ))
					{
						if (nn::fs::TryRenameFile( tmpPath, ToReal(dstPath) ).IsSuccess())
							return true;
					}

					// sꍇe|t@C͏
					nn::fs::TryDeleteFile( tmpPath );
				}
			}
		}
		else
		{
			return true;
		}

		// Rs[ɎŝŃf[^߂
		m_twlVoice[ indexDst ].SetUse( false );
	}
	else
	{
		VoiceData* pSrcVoice = GetVoiceData( pageSrc, indexSrc );

		// Rs[̃{CX\z
		VoiceData voice(
			pageDst,
			indexDst,
			pSrcVoice->GetColor(),
			pSrcVoice->GetFill(),
			pSrcVoice->GetForm()
		);

		if (SetVoiceData( pageDst, indexDst, voice ))
		{
			if (withFile)
			{
				// f[^̃Rs[
				VoicePath srcPath;
				VoicePath dstPath;

				if (GetVoicePath( srcPath, pageSrc, indexSrc ))
				{
					if (GetVoicePath( dstPath, pageDst, indexDst ))
					{
						// e|
						VoicePath tmpPath;
						GetVoiceTempPath( tmpPath, pageDst );
						ToReal(tmpPath);

						if (CopyFile( ToReal(srcPath), tmpPath ))
						{
							if (nn::fs::TryRenameFile( tmpPath, ToReal(dstPath) ).IsSuccess())
								return true;
						}

						// sꍇe|t@C͏
						nn::fs::TryDeleteFile( tmpPath );
					}
				}
			}
			else
			{
				return true;
			}
		}

		// Rs[ɎŝŖɂ
		m_voice[ pageDst ][ indexDst ].SetUse( false );
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		e|t@C̍폜
	@return		Ȃ
	@par		[]
				e|t@C̍폜s܂

----------------------------------------------------------------------------- */
bool VoiceManager::DeleteTemp( VoicePage page )
{
	VoicePath path;

	// CTR̈ȂTWL
	if (page == VOICE_PAGE_CTR_NAND)
	{
		DeleteTemp( VOICE_PAGE_TWL_NAND );
	}

	if (GetVoiceTempPath( path, page ))
	{
		ToReal( path );

		nn::Result result = nn::fs::TryDeleteFile( path );

		if (result <= nn::fs::ResultNotFound())
		{
			// t@CȂ̂Ő
			return true;
		}
		else if (result.IsFailure())
		{
			result = nn::fs::TryDeleteDirectoryRecursively( path );
			if (result.IsFailure())
			{
				// fBNgƂč폜悤ƂĂs
				return false;
			}
		}

		// 폜
		return true;
	}

	// pX擾łȂ
	return false;
}


/**-----------------------------------------------------------------------------

	@brief		A[JCu܂
	@param[in]	path			pX
	@return		A[JCuׂꍇtrueAׂȂꍇfalse
	@par		[]
				w肳ꂽy[WA[JCuׂł邩Ԃ܂

----------------------------------------------------------------------------- */
bool VoiceManager::LookStrideArchive( VoicePage page1, VoicePage page2 ) const
{
	if (page1 == VOICE_PAGE_TWL_NAND && page2 != VOICE_PAGE_TWL_NAND)
		return true;

	if (page1 != VOICE_PAGE_TWL_NAND && page2 == VOICE_PAGE_TWL_NAND)
		return true;

	if (page1 == VOICE_PAGE_CTR_NAND && page2 != VOICE_PAGE_CTR_NAND)
		return true;

	if (page1 != VOICE_PAGE_CTR_NAND && page2 == VOICE_PAGE_CTR_NAND)
		return true;

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		KV[{CXǂݍ
	@return		Ȃ
	@par		[]
				TWLTEhA[JCüɂ{CXf[^ǂݍ݂܂

----------------------------------------------------------------------------- */
void VoiceManager::ReadTwlNandVoice()
{
	nn::fs::Directory dir;
	nn::Result result = dir.TryInitialize( c_dirTwlNand );

	if (result.IsSuccess())
	{
		nn::fs::DirectoryEntry dirEntry;
		while (0 < dir.Read( &dirEntry, 1 ))
		{
			if (ValidateTwlVoice( dirEntry.entryName ))
			{
				TwlVoiceData voice( dirEntry.entryName );

				VoicePage page = voice.GetPage();

				if (page == VOICE_PAGE_TWL_NAND)
				{
					int index = voice.GetIndex();

					if (Contains( page, index ))
					{
						// ɂꍇ͍폜
						Delete( page, index );
					}

					// o^
					m_twlVoice[ index ] = voice;
				}
			}
		}
	}
}


/**-----------------------------------------------------------------------------

	@brief		NAND̈̃{CXǂݍ
	@return		Ȃ
	@par		[]
				CTRL̈̃{CXf[^ǂݍ݂܂<br/>
				KV[̈Ƀf[^݂Ȃꍇ͐ȂƂāA
				f[^폜܂

----------------------------------------------------------------------------- */
void VoiceManager::ReadCtrNandVoice()
{
	nn::fs::Directory dir;
	nn::Result result = dir.TryInitialize( c_dirCtrNand );

	m_twlConsistency = true;

	if (result.IsSuccess())
	{
		m_validatePage[ VOICE_PAGE_CTR_NAND ] = true;

		nn::fs::DirectoryEntry dirEntry;

		s32 numEntries = 0;

		do
		{
			numEntries = 0;
			result = dir.TryRead( &numEntries, &dirEntry, 1 );

			if (result.IsFailure())
			{
				return;
			}

			if (0 < numEntries)
			{
				EntryVoiceData( &dirEntry, VOICE_PAGE_CTR_NAND );
			}

		} while (0 < numEntries);
	}
	else
	{
		m_validatePage[ VOICE_PAGE_CTR_NAND ] = false;
	}
}


/**-----------------------------------------------------------------------------

	@brief		SD̈̃{CXǂݍ
	@return		Ȃ
	@par		[]
				SD̈ɂ{CXf[^ǂݍ݂܂<br/>
				SDɃ{CXuĂfBNg΁A
				fBNg쐬܂

----------------------------------------------------------------------------- */
Result VoiceManager::ReadSDVoice()
{
	nn::Result result;

	for (size_t i = 0; i < nn::util::GetArraySize(c_filePathSD); i++)
	{
		m_validatePage[ i + VOICE_PAGE_SD_01 ] = false;
	}

	if (IsExistDirectory( c_dirSDRoot ) == false)
	{
		if (nn::fs::IsSdmcInserted() && nn::fs::IsSdmcWritable())
		{
			// fBNg΍쐬ďI
			result = CreateDirectory( c_dirSDRoot );

			if (result.IsFailure())
			{
				// fBNgȂ΁A\Ⴂ̂ōȂ
				return GetFsWholeError( result );
			}

			for (size_t i = 0; i < nn::util::GetArraySize(c_filePathSD); i++)
			{
				VoicePath dirPath;
				std::wcsncpy(dirPath, c_filePathSD[i], nn::fs::MAX_FILE_PATH_LENGTH );

				// zpX猻pX֕ϊ
				ToReal( dirPath );

				result = CreateDirectory( dirPath );

				// ANZXłȂꍇɏ𒆒f
				if (result <= nn::fs::ResultAccessDenied())
				{
					return GetFsWholeError( result );
				}
				else if (result <= nn::fs::ResultOperationDenied())
				{
					return GetFsWholeError( result );
				}
				else if (result <= nn::fs::ResultOutOfResource())
				{
					return GetFsWholeError( result );
				}
				else if (result <= nn::fs::ResultNotEnoughSpace())
				{
					return GetFsWholeError( result );
				}
				else if (result <= nn::fs::ResultWriteProtected())
				{
					return GetFsWholeError( result );
				}

				if (result.IsSuccess())
				{
					m_validatePage[ i + VOICE_PAGE_SD_01 ] = true;
				}
				else
				{
					m_validatePage[ i + VOICE_PAGE_SD_01 ] = false;
				}
			}
		}
		return ResultSuccess;
	}

	for (int i = 0; i < nn::util::GetArraySize(c_filePathSD); i++)
	{
		VoicePath dirPath;
		std::wcsncpy(dirPath, c_filePathSD[i], nn::fs::MAX_FILE_PATH_LENGTH );

		// zpX猻pX֕ϊ
		ToReal( dirPath );

		if (IsExistDirectory( dirPath ))
		{
			m_validatePage[ i + VOICE_PAGE_SD_01 ] = true;

			Result sresult;
			sresult = ReadSDVoicePage( dirPath, static_cast<sndio::VoicePage>(i + VOICE_PAGE_SD_01) );
			if (sresult != ResultSuccess)
			{
				return sresult;
			}
		}
		else
		{
			result = CreateDirectory( dirPath );

			// ANZXłȂꍇɏ𒆒f
			if (result <= nn::fs::ResultAccessDenied())
			{
				return GetFsWholeError( result );
			}
			else if (result <= nn::fs::ResultOperationDenied())
			{
				return GetFsWholeError( result );
			}
			else if (result <= nn::fs::ResultOutOfResource())
			{
				return GetFsWholeError( result );
			}
			else if (result <= nn::fs::ResultNotEnoughSpace())
			{
				return GetFsWholeError( result );
			}
			else if (result <= nn::fs::ResultWriteProtected())
			{
				return GetFsWholeError( result );
			}

			if (result.IsSuccess())
			{
				m_validatePage[ i + VOICE_PAGE_SD_01 ] = true;
			}
			else
			{
				// sȃy[W
				m_validatePage[ i + VOICE_PAGE_SD_01 ] = false;
			}
		}
	}

	return ResultSuccess;
}


/**-----------------------------------------------------------------------------

	@brief		SD̈̃{CXǂݍ
	@param[in]	path			pX
	@param[in]	page			y[W
	@return		Ȃ
	@par		[]
				SD̈ɂ{CXf[^ǂݍ݂܂<br/>
				SDɃ{CXuĂfBNg΁A
				fBNg쐬܂

----------------------------------------------------------------------------- */
Result VoiceManager::ReadSDVoicePage( const wchar_t* path, VoicePage page )
{
	nn::fs::Directory dir;
	nn::Result result = dir.TryInitialize( path );

	if (result.IsSuccess())
	{
		nn::fs::DirectoryEntry dirEntry;

		s32 numEntries = 0;

		do
		{
			numEntries = 0;
			result = dir.TryRead( &numEntries, &dirEntry, 1 );

			if (result.IsFailure())
			{
				return GetFsWholeError( result );
			}

			if (0 < numEntries)
			{
				EntryVoiceData( &dirEntry, page );
			}

		} while (0 < numEntries);

		return ResultSuccess;
	}

	return GetFsWholeError( result );
}


/**-----------------------------------------------------------------------------

	@brief		{CXo^
	@param[in]	entryName		t@CGg
	@param[in	page			o^y[W
	@retval		true 
	@retval		false s
	@par		[]
				{CXf[^̓o^s܂

----------------------------------------------------------------------------- */
bool VoiceManager::EntryVoiceData( nn::fs::DirectoryEntry* entry, VoicePage page )
{
	NN_NULL_TASSERT_( entry );

	const wchar_t* entryName = entry->entryName;

	if (ValidateVoice( entryName ))
	{
		VoiceData voice( entryName, page );

		int index = voice.GetIndex();
		if (Contains( page, index ))
		{
			// ɂꍇ͖
			GetVoiceData( page, index )->SetValidVoice( false );

			// ʂ̃t@CǂݎpAfBNgłꍇ
			// ǂݎpƂĈ
			if (entry->attributes.isReadOnly || entry->attributes.isDirectory)
			{
				GetVoiceData( page, index )->SetReadOnly( true );
			}

			return true;
		}

		if (SetVoiceData( page, index, voice ))
		{
			// ǂݎpAfBNgłꍇɓǂݎpƂĈ
			// st@Cɂ
			if (entry->attributes.isReadOnly || entry->attributes.isDirectory)
			{
				GetVoiceData( page, index )->SetReadOnly( true );
				GetVoiceData( page, index )->SetValidVoice( false );
			}

			// ͈̓`FbN
			if (voice.GetColor() < 0 || VOICE_COLOR_NUM <= voice.GetColor())
			{
				// F
				GetVoiceData( page, index )->SetValidVoice( false );
			}
			else if (voice.GetForm() < VOICE_FORM_NORMAL || VOICE_FORM_NUM <= voice.GetForm())
			{
				// `
				GetVoiceData( page, index )->SetValidVoice( false );
			}
			else if (voice.GetFill() < 0 || VOICE_FILL_100 < voice.GetFill())
			{
				// ܂
				GetVoiceData( page, index )->SetValidVoice( false );
			}

			// t@C̐`FbN
			VoicePath path;
			if (GetVoicePath( path, page, index ))
			{
				nn::fs::FileInputStream fs;
				nn::Result result;
				result = fs.TryInitialize( ToReal(path) );
				if (result.IsFailure())
				{
					GetVoiceData( page, index )->SetValidVoice( false );
				}
				else
				{
					s64 fileSize = 0ll;
					result = fs.TryGetSize( &fileSize );
					if (result.IsFailure())
					{
						GetVoiceData( page, index )->SetValidVoice( false );
					}

					if (VOICE_LIMIT_BYTES <= s32(fileSize))
					{
						GetVoiceData( page, index )->SetValidVoice( false );
					}
				}

				fs.Finalize();
			}

			if (IsTwlSupport() &&
				page == VOICE_PAGE_CTR_NAND &&
				Contains(VOICE_PAGE_TWL_NAND, index) == false)
			{
				// ƂĂȂ̂ŕKv
				m_twlConsistency = false;
			}

			return true;
		}
	}

	return false;
}


/**-----------------------------------------------------------------------------

	@brief		{CXݒ
	@param[in]	page			wy[Wԍ
	@param[in]	index			wCfbNX(0`17)
	@param[in]	pVoice			{CXf[^ւ̃|C^
	@retval		true 
	@retval		false s
	@par		[]
				{CXւ̃f[^|C^Aw肳ꂽy[Wԍ
				CfbNXɐݒ肵܂

----------------------------------------------------------------------------- */
bool VoiceManager::SetVoiceData( VoicePage page, int index, const VoiceData& voice )
{
	NN_TASSERT_( page != VOICE_PAGE_TWL_NAND );

	if (page < 0 || VOICE_PAGE_NUM <= page)
		return false;

	if (index < 0 || VOICE_NUM_IN_PAGE <= index)
		return false;

	if (m_voice[ page ][ index ].IsUse())
		return false;

	m_voice[ page ][ index ] = voice;
	return true;
}


/**-----------------------------------------------------------------------------

	@brief		wpX
	@param[out]	path			pX
	@return		Ȃ
	@par		[]
				w肩pX𐶐ApathɌʂԂ܂

----------------------------------------------------------------------------- */
void VoiceManager::Format( VoicePath& path, const wchar_t* fmt, ... ) const
{
	std::va_list vlist;

	va_start(vlist, fmt);

	(void)std::vswprintf( path, nn::fs::MAX_FILE_PATH_LENGTH, fmt, vlist );

	va_end(vlist);
}


}} // namespace sndio::detail
