/**-----------------------------------------------------------------------------

	@file		sndio_VoiceManager.h
	@author		tetsu
	@par		[]
				^{CX̃f[^Ǘ`(J)

				Copyright (c)2010 Nintendo All rights reserved.

----------------------------------------------------------------------------- */

#ifndef SNDIO_VOICE_MANAGER_H
#define SNDIO_VOICE_MANAGER_H

#include <wchar.h>
#include <nn/types.h>
#include <nn/fs.h>
#include "sndio_Result.h"
#include "sndio_VoiceTypes.h"


namespace sndio { namespace detail {


////////////////////////////////////////////////////////////////////////////////
// f[^NX
////////////////////////////////////////////////////////////////////////////////


//! ʏ{CX
class VoiceData
{
public:
	VoiceData();

	//! f[^ŏ
	VoiceData( VoicePage page, int index, int color, int fill, int form );

	//! Ggŏ
	VoiceData( const wchar_t* entryName, VoicePage page );
	
	//! gpĂ邩
	bool IsUse() const;

	//! Lȉ
	bool IsValidVoice() const;

	//! ǂݎp
	bool IsReadOnly() const;

	//! ۑĂy[W
	VoicePage GetPage() const;

	//! y[WɕۑĂʒu (0`17)
	int GetIndex() const;

	//! J[
	int GetColor() const;

	//! ܂
	int GetFill() const;

	//! ӂ
	int GetForm() const;

	//! gp邩
	void SetUse( bool use );

	//! Lȉ
	void SetValidVoice( bool valid );

	//! ǂݎp
	void SetReadOnly( bool readOnly );

	//! y[Wݒ
	void SetPage( VoicePage page );

	//! CfbNXݒ(0`17)
	void SetIndex( int index );

	//! y[WƃCfbNX̓ւ
	void Swap( VoiceData* pFrom );

protected:

	s8					m_use;				//!< gpĂ邩
	s8					m_valid;			//!< Lȉ
	s8					m_readOnly;			//!< ǂݎp
	s8					m_page;				//!< ۑy[W
	s8					m_index;			//!< ۑCfbNX
	s8					m_color;			//!< J[
	s8					m_fill;				//!< tL_VACR̖܂
	s8					m_form;				//!< tL_VACŘ`
};


//! KV[{CX(DSiTEh)
class TwlVoiceData : public VoiceData
{
public:

	TwlVoiceData() : VoiceData()
	{
	}

	//! f[^ŏ
	TwlVoiceData( int index, int color, int fill, int form,
		const nn::fnd::DateTime& dateTime );

	//! Ggŏ
	explicit TwlVoiceData( const wchar_t* entryName );

	//! t
	const nn::fnd::DateTime& GetDateTime() const;

	//! Rs[
	int GetCopy() const;

	//! o[W
	int GetVersion() const;

protected:
	nn::fnd::DateTime	m_dateTime;			//!< XV
	u8					m_copy;				//!< Rs[
	u8					m_version;			//!< o[W

	NN_PADDING2;
	NN_PADDING4;
};


////////////////////////////////////////////////////////////////////////////////
// ǗNX
////////////////////////////////////////////////////////////////////////////////

class VoiceManager
{
public:

	//! t@CpX
	typedef wchar_t VoicePath[nn::fs::MAX_FILE_PATH_LENGTH + 1];

	//! {CXRei
	typedef VoiceData VoiceContainer[VOICE_NUM_IN_PAGE];

	//! {CXRei
	typedef TwlVoiceData TwlVoiceContainer[VOICE_NUM_IN_PAGE];

	//! CTRL̈pX擾
	static const wchar_t* GetNandPath();

	//! {CXpX擾
	static const wchar_t* GetTwlPath();

	//! {CXpX擾(zpX)
	static const wchar_t* GetDirectoryPath( VoicePage page );

	//! SDe|pX擾(zpX)
	static const wchar_t* GetDirectoryTempPath();

	//! {CXÓ`FbN
	static bool ValidateVoice( const wchar_t* entryName );

	//! {CX̑Ó`FbN
	static bool ValidateTwlVoice( const wchar_t* entryName );

public:

	VoiceManager();
	~VoiceManager();

	//! 
	Result Initialize( bool useSD = true, bool useCtrNand = true, bool useTwlNand = false );

	//! I
	void Finalize();

	//! sĂ邩
	bool IsInitialized() const;

	//! {CXT|[g邩
	bool IsTwlSupport() const;

	//! TWL̈ƐƂĂȂ̂ŁAKv
	bool IsTwlNeedRestore() const;

	//! {CXł݂邩
	bool IsExistsVoice() const;

	//! w̃y[W
	bool IsInvalidPage( VoicePage page );

	//! 
	bool IsBusy() const;

	//! č\z
	Result Restore();

	//! SD؂藣
	void PurgeSD();

	//! SDXgA
	Result RestoreSD();

	//! {CXf[^̎擾( index = 0 ` VOICE_NUM_IN_PAGE-1 )
	//{@
	VoiceData* GetVoiceData( VoicePage page, int index );
	const VoiceData* GetVoiceData( VoicePage page, int index ) const;
	//}@

	//! {CXf[^擾( index = 0 ` VOICE_NUM_IN_PAGE-1 )
	//{@
	TwlVoiceData* GetTwlVoiceData( int index );
	const TwlVoiceData* GetTwlVoiceData( int index ) const;
	//}@

	//! {CXǉ
	bool AddVoice( VoicePage page, int index, int color, int fill, int form );

	//! {CXǉ
	bool AddTwlVoice( int index, int color, int fill, int form,
		const nn::fnd::DateTime& dateTime );

	//! {CX̃pX擾( index = 0 ` VOICE_NUM_IN_PAGE-1 )
	bool GetVoicePath( VoicePath& path, VoicePage page, int index ) const;

	//! {CX̃e|pt@CpX擾
	bool GetVoiceTempPath( VoicePath& path, VoicePage page ) const;

	//! {CX݂邩ǂ擾( index = 0 ` VOICE_NUM_IN_PAGE-1 )
	bool Contains( VoicePage page, int index ) const;

	//! 폜
	bool Delete( VoicePage page, int index, bool withFile = true );

	//! ւ
	bool Swap( VoicePage page1, int index1, VoicePage page2, int index2, bool withFile = true );

	//! Rs[
	bool Copy( VoicePage pageSrc, int indexSrc, VoicePage pageDst, int indexDst, bool withFile = true );

	//! e|t@C̍폜
	bool DeleteTemp( VoicePage page );
	
	//! A[JCu܂
	bool LookStrideArchive( VoicePage page1, VoicePage page2 ) const;

private:

	//! TWL̈ǂݍ
	void ReadTwlNandVoice();

	//! NAND̈ǂݍ
	void ReadCtrNandVoice();

	//! SD̈ǂݍ
	Result ReadSDVoice();

	//! SD̈̃tH_Pʓǂݍ
	Result ReadSDVoicePage( const wchar_t* path, VoicePage page );

	//! {CXf[^ݒ
	bool SetVoiceData( VoicePage page, int index, const VoiceData& pVoice );

	//! {CXf[^o^
	bool EntryVoiceData( nn::fs::DirectoryEntry* entry, VoicePage page );

	//! pX
	void Format( VoicePath& path, const wchar_t* fmt, ... ) const;

private:

	//! {CXf[^Rei
	VoiceContainer			m_voice[ VOICE_PAGE_NUM ];

	//! {CXf[^Rei
	TwlVoiceContainer		m_twlVoice;

	//! ς݂
	bool					m_initialized;

	//! NANDgp
	bool					m_useCtrNand;

	//! SDgp
	bool					m_useSD;

	//! {CXT|[g邩
	bool					m_twlSupport;

	//! {CXƂ̐ƂĂ邩
	bool					m_twlConsistency;

	//! Lȃy[Wǂ
	bool					m_validatePage[ VOICE_PAGE_NUM ];

	NN_PADDING4;

	//! rی
	mutable nn::os::CriticalSection	m_cs;
};


}}


#endif // SNDIO_VOICE_MANAGER_H
