/*---------------------------------------------------------------------------*
  Project:  sndio
  File:     sndio_Api.cpp

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 21647 $
 *---------------------------------------------------------------------------*/

#include <ctype.h>
#include <nn/assert.h>
#include <nn/aacenc.h>
#include <nn/cfg.h>
#include <nn/dsp.h>
#include <nn/os/os_CriticalSection.h>

#include "sndio_Api.h"
#include "detail/sndio_Mount.h"
#include "detail/sndio_VoiceManager.h"
#include "detail/sndio_Util.h"

namespace sndio {


using detail::VoiceManager;
using detail::VoiceData;


namespace
{

//! {CXǗ
VoiceManager    s_voiceManager;

//! ς݂
bool            s_IsInitialized = false;

//! }Eg̎
bool            s_useSD = true;
bool            s_useCtrNand = false;
bool            s_useTwlNand = false;


/*!
    @brief      Xe[^XXV

    @return     ɊȂ true 
                łȂ false Ԃ܂B
*/
Result CheckSDError( nn::Result result )
{
    if (result.IsFailure())
    {
        if (result <= nn::fs::ResultNotFound())
        {
            if (result <= nn::fs::ResultMediaNotFound())
            {
                // fBAȂA͔Fł܂B
                // ResultNotFound ɑG[łB

                // fBAs
                if (nn::fs::IsSdmcInserted())
                    return ResultMediaInvalid;

                // fBA
                return ResultMediaNotFound;
            }

            return ResultMediaNotFound;
        }
        else if (result <= nn::fs::ResultNotFormatted())
        {
            return ResultNotFormatted;
        }
        else if (result <= nn::fs::ResultBadFormat())
        {
            return ResultBadFormat;
        }
        else if (result <= nn::fs::ResultArchiveInvalidated())
        {
            // }EgsȂĂŒ SD J[hꂽ\܂B
            // nn::fs::ResultMediaNotFound() ƓłB
            return ResultArchiveInvalidated;
        }
        else if (result <= nn::fs::ResultOperationDenied())
        {
            if(result <= nn::fs::ResultWriteProtected())
            {
                // SD J[h݋֎~ɂĂ܂B
            }
            else if(result <= nn::fs::ResultMediaAccessError())
            {
                // }Ă̂ SD J[hł͂ȂASD J[hĂ܂B
                return ResultMediaAccessError;
            }
            else
            {
                // ȊO̗RŃ}Egۂ邱Ƃ́A܂܂B
                // AꍇɃJ[h̑}{̂̍ċNŕA\܂B
                // FATALH
                return ResultOperationDenied;
            }
        }
        else
        {
            // LȊÕG[͔܂B
            // FATALH
            return ResultError;
        }
    }

    return ResultSuccess;
}


}


/*!
    @brief      Ƀ}EgstOݒ肵܂B

                ̊֐͔J֐łB

    @param[in]  useSD       SDJ[ḧgp邩
    @param[in]  useCtrNand  {̕ۑ̈gp邩
    @param[in]  useTwlNand  TWLA[JCügp邩
*/
void SetupMountArchive( bool useSD, bool useCtrNand, bool useTwlNand )
{
    s_useSD = useSD;
    s_useCtrNand = useCtrNand;
    s_useTwlNand = useTwlNand;
}


/*!
    @brief      ܂B

                }Eg͓ōs܂B<br/>

    @retval     ResultSuccess               
    @retval     ResultMediaNotFound         fBAĂȂ
    @retval     ResultMediaInvalid          fBAs
    @retval     ResultNotFormatted          tH[}bg
    @retval     ResultBadFormat             stH[}bg
    @retval     ResultArchiveInvalidated    A[JCus
    @retval     ResultMediaAccessError      ANZXG[
    @retval     ResultOperationDenied       ANZX
    @retval     ResultError                 sȃG[
*/
Result Initialize()
{
    if (s_IsInitialized)
    {
        return ResultNotFinalize;
    }
    s_IsInitialized = true;

    // SDւ̃}Eg
    if (s_useSD)
    {
        nn::Result result = detail::MountSD();

        if (result.IsFailure())
        {
            s_useSD = false;
            return CheckSDError( result );
        }
    }

    // CTR NANDւ̃}Eg
    if (s_useCtrNand)
    {
        nn::Result result = detail::MountCtrNand();
        if (result.IsFailure())
        {
            s_useCtrNand = false;
        }
    }

    // TWL NANDւ̃}Eg
    if (s_useTwlNand)
    {
        nn::Result result = detail::MountTwlNand();
        if (result.IsFailure())
        {
            s_useTwlNand = false;
        }
    }

    return s_voiceManager.Initialize( s_useSD, s_useCtrNand, s_useTwlNand );
}


/*!
    @brief      SDJ[hɂ鉹̍č\z܂B

                }Egɍs܂B

    @retval     ResultSuccess               
    @retval     ResultMediaNotFound         fBAĂȂ
    @retval     ResultMediaInvalid          fBAs
    @retval     ResultNotFormatted          tH[}bg
    @retval     ResultBadFormat             stH[}bg
    @retval     ResultArchiveInvalidated    A[JCus
    @retval     ResultMediaAccessError      ANZXG[
    @retval     ResultOperationDenied       ANZX
    @retval     ResultError                 sȃG[
*/
Result Restore()
{
    if (IsInitialized())
    {
        nn::Result result;

        // SD̂ݍďs
        if (s_useSD)
        {
            s_voiceManager.PurgeSD();
            (void)detail::UnmountSD();

            result = detail::MountSD();
            s_voiceManager.RestoreSD();
            if (result.IsFailure())
            {
                return CheckSDError( result );
            }
        }

        return ResultSuccess;
    }

    return ResultNotInitialize;
}


/*!
    @brief      j܂B
*/
void Finalize()
{
    if (IsInitialized())
    {
        s_voiceManager.Finalize();

        if (s_useSD)
        {
            detail::UnmountSD();
        }

        if (s_useCtrNand)
        {
            detail::UnmountCtrNand();
        }

        if (s_useTwlNand)
        {
            detail::UnmountTwlNand();
        }

        s_useSD = false;
        s_useCtrNand = false;
        s_useTwlNand = false;
        s_IsInitialized = false;
    }
}


/*!
    @brief      ۑNXĂ邩ǂԂ܂B

    @return     ۑNX̏Ă true A
                łȂ false Ԃ܂B
*/
bool IsInitialized()
{
    return s_IsInitialized;
}


/*!
    @brief      y[WsȏԂǂԂ܂

    @retval     true    y[Ws(fBNg)Ԃł
    @retval     false   y[Wł
*/
bool IsInvalidPage( VoicePage page )
{
    return s_voiceManager.IsInvalidPage( page );
}


/*!
    @brief      w肳ꂽʒuɉ݂邩Ԃ܂B

                CfbNXɂ̓y[Wɑ݂鉹CfbNXԍw肵܂B<br/>
                wCfbNXɎwł͈͂[0-17]ɂȂ܂B<br/>

    @param[in]  page        wy[Wԍ
    @param[in]  index       wCfbNX([0-17])

    @return     trueȂ特݂܂BfalseȂ特݂܂B
*/
bool IsExistsEntry( VoicePage page, s32 index )
{
    NN_TASSERT_( IsInitialized() );

    return s_voiceManager.Contains( page, index );
}


/*!
    @brief      o^Ă鉹̃Gg擾܂B

                wCfbNXɎwł͈͂[0-17]ɂȂ܂B<br/>
                wy[WAwCfbNXɃf[^ꍇfalseԂ܂<br/>

    @param[out] pOut        Gg
    @param[in]  page        wy[W
    @param[in]  index       wCfbNX

    @return     trueœǂݍ݂BfalseœǂݍݎsB
*/
bool ReadEntry( VoiceEntry* pOut, VoicePage page, s32 index )
{
    NN_TASSERT_( IsInitialized() );
    NN_NULL_TASSERT_( pOut );

    VoiceManager::VoicePath voicePath;

    if (s_voiceManager.GetVoicePath( voicePath, page, index ))
    {
        std::memset( pOut->entryName, 0, sizeof(pOut->entryName) );
        std::memcpy( pOut->entryName, voicePath, sizeof(pOut->entryName) );

        if ( VoiceData* pVoiceData = s_voiceManager.GetVoiceData( page, index ) )
        {
            pOut->color   = static_cast<VoiceColor>(pVoiceData->GetColor());
            pOut->fill    = static_cast<VoiceFill>(pVoiceData->GetFill());
            pOut->form    = static_cast<VoiceForm>(pVoiceData->GetForm());
            pOut->invalid = !(pVoiceData->IsValidVoice());
            pOut->readOnly= pVoiceData->IsReadOnly();
        }
        else
        {
            return false;
        }

        return true;
    }

    return false;
}


/*!
    @brief      ۑΏۂ̉̃Gg쐬܂

                󂢂Ă鉹ӏɑ΂ĉۑ邽߂̃pX쐬AGgƂĕԂ܂B<br/>
                wCfbNXɎwł͈͂[0-17]ɂȂ܂B<br/>
                wy[WAwCfbNXɊɉ݂ꍇ̓Gg쐬ɎsfalseԂ܂B<br/>

    @param[out] pOut        Gg
    @param[in]  page        wy[W
    @param[in]  index       wCfbNX([0-17])
    @param[in]  color       쐬J[
    @param[in]  fill        쐬邤܂
    @param[in]  form        쐬`

    @return     trueŃGg쐬BfalseŃGg쐬sB
*/
bool CreateEntry( VoiceEntry* pOut, VoicePage page, s32 index, VoiceColor color, VoiceFill fill, VoiceForm form )
{
    NN_TASSERT_( IsInitialized() );

    // łɓo^ĂΎs
    if (IsExistsEntry(page, index))
    {
        return false;
    }

    // ̂ӂȂ玸s(̂ӂ͍Ȃ)
    if (form == VOICE_FORM_EMPTY)
    {
        return false;
    }

    if (s_voiceManager.AddVoice( page, index, color, fill, form ))
    {
        VoiceManager::VoicePath voicePath;

        if (s_voiceManager.GetVoicePath( voicePath, page, index ))
        {
            std::memset( pOut->entryName, 0, sizeof(pOut->entryName) );
            std::memcpy( pOut->entryName, voicePath, sizeof(pOut->entryName) );
            pOut->color = color;
            pOut->fill = fill;
            pOut->form = form;
            return true;
        }
    }

    return true;
}


/*!
    @brief      w肳ꂽGg폜܂

                wCfbNXɎwł͈͂[0-17]ɂȂ܂B<br/>
                wy[WAwCfbNXɃf[^ꍇfalseԂ܂<br/>

    @param[in]  page        wy[W
    @param[in]  index       wCfbNX([0-17])

    @return     trueŃGg폜BfalseŃGg폜sB
*/
bool DeleteEntry( VoicePage page, s32 index )
{
    // o^ĂȂΎs
    if (!IsExistsEntry(page, index))
    {
        return false;
    }

    if (s_voiceManager.Delete( page, index ))
    {
        return true;
    }

    return false;
}


/*!
    @brief      w肳ꂽGg̉f[^ǂݍ݂܂

                wy[WSDJ[h̃y[Ŵݎwł܂B<br/>
                wCfbNXɎwł͈͂[0-17]ɂȂ܂B<br/>
                ndatãobt@VOICE_LIMIT_BYTESŒŊmۂ̂nĂ

    @param[out] readCount   ǂݎTCY
    @param[in]  page        wy[W
    @param[in]  index       wCfbNX([0-17])
    @param[in]  data        ǂݎf[^(m4ãf[^)
    @param[in]  dataSize    VOICE_LIMIT_BYTESŒ

    @retval     ResultSuccess               ݐ
    @retval     ResultMediaAccessError      fBAANZXG[
    @retval     ResultError                 ̑̕sȃG[
*/
Result ReadVoice( s32* readCount, VoicePage page, s32 index, void* data, size_t dataSize )
{
    if (dataSize != VOICE_LIMIT_BYTES)
    {
        return ResultError;
    }

    VoiceEntry entry;
    nn::Result result;

    while (ReadEntry( &entry, page, index ))
    {
        nn::fs::FileInputStream input;

        detail::ToReal( entry.entryName );

        result = input.TryInitialize( entry.entryName );
        if (result.IsFailure())
        {
            break;
        }

        s64 fileSize = 0ll;
        result = input.TryGetSize( &fileSize );
        if (result.IsFailure())
        {
            break;
        }

        dataSize = static_cast<size_t>( fileSize );

        result = input.TryRead( readCount, data, dataSize );
        if (result.IsFailure())
        {
            break;
        }

        return ResultSuccess;
    }

    if (result <= nn::fs::ResultOperationDenied())
    {
        if (result <= nn::fs::ResultMediaAccessError())
        {
            // fBAւ̃ANZXɃG[܂B
            // ResultOperationDenied ɑG[łB
            return ResultMediaAccessError;
        }

        // 삪ۂ܂B
        return ResultError;
    }

    return ResultError;
}


/*!
    @brief      w肳ꂽGg։f[^݂܂

                w肳ꂽy[WACfbNXɁAdataɓnꂽm4af[^݂܂<br/>
                wy[WSDJ[h̃y[Ŵݎwł܂B<br/>
                wCfbNXɎwł͈͂[0-17]ɂȂ܂B<br/>

    @param[in]  page        wy[W
    @param[in]  index       wCfbNX([0-17])
    @param[in]  color       쐬J[
    @param[in]  fill        쐬邤܂
    @param[in]  form        쐬`
    @param[in]  data        쐬f[^(m4ãf[^)
    @param[in]  dataSize    쐬f[^TCY

    @retval     ResultSuccess               ݐ
    @retval     ResultNotEnoughSpace        eʕsG[
    @retval     ResultWriteProtected        ݋֎~G[
    @retval     ResultMediaAccessError      fBAANZXG[
    @retval     ResultError                 ̑̕sȃG[
*/
Result WriteVoice(
    VoicePage page, s32 index,
    VoiceColor color, VoiceFill fill, VoiceForm form,
    const void* data, size_t dataSize )
{
    NN_TASSERT_( page != VOICE_PAGE_CTR_NAND );
    NN_TASSERT_( page != VOICE_PAGE_TWL_NAND );

    VoiceEntry entry;
    nn::Result result;

    while (CreateEntry( &entry, page, index, color, fill, form ))
    {
        nn::fs::FileOutputStream output;

        detail::ToReal( entry.entryName );

        result = nn::fs::TryCreateFile( entry.entryName, s64(dataSize) );
        if (result.IsFailure())
        {
            break;
        }

        result = output.TryInitialize( entry.entryName, false );
        if (result.IsFailure())
        {
            break;
        }

        s32 writeCount = 0;
        result = output.TryWrite( &writeCount, data, dataSize, true );
        if (result.IsFailure() || writeCount != dataSize)
        {
            break;
        }

        return ResultSuccess;
    }

    DeleteEntry( page, index );

    if (result <= nn::fs::ResultNotEnoughSpace())
    {
        // 󂫗eʂ܂B
        return ResultNotEnoughSpace;
    }
    else if (result <= nn::fs::ResultOperationDenied())
    {
        if (result <= nn::fs::ResultWriteProtected())
        {
            // ݂֎~ɂĂ܂B
            // ResultOperationDenied ɑG[łB
            return ResultWriteProtected;
        }
        else if (result <= nn::fs::ResultMediaAccessError())
        {
            // fBAւ̃ANZXɃG[܂B
            // ResultOperationDenied ɑG[łB
            return ResultMediaAccessError;
        }

        // 삪ۂ܂B
        return ResultError;
    }

    // sȃG[
    return ResultError;
}


/*!
    @brief      {ID擾

                {ID̉32Bit16i8(I[NULL܂܂Ȃ)ƂĎ擾܂B

    @param[out] 擾ꂽ{ID
*/
void GetBodyId( u8 bodyId[8] )
{
    /* sndioŗL̖{ID擾(32bit͔j) */
    u32 value = static_cast<u32>( nn::cfg::CTR::GetTransferableId( 0x00205 ) );

    for (int i = 7; 0 <= i; i--)
    {
        u32 hex = u32( value & 0xf );

        if (0x0 <= hex && hex < 0xA)
        {
            bodyId[i] = static_cast<u8>( '0' + hex );
        }
        else if (0xA <= hex && hex <= 0xF)
        {
            bodyId[i] = static_cast<u8>( 'A' + hex - 0xA );
        }
        else
        {
            NN_TASSERT_( false );
        }

        value >>= 4;
    }
}


} // namespace sndio
