/*---------------------------------------------------------------------------*
  Project:  sndio
  File:     sndio_SoundReader.h

  Copyright (C)2010 Nintendo Co., Ltd. All rights reserved

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 19445 $
 *---------------------------------------------------------------------------*/

#ifndef SNDIO_SOUND_READER_H
#define SNDIO_SOUND_READER_H

#include <wchar.h>
#include <nn/types.h>
#include <nn/fs.h>
#include <nn/dbg.h>
#include <nn/util/util_NonCopyable.h>
#include "detail/sndio_Result.h"


namespace sndio {


//! TEhtH[}bg
enum SoundFormat
{
    SOUND_FORMAT_NONE,
    SOUND_FORMAT_AAC,           //!< AAC(MP4, M4A, 3GP)
    SOUND_FORMAT_MP3,           //!< MP3
    SOUND_FORMAT_DAT,           //!< DAT(TWLSNDtH[}bg)
    SOUND_FORMAT_PCM            //!< PCM
};


/*!
    @brief      yf[^NXłB
*/
class SoundReader : private nn::util::NonCopyable<SoundReader>
{
public:

    //! AAĈPt[Tv
    static const u32 AAC_FRAME_SAMPLES = 1024;

    //! MP3̂Pt[Tv
    static const u32 MP3_FRAME_SAMPLES = 1152;

    /*!
        @brief      \z

                    RXgN^łB<br/>
    */
    SoundReader();

    /*!
        @brief      

                    fXgN^łB<br/>
    */
    ~SoundReader();

    /*!
        @brief      s܂B

                    AP[^̓foCX̊mۂłKv܂B<br/>
                    ADTS-AAC(.aac)t@C͎wł܂B<br/>

        @param[in]  allocator   AP[^(foCX)
        @param[in]  filePath    t@CpX(M4A,MP4,3GP,MP3̃t@C)

        @retval     ResultSuccess               
        @retval     ResultError                 sȃG[
        @retval     ResultFileNotFound          t@C݂Ȃ
        @retval     ResultMediaNotFound         fBAĂȂ
        @retval     ResultMediaInvalid          fBAs
        @retval     ResultAlreadyExists         t@Cɑ
        @retval     ResultNotEnoughSpace        eʕsG[
        @retval     ResultBadFormat             stH[}bg
        @retval     ResultArchiveInvalidated    A[JCus
        @retval     ResultOperationDenied       ANZX
        @retval     ResultWriteProtected        ݋֎~G[
        @retval     ResultMediaAccessError      fBAANZXG[
        @retval     ResultNotFormatted          tH[}bg
        @retval     ResultOutOfResource         \[Xs
        @retval     ResultNotFinalize           JĂȂ
        @retval     ResultInvalidFile           T|[g̃t@C
        @retval     ResultInvalidDecoder        fR[_[G[
    */
    Result Initialize( nn::fnd::IAllocator& allocator, const wchar_t* filePath );

    /*!
        @brief      Xg[珉s܂B

                    AP[^̓foCX̊mۂłKv܂B<br/>
                    ADTS-AAC(.aac)t@C͎wł܂B<br/>

        @param[in]  allocator   AP[^(foCX)
        @param[in]  stream      Xg[
        @param[in]  format      TEhtH[}bg

        @retval     ResultSuccess               
        @retval     ResultError                 sȃG[
        @retval     ResultFileNotFound          t@C݂Ȃ
        @retval     ResultMediaNotFound         fBAĂȂ
        @retval     ResultMediaInvalid          fBAs
        @retval     ResultAlreadyExists         t@Cɑ
        @retval     ResultNotEnoughSpace        eʕsG[
        @retval     ResultBadFormat             stH[}bg
        @retval     ResultArchiveInvalidated    A[JCus
        @retval     ResultOperationDenied       ANZX
        @retval     ResultWriteProtected        ݋֎~G[
        @retval     ResultMediaAccessError      fBAANZXG[
        @retval     ResultNotFormatted          tH[}bg
        @retval     ResultOutOfResource         \[Xs
        @retval     ResultNotFinalize           JĂȂ
        @retval     ResultInvalidFile           T|[g̃t@C
        @retval     ResultInvalidParameter      p[^s
        @retval     ResultInvalidDecoder        fR[_[G[
    */
    Result Initialize(
        nn::fnd::IAllocator& allocator,
        nn::fs::IInputStream& stream,
        SoundFormat format );

    /*!
        @brief      j܂B

                    fXgN^̒łĂ΂܂B<br/>
                    IɌĂł肠܂B<br/>
    */
    void Finalize();

    /*!
        @brief      ^Cg擾܂B
        @return     ^Cgւ̃|C^
    */
    const wchar_t* GetTitleName() const;

    /*!
        @brief      Ao擾܂
        @return     Aoւ̃|C^
    */
    const wchar_t* GetAlbumName() const;

    /*!
        @brief      A[eBXg擾܂
        @return     A[eBXgւ̃|C^
    */
    const wchar_t* GetArtistName() const;

    /*!
        @brief      g^O񂪗L`FbN܂
        @return     Lȏꍇtrue
    */
    bool IsExistExtendTag() const;

    /*!
        @brief      ۑ\tgEFAID擾܂

        @param[out] \tgEFAID擾
        @return     擾łꍇtrue
    */
    bool GetSFTW( u8 sftw[5] ) const;

    /*!
        @brief      ^{ID擾܂

        @param[out] ^{ID擾
        @return     擾łꍇtrue
    */
    bool GetRCID( u8 rcid[8] ) const;

    /*!
        @brief      CTRAvŕҏW{ID擾܂

        @param[out] CTRAvŕҏW{ID擾
        @return     擾łꍇtrue
    */
    bool GetEDID( u8 edid[8] ) const;

    /*!
        @brief      Abv[hĂ邩擾܂
        @return     0: 1:C^[lbgAbv[h
    */
    u8 GetPermitUpload() const;

    /*!
        @brief      ό̉zzĂ邩擾܂
        @return     0: 1:zz
    */
    u8 GetPermitEditDistribute() const;

    /*!
        @brief      \擾܂
        @return     0:s 1:
    */
    u8 GetPermitFirstDistribute() const;

    /*!
        @brief      񎟔zz\擾܂
        @return     0:񎟔zzs 1:񎟔zz
    */
    u8 GetPermitSecondDistribute() const;

    /*!
        @brief      TEhtH[}bg擾܂B
        @return     TEhtH[}bg
    */
    SoundFormat GetSoundFormat() const;

    /*!
        @brief      Pt[̃Tv擾܂B
        @return     Tv
    */
    u32 GetFrameSamples() const;

    /*!
        @brief      Pt[̃obt@TCY(Byte)擾܂B
        @return     obt@TCY(Byte)
    */
    u32 GetFrameBufferSize() const;

    /*!
        @brief      TvO[g擾܂B
        @return     TvO[g
    */
    s32 GetSamplingRate() const;

    /*!
        @brief      `l擾܂B
        @return     `l
    */
    s32 GetChannelCount() const;

    /*!
        @brief      ĐԂ擾܂(msecP)
        @return     Đ(msecP)
    */
    s32 GetTotalTime() const;

    /*!
        @brief      V[NԂ擾܂(msecP)
        @return     V[N(msecP)
    */
    s32 GetSeekTime() const;

    /*!
        @brief      V[N\
        @return     true:V[N\, false:V[Ns
    */
    bool CanSeek() const;

    /*!
        @brief      (msecP)ŃV[Nʒuݒ肵܂
        @param[in]  V[N(msecP)
    */
    void SetSeekTime( s32 seekTime );

    /*!
        @brief      tĐs(AAC-DSP̂)
        @param[in]  true tĐw, false Đw(ftHg)
    */
    void SetReverse( bool reverse );

    /*!
        @brief      TEh̔g`f[^(PCM16)擾܂B

                    fR[hŝŎsɂ͎Ԃ܂B<br/>
                    󂯎obt@32oCgACg̃foCX
                    Kv܂B<br/>
                    m̏ꍇ́ApBufR𖳉(0)ɐݒ肵܂B<br/>
                    AACłΎ擾Tv𒴂Ȃ1024̔{Ŏ擾܂B<br/>
                    MP3łΎ擾Tv𒴂Ȃ576,1152̔{Ŏ擾܂B

        @param[out] readSamples ǂ݂ƂTv
        @param[out] pBufL       `l󂯎obt@
        @param[out] pBufR       E`l󂯎obt@
        @param[in]  samples     擾Tv

        @retval     ResultSuccess               ǂݎ萬
        @retval     ResultError                 sȃG[
        @retval     ResultMediaNotFound         fBAĂȂ
        @retval     ResultMediaInvalid          fBAs
        @retval     ResultAlreadyExists         t@Cɑ
        @retval     ResultNotEnoughSpace        eʕsG[
        @retval     ResultBadFormat             stH[}bg
        @retval     ResultArchiveInvalidated    A[JCus
        @retval     ResultOperationDenied       ANZX
        @retval     ResultWriteProtected        ݋֎~G[
        @retval     ResultMediaAccessError      fBAANZXG[
        @retval     ResultNotFormatted          tH[}bg
        @retval     ResultOutOfResource         \[Xs
        @retval     ResultInvalidFile           T|[g̃t@C
        @retval     ResultInvalidParameter      p[^s
        @retval     ResultInvalidDecoder        fR[_[G[
    */
    Result TryRead( s32* readSamples, s16* pBufL, s16* pBufR, s32 samples );

private:

    void*                   m_kernel;
    nn::fs::IInputStream*   m_stream;
    wchar_t                 m_filePath[ nn::fs::MAX_FILE_PATH_LENGTH + 1 ];

};


} // namespace sndio


#endif /* SNDIO_SOUND_READER_H */
