/*---------------------------------------------------------------------------*
  Project:  sndio
  File:     sndio_VoiceWriter.cpp

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 21647 $
 *---------------------------------------------------------------------------*/

#include "sndio_VoiceWriter.h"
#include "detail/sndio_Mount.h"
#include "detail/sndio_VoiceExtendTag.h"
#include "detail/sndio_VoiceManager.h"
#include "detail/sndio_CachedWriter.h"
#include "detail/sndio_Util.h"

#include <nn/assert.h>
#include <nn/dsp.h>
#include <nn/os/os_CriticalSection.h>
#include <nn/aacenc.h>
#include <tmc_m4amux.h>


namespace
{
    //! fR[_[̏sĂ邩
    static bool s_decoderInitialize = false;

    //! ^OR[hŒl
    static const char c_tagCode[] = "3DS1";

    //! g^O̐
    static const u32 TAG_NUM = 5;

    /*!
        @brief      UTF8̕obt@ɏo܂B
                    dstNULLw肵ꍇɂ́ATCŶݕԂ܂B

        @param[out] dst         o͐̃obt@
        @param[in]  dst_len     o͐̃obt@TCY
        @param[in]  code        UTF32̕R[h
        @return                 񂾃oCgԂ܂B
                                ݐ̃obt@TCYȂꍇɂ́A
                                ̃G[R[hԂ܂B
    */
    static s32 WriteUTF8( u8* dst, s32 dst_len, u32 code )
    {
        s32 len, i;

        if (code < 0x00000080)      { len = 1; }
        else if (code < 0x00000800) { len = 2; }
        else if (code < 0x00010000) { len = 3; }
        else if (code < 0x02000000) { len = 4; }
        else { return -2; }

        if (!dst) { return len; }
        if (dst_len <= 0) { return -1; }

        switch (len)
        {
        case 1: { dst[ 0 ] = (u8)code;                  } break;
        case 2: { dst[ 0 ] = (u8)((code >> 6) + 0xC0);  } break;
        case 3: { dst[ 0 ] = (u8)((code >> 12) + 0xE0); } break;
        case 4: { dst[ 0 ] = (u8)((code >> 18) + 0xF0); } break;
        }

        if (dst_len < len)
        {
            return -1;
        }

        for (i = 1; i < len; ++i)
        {
            dst[ i ] = (u8)((code >> (6 * (len - i - 1)) & 0x3F) + 0x80);
        }

        return len;
    }

    /*!
        @brief      obt@UTF16LE̕ǂݍ݂܂B

        @param[in]  src         ͌̃obt@
        @param[in]  dst_len     ͐̎c̒
        @param[out] utf32_code  UTF32̕R[h
        @return                 ǂݍ񂾃oCgԂ܂B
                                zO̕R[h̏ꍇɂ̓G[Ƃ
                                ̃G[R[hԂ܂B
    */
    static s32 ReadUTF16LE( const u8 *src, s32 src_len, u32* utf32_code )
    {
        NN_NULL_TASSERT_( src );
        NN_NULL_TASSERT_( utf32_code );

        u32 code;
        if (src_len < 2) { return -1; }

        code = *src + (*(src + 1) << 8);

        *utf32_code = code;
        return 2;
    }

    /*!
        @brief        UTF16UTF8֕ϊs܂B

        @param[in]  src         ͌̃obt@
        @param[in]  dst_len     ͐̎c̒
        @return                 ϊꂽoUtf8̃oCgԂ܂
                                ̃G[R[hԂ܂B
    */
    static s32 ToUTF8( char* dst, s32 dstLen, const wchar_t* src )
    {
        s32 offset = 0;
        s32 len = std::wcslen( src );

        for (s32 i = 0; i < len;)
        {
            u32 utf32 = 0;

            i += ReadUTF16LE(
                reinterpret_cast<const u8*>(src + i),
                sizeof(wchar_t),
                &utf32 ) / sizeof(wchar_t);

            u8 work[4];
            s32 wlen = WriteUTF8( work, 4, utf32 );

            // oCg݂
            if (offset + wlen < dstLen)
            {
                std::memcpy( dst + offset, work, wlen );
                offset += wlen;
            }
            else
            {
                break;
            }
        }

        return offset;
    }
}


namespace sndio {

using namespace detail;

namespace
{
    struct Kernel
    {
        TMC_M4AMuxHandle        hM4AMux;                        //!< m4anh
        TMC_M4AContOption       m4aContOption;                  //!< m4a
        TMC_M4ATagInfo          m4aTagInfo;                     //!< m4a^O
        TMC_M4ATagExtendInfo    m4aTagExtendInfo[ TAG_NUM ];    //!< m4ag^O
        bool                    needDelete;                     //!< Xg[邩
        NN_PADDING3;
        wchar_t                 filePath[ nn::fs::MAX_FILE_PATH_LENGTH + 1 ];

        Kernel() :
        hM4AMux( NULL ),
        needDelete( false )
        {
            std::memset( &m4aContOption, 0, sizeof(m4aContOption) );
            std::memset( &m4aTagInfo, 0, sizeof(m4aTagInfo) );
            std::memset( m4aTagExtendInfo, 0, sizeof(m4aTagExtendInfo) );
            std::memset( &filePath, 0, sizeof(filePath) );
        }
    };
}
#define member( ins ) reinterpret_cast<Kernel*>(ins)


/*!
    @brief      \z

                RXgN^łB<br/>

    @param[in]  allocator   AP[^
*/
VoiceWriter::VoiceWriter( nn::fnd::IAllocator& allocator ) :
m_allocator( allocator ),
m_outputStream( NULL ),
m_streamBuffer( NULL ),
m_writeSamples( 0 )
{
    m_kernel = m_allocator.Allocate( sizeof(Kernel), 32 );
    m_kernel = new(m_kernel) Kernel();

    std::memset( &m_parameter, 0, sizeof(m_parameter) );
    std::memset( m_prmt, 0, sizeof(m_prmt) );

    m_parameter.permitUpload = true;
    m_parameter.permitEditDistribute = true;
    m_parameter.permitFirstDistribute = true;
    m_parameter.permitSecondDistribute = true;
}


/*!
    @brief      

                fXgN^łB<br/>
*/
VoiceWriter::~VoiceWriter()
{
    FlushStream();

    SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
    SafeFree( &m_allocator, m_streamBuffer );
    SafeFree( &m_allocator, m_kernel );

    if (s_decoderInitialize)
    {
        if (nn::aacenc::CTR::Finalize != 0)
            nn::aacenc::CTR::Finalize();
        s_decoderInitialize = false;
    }
}


/*!
    @brief      ݊Jns܂

                ʂVoiceWriterɏ݊JnsĂƂ͎s܂B<br/>

    @param[in]  pathName     t@CpX

    @retval     ResultSuccess               Jn
    @retval     ResultError                 sȃG[
    @retval     ResultMediaNotFound         fBAĂȂ
    @retval     ResultMediaInvalid          fBAs
    @retval     ResultAlreadyExists         t@Cɑ
    @retval     ResultNotEnoughSpace        eʕsG[
    @retval     ResultBadFormat             stH[}bg
    @retval     ResultArchiveInvalidated    A[JCus
    @retval     ResultOperationDenied       ANZX
    @retval     ResultWriteProtected        ݋֎~G[
    @retval     ResultMediaAccessError      fBAANZXG[
    @retval     ResultNotFormatted          tH[}bg
    @retval     ResultOutOfResource         \[Xs
    @retval     ResultInvalidEncoder        GR[_[G[
    @retval     ResultInvalidParameter      p[^s
*/
Result VoiceWriter::StartStream( const wchar_t* pathName )
{
    // zpXpX֕ϊ
    std::wcscpy( member( m_kernel )->filePath, pathName );

    void* place = m_allocator.Allocate( sizeof(detail::CachedWriter), 32 );
    detail::CachedWriter* stream = new (place) detail::CachedWriter( m_allocator );

    member( m_kernel )->needDelete = true;

    // t@C𐶐
    detail::ToReal( member( m_kernel )->filePath );

    nn::Result result;
    result = stream->TryInitialize( member( m_kernel )->filePath );
    if (result.IsFailure())
    {
        member( m_kernel )->needDelete = false;
        m_allocator.Free( stream );
        return GetFsWholeError( result );
    }

    m_outputStream = stream;
    return StartStream( *m_outputStream );
}


/*!
    @brief      ݊Jns܂(OutputStream)

                ʂVoiceWriterɏ݊JnsĂƂ͎s܂B<br/>

    @param[in]  outputStream ݐ̏o̓Xg[ւ̃|C^

    @retval     ResultSuccess               Jn
    @retval     ResultError                 sȃG[
    @retval     ResultMediaNotFound         fBAĂȂ
    @retval     ResultMediaInvalid          fBAs
    @retval     ResultAlreadyExists         t@Cɑ
    @retval     ResultNotEnoughSpace        eʕsG[
    @retval     ResultBadFormat             stH[}bg
    @retval     ResultArchiveInvalidated    A[JCus
    @retval     ResultOperationDenied       ANZX
    @retval     ResultWriteProtected        ݋֎~G[
    @retval     ResultMediaAccessError      fBAANZXG[
    @retval     ResultNotFormatted          tH[}bg
    @retval     ResultOutOfResource         \[Xs
    @retval     ResultInvalidEncoder        GR[_[G[
    @retval     ResultInvalidParameter      p[^s
*/
Result VoiceWriter::StartStream( nn::fs::IOutputStream& outputStream )
{
    m_writeSamples = 0;
    m_outputStream = &outputStream;
    NN_NULL_TASSERT_( m_outputStream );

    // TEh
    TMC_M4AContOption& contOption = member( m_kernel )->m4aContOption;
    contOption.num_ch = 1;              // m
    contOption.sampling_rate_idx = 5;   // 16kHz
    contOption.bit_rate_idx = 3;        // 16kbps

    // t
    contOption.creation_time.tm_sec     = m_parameter.createDate.second;
    contOption.creation_time.tm_min     = m_parameter.createDate.minute;
    contOption.creation_time.tm_hour    = m_parameter.createDate.hour;
    contOption.creation_time.tm_mday    = m_parameter.createDate.day;
    contOption.creation_time.tm_mon     = m_parameter.createDate.month;
    contOption.creation_time.tm_year    = m_parameter.createDate.year;
    contOption.creation_time.tm_wday    = m_parameter.createDate.week;
    contOption.creation_time.tm_yday    = 0;        // NʎZ gpȂ
    contOption.creation_time.tm_isdst   = 0;        // Ď gpȂ

    // ^O
    TMC_M4ATagInfo& tagInfo = member( m_kernel )->m4aTagInfo;
    tagInfo.album                       = m_parameter.albumName;
    tagInfo.album_len                   = m_parameter.albumNameLen;
    tagInfo.album_artist                = "";
    tagInfo.album_artist_len            = 0;
    tagInfo.artist                      = m_parameter.artistName;
    tagInfo.artist_len                  = m_parameter.artistNameLen;
    tagInfo.genre                       = 0;
    tagInfo.title                       = m_parameter.titleName;
    tagInfo.title_len                   = m_parameter.titleNameLen;
    tagInfo.track_number                = 0;
    tagInfo.track_number_total          = 0;
    tagInfo.year                        = m_parameter.createDate.year;
    tagInfo.genre_custom                = NULL;
    tagInfo.genre_custom_len            = 0;

    // g^O
    TMC_M4ATagExtendInfo* extendInfo = member( m_kernel )->m4aTagExtendInfo;
    std::memset( extendInfo, 0, sizeof(extendInfo) );
    extendInfo[0].tagname           = TAG_CODE;
    extendInfo[0].value             = const_cast<char*>( c_tagCode );
    extendInfo[0].value_len         = sizeof(c_tagCode);
    extendInfo[0].next              = &extendInfo[1];

    extendInfo[1].tagname           = TAG_SFTW;
    extendInfo[1].value             = m_parameter.sftw;
    extendInfo[1].value_len         = sizeof(m_parameter.sftw);
    extendInfo[1].next              = &extendInfo[2];

    extendInfo[2].tagname           = TAG_RCID;
    extendInfo[2].value             = m_parameter.rcid;
    extendInfo[2].value_len         = sizeof(m_parameter.rcid);
    extendInfo[2].next              = &extendInfo[3];

    extendInfo[3].tagname           = TAG_EDID;
    extendInfo[3].value             = m_parameter.edid;
    extendInfo[3].value_len         = sizeof(m_parameter.edid);
    extendInfo[3].next              = &extendInfo[4];

    // g^Op[~bV
    std::memset( m_prmt, 0, sizeof(m_prmt) );

    bool first = true;
    if (m_parameter.permitUpload)
    {
        std::strcat( m_prmt, PRMT_UPLOAD );
        first = false;
    }

    if (m_parameter.permitEditDistribute)
    {
        if (!first) std::strcat( m_prmt, "," );
        std::strcat( m_prmt, PRMT_EDIT_DISTRIBUTE );
        first = false;
    }

    if (m_parameter.permitFirstDistribute)
    {
        if (!first) std::strcat( m_prmt, "," );
        std::strcat( m_prmt, PRMT_FIRST_DISTRIBUTE );
        first = false;
    }

    if (m_parameter.permitSecondDistribute)
    {
        if (!first) std::strcat( m_prmt, "," );
        std::strcat( m_prmt, PRMT_SECOND_DISTRIBUTE );
    }

    extendInfo[4].tagname   = TAG_PRMT;
    extendInfo[4].value     = m_prmt;
    extendInfo[4].value_len = std::strlen( m_prmt ) + 1;
    extendInfo[4].next      = NULL;

    int work_size = TMC_M4AMux_GetHandleSize(
        10 + 1,
        contOption.sampling_rate_idx,
        &tagInfo,
        &extendInfo[0] );

    if (work_size <= 0)
    {
        // nhTCY擾G[
        return ResultInvalidEncoder;
    }

    SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
    member( m_kernel )->hM4AMux = m_allocator.Allocate( work_size, 32 );
    if (member( m_kernel )->hM4AMux == NULL)
    {
        // nhmۃG[
        return ResultOutOfResource;
    }

    SafeFree( &m_allocator, m_streamBuffer );
    m_streamBuffer = reinterpret_cast<u8*>( m_allocator.Allocate( NN_DSP_CODEC_AAC_BS_SIZE, 32 ) );
    if (m_streamBuffer == NULL)
    {
        // Xg[obt@mۃG[
        return ResultOutOfResource;
    }
    std::memset( m_streamBuffer, 0, NN_DSP_CODEC_AAC_BS_SIZE );

    // MUXCu
    int status = -1;

    status = TMC_M4AMux_Init(
        member( m_kernel )->hM4AMux,
        work_size,
        &contOption,
        &tagInfo,
        &extendInfo[0] );

    if (status != TMC_M4AMUX_SUCCESS)
    {
        // MUXCuG[
        switch (status)
        {
        case TMC_M4AMUX_PARAM_ERROR:
            return ResultInvalidParameter;

        case TMC_M4AMUX_MEMOVER_ERROR:
            return ResultOutOfResource;

        case TMC_M4AMUX_HANDLE_ERROR:
            return ResultInvalidEncoder;

        default:
            NN_TASSERT_( false );
            break;
        }
    }

    // GR[_Cu
    bool result = false;

    if (s_decoderInitialize)
    {
        nn::aacenc::CTR::Finalize();
    }

    result = nn::aacenc::CTR::Initialize(
        nn::aacenc::MODE_CH1,
        nn::aacenc::FS_16,
        nn::aacenc::RATE_M_H,
        nn::aacenc::FORM_RAW );
    s_decoderInitialize = true;

    if (!result)
    {
        // CuG[
        return ResultInvalidEncoder;
    }

    return ResultSuccess;
}


/*!
    @brief      ݏIs܂

    @retval     ResultSuccess               I
    @retval     ResultError                 sȃG[
    @retval     ResultMediaNotFound         fBAĂȂ
    @retval     ResultMediaInvalid          fBAs
    @retval     ResultAlreadyExists         t@Cɑ
    @retval     ResultNotEnoughSpace        eʕsG[
    @retval     ResultBadFormat             stH[}bg
    @retval     ResultArchiveInvalidated    A[JCus
    @retval     ResultOperationDenied       ANZX
    @retval     ResultWriteProtected        ݋֎~G[
    @retval     ResultMediaAccessError      fBAANZXG[
    @retval     ResultNotFormatted          tH[}bg
    @retval     ResultInvalidEncoder        GR[_[G[
*/
Result VoiceWriter::EndStream()
{
    // GR[_CuI
    if (s_decoderInitialize)
    {
        s_decoderInitialize = false;
        nn::aacenc::CTR::Finalize();
    }

    if (m_streamBuffer == NULL)
    {
        FlushStream();

        // ݂ɎsĂ
        return ResultInvalidEncoder;
    }

    nn::Result result;
    int status;
    int offset;
    u8* out_stream;
    int out_stream_size;
    s32 writeCount = 0;

    // I[̃Xg[
    status = TMC_M4AMux_GetEndStream( member( m_kernel )->hM4AMux, &out_stream, &out_stream_size);

    if (TMC_M4AMUX_SUCCESS != status)
    {
        FlushStream();
        SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
        SafeFree( &m_allocator, m_streamBuffer );

        // G[
        return ResultInvalidEncoder;
    }

    result = m_outputStream->TryWrite(
        &writeCount,
        out_stream,
        out_stream_size,
        true );
    if (result.IsFailure())
    {
        FlushStream();
        SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
        SafeFree( &m_allocator, m_streamBuffer );
        return GetFsWholeError( result );
    }

    // Xg[ҏW
    status = TMC_M4AMux_EditHeader(member( m_kernel )->hM4AMux, &out_stream, &out_stream_size, &offset);

    if (TMC_M4AMUX_SUCCESS != status)
    {
        FlushStream();
        SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
        SafeFree( &m_allocator, m_streamBuffer );
        // G[
        return ResultInvalidEncoder;
    }

    result = m_outputStream->TrySetPosition( s64(offset) );
    if (result.IsFailure())
    {
        FlushStream();
        SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
        SafeFree( &m_allocator, m_streamBuffer );
        return GetFsWholeError( result );
    }

    result = m_outputStream->TryWrite(
        &writeCount,
        out_stream,
        out_stream_size,
        true );

    if (result.IsFailure())
    {
        FlushStream();
        SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
        SafeFree( &m_allocator, m_streamBuffer );
        return GetFsWholeError( result );
    }

    Result sresult = FlushStream();
    SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
    SafeFree( &m_allocator, m_streamBuffer );

    return sresult;
}


/*!
    @brief      ݂s܂

                StartStreamĂяoĂKv܂B<br/>
                p[^̐ݒK؂ɍsĂKv܂B<br/>

    @param[out] pOut        ߂Tv
    @param[in]  pBufPCM16   ݃obt@(PCM16/TvO[g 16KHz)
    @param[in]  samples     ރTv(1024̔{łȂ΂ȂȂ)
    @param[in]  flush       ݌ɃtbVs(ftHgtrue)

    @retval     ResultSuccess               ݐ
    @retval     ResultError                 sȃG[
    @retval     ResultMediaNotFound         fBAĂȂ
    @retval     ResultMediaInvalid          fBAs
    @retval     ResultAlreadyExists         t@Cɑ
    @retval     ResultNotEnoughSpace        eʕsG[
    @retval     ResultBadFormat             stH[}bg
    @retval     ResultArchiveInvalidated    A[JCus
    @retval     ResultOperationDenied       ANZX
    @retval     ResultWriteProtected        ݋֎~G[
    @retval     ResultMediaAccessError      fBAANZXG[
    @retval     ResultNotFormatted          tH[}bg
    @retval     ResultOutOfResource         \[Xs
    @retval     ResultInvalidEncoder        GR[_[G[
*/
Result VoiceWriter::TryWrite( s32* pOut, s16* pBufPCM16, s32 samples, bool flush )
{
    NN_TASSERT_( (samples % NN_DSP_CODEC_AAC_PCM_SAMPLES) == 0 );

    s64 fileSize = 0;
    nn::Result result;

    if (pOut)
    {
        *pOut = 0;
    }

    for (s32 offset = 0; offset < samples; offset += NN_DSP_CODEC_AAC_PCM_SAMPLES)
    {
        s32 streamSize = 0;

        // TCY
        result = m_outputStream->TryGetSize( &fileSize );
        if (result.IsFailure())
        {
            SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
            SafeFree( &m_allocator, m_streamBuffer );
            return GetFsWholeError( result );
        }

        // t@CTCYɈ
        if (VOICE_LIMIT_BYTES < fileSize + NN_DSP_CODEC_AAC_PCM_SAMPLES)
        {
            // ȍ~̏̓X[
            return ResultSuccess;
        }

        // GR[_[ɓnOɃtbV
        nn::dsp::CTR::FlushDataCache( reinterpret_cast<uptr>(pBufPCM16),
            NN_DSP_CODEC_AAC_PCM_SAMPLES * sizeof(s16) );

        // GR[h
        if (nn::aacenc::CTR::Encode( pBufPCM16, pBufPCM16, m_streamBuffer ))
        {
            nn::aacenc::CTR::Wait();
            streamSize = nn::aacenc::CTR::GetStreamSize();

            NN_TASSERT_( 0 < streamSize );
        }
        else
        {
            SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
            SafeFree( &m_allocator, m_streamBuffer );

            return ResultInvalidEncoder;
        }

        u8* out_stream = NULL;
        int out_stream_size = 0;
        int status = TMC_M4AMux_SetStream(
            member( m_kernel )->hM4AMux,
            m_streamBuffer,
            streamSize,
            &out_stream,
            &out_stream_size );

        NN_TASSERT_( 0 < out_stream_size );

        if (status != TMC_M4AMUX_SUCCESS)
        {
            SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
            SafeFree( &m_allocator, m_streamBuffer );

            // MUXCuG[
            switch (status)
            {
            case TMC_M4AMUX_PARAM_ERROR:
                return ResultInvalidEncoder;

            case TMC_M4AMUX_MEMOVER_ERROR:
                return ResultOutOfResource;

            case TMC_M4AMUX_HANDLE_ERROR:
                return ResultInvalidEncoder;

            case TMC_M4AMUX_ENDSTREAM_ERROR:
                return ResultInvalidEncoder;

            default:
                NN_TASSERT_( false );
                break;
            }
        }

        s32 writeCount = 0;
        result = m_outputStream->TryWrite(
            &writeCount,
            out_stream,
            out_stream_size,
            flush );

        if (result.IsFailure())
        {
            SafeFree( &m_allocator, member( m_kernel )->hM4AMux );
            SafeFree( &m_allocator, m_streamBuffer );
            return GetFsWholeError( result );
        }

        if (pOut)
        {
            *pOut += NN_DSP_CODEC_AAC_PCM_SAMPLES;
        }

        pBufPCM16 += NN_DSP_CODEC_AAC_PCM_SAMPLES;
        m_writeSamples += NN_DSP_CODEC_AAC_PCM_SAMPLES;
    }

    return ResultSuccess;
}


/*!
    @brief      ܂ꂽTv擾܂

                TryWriteŏ݂sTv擾ł܂<br/>

    @return     ݃Tv
*/
s32 VoiceWriter::GetWriteSamples() const
{
    return m_writeSamples;
}


/*!
    @brief      p[^ݒ

                쐬p[^\̂gĐݒ肵܂<br/>

    @param[in]  param       쐬p[^
*/
void VoiceWriter::SetParameter( const VoiceParameter& param )
{
    std::memcpy( &m_parameter, &param, sizeof(VoiceParameter) );
}


/*!
    @brief      쐬w肵܂

    @param[in]  date        쐬̃p[^
*/
void VoiceWriter::SetCreateDate( const nn::fnd::DateTimeParameters& date )
{
    m_parameter.createDate = date;
}


/*!
    @brief      ^Cgݒ肵܂

    @param[in]  titleName   ^Cg
*/
void VoiceWriter::SetTitleName( const wchar_t* titleName )
{
    m_parameter.titleNameLen = ToUTF8( m_parameter.titleName, 128, titleName );
}


/*!
    @brief      Aoݒ肵܂

    @param[in]  albumName   Ao
*/
void VoiceWriter::SetAlbumName( const wchar_t* albumName )
{
    m_parameter.albumNameLen = ToUTF8( m_parameter.albumName, 128, albumName );
}


/*!
    @brief      A[eBXgݒ肵܂

    @param[in]  artistName  A[eBXg
*/
void VoiceWriter::SetArtistName( const wchar_t* artistName )
{
    m_parameter.artistNameLen = ToUTF8( m_parameter.artistName, 128, artistName );
}


/*!
    @brief      ۑ\tgEFAIDݒ肵܂

    @param[in]  \tgEFAID
*/
void VoiceWriter::SetSFTW( u8 sftw[5] )
{
    std::memcpy( m_parameter.sftw, sftw, 5 );
    m_parameter.sftw[5] = 0;
}


/*!
    @brief      ^{IDݒ肵܂

    @param[in]  ^{ID
*/
void VoiceWriter::SetRCID( u8 rcid[8] )
{
    std::memcpy( m_parameter.rcid, rcid, 8 );
    m_parameter.rcid[8] = 0;
}


/*!
    @brief      CTRAvŕҏW{IDݒ肵܂

    @param[in]  CTRAvŕҏW{ID
*/
void VoiceWriter::SetEDID( u8 edid[8] )
{
    std::memcpy( m_parameter.edid, edid, 8 );
    m_parameter.edid[8] = 0;
}


/*!
    @brief      Abv[hĂ邩ݒ肵܂

    @param[in]  true:C^[lbgAbv[h false:
*/
void VoiceWriter::SetPermitUpload( bool permitUpload )
{
    m_parameter.permitUpload = permitUpload;
}


/*!
    @brief      ό̉zz邩ݒ肵܂

    @param[in]  true:zz false:zzs
*/
void VoiceWriter::SetPermitEditDistribute( bool permitEditDistribute )
{
    m_parameter.permitEditDistribute = permitEditDistribute;
}


/*!
    @brief      \ݒ肵܂

    @param[in]  true: false:s
*/
void VoiceWriter::SetPermitFirstDistribute( bool permitFirstDistribute )
{
    m_parameter.permitFirstDistribute = permitFirstDistribute;
}


/*!
    @brief      񎟔zz\ݒ肵܂

    @param[in]  true:񎟔zz false:񎟔zzs
*/
void VoiceWriter::SetPermitSecondDistribute( bool permitSecondDistribute )
{
    m_parameter.permitSecondDistribute = permitSecondDistribute;
}


/*!
    @brief      t@CI܂
*/
Result VoiceWriter::FlushStream()
{
    if (member( m_kernel )->needDelete)
    {
        member( m_kernel )->needDelete = false;

        detail::CachedWriter* stream =
            reinterpret_cast<detail::CachedWriter*>( m_outputStream );

        nn::Result result = stream->FlushWrite();

        stream->~CachedWriter();
        m_allocator.Free( stream );
        m_outputStream = NULL;

        if (result.IsFailure())
        {
            return GetFsWholeError( result );
        }
    }

    return ResultSuccess;
}


} // namespace sndio
