/*---------------------------------------------------------------------------*
  Project:  sndio
  File:     sndio_VoiceWriter.h

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 19504 $
 *---------------------------------------------------------------------------*/

#ifndef SNDIO_VOICE_WRITER_H
#define SNDIO_VOICE_WRITER_H

#include <wchar.h>
#include <nn/types.h>
#include <nn/fnd.h>
#include <nn/fs.h>
#include <nn/dbg.h>
#include <nn/util/util_Color.h>
#include <nn/util/util_NonCopyable.h>
#include "detail/sndio_Result.h"
#include "detail/sndio_VoiceTypes.h"


/*!
    @brief      TEhIO֘A`
*/
namespace sndio {


/*!
    @brief      ރNXłB
*/
class VoiceWriter : private nn::util::NonCopyable<VoiceWriter>
{
public:

    /*!
        @brief      \z

                    RXgN^łB<br/>

        @param[in]  allocator   AP[^
    */
    explicit VoiceWriter( nn::fnd::IAllocator& allocator );

    /*!
        @brief      

                    fXgN^łB<br/>
    */
    ~VoiceWriter();

    /*!
        @brief      ݊Jns܂

                    ʂVoiceWriterɏ݊JnsĂƂ͎s܂B<br/>

        @param[in]  pathName     t@CpX

        @retval     ResultSuccess               Jn
        @retval     ResultError                 sȃG[
        @retval     ResultMediaNotFound         fBAĂȂ
        @retval     ResultMediaInvalid          fBAs
        @retval     ResultAlreadyExists         t@Cɑ
        @retval     ResultNotEnoughSpace        eʕsG[
        @retval     ResultBadFormat             stH[}bg
        @retval     ResultArchiveInvalidated    A[JCus
        @retval     ResultOperationDenied       ANZX
        @retval     ResultWriteProtected        ݋֎~G[
        @retval     ResultMediaAccessError      fBAANZXG[
        @retval     ResultNotFormatted          tH[}bg
        @retval     ResultOutOfResource         \[Xs
        @retval     ResultInvalidEncoder        GR[_[G[
        @retval     ResultInvalidParameter      p[^s
    */
    Result StartStream( const wchar_t* pathName );

    /*!
        @brief      ݊Jns܂(OutputStream)

                    ʂVoiceWriterɏ݊JnsĂƂ͎s܂B<br/>

        @param[in]  outputStream ݐ̏o̓Xg[ւ̃|C^

        @retval     ResultSuccess               Jn
        @retval     ResultError                 sȃG[
        @retval     ResultMediaNotFound         fBAĂȂ
        @retval     ResultMediaInvalid          fBAs
        @retval     ResultAlreadyExists         t@Cɑ
        @retval     ResultNotEnoughSpace        eʕsG[
        @retval     ResultBadFormat             stH[}bg
        @retval     ResultArchiveInvalidated    A[JCus
        @retval     ResultOperationDenied       ANZX
        @retval     ResultWriteProtected        ݋֎~G[
        @retval     ResultMediaAccessError      fBAANZXG[
        @retval     ResultNotFormatted          tH[}bg
        @retval     ResultOutOfResource         \[Xs
        @retval     ResultInvalidEncoder        GR[_[G[
        @retval     ResultInvalidParameter      p[^s
    */
    Result StartStream( nn::fs::IOutputStream& outputStream );

    /*!
        @brief      ݏIs܂

        @retval     ResultSuccess               I
        @retval     ResultError                 sȃG[
        @retval     ResultMediaNotFound         fBAĂȂ
        @retval     ResultMediaInvalid          fBAs
        @retval     ResultAlreadyExists         t@Cɑ
        @retval     ResultNotEnoughSpace        eʕsG[
        @retval     ResultBadFormat             stH[}bg
        @retval     ResultArchiveInvalidated    A[JCus
        @retval     ResultOperationDenied       ANZX
        @retval     ResultWriteProtected        ݋֎~G[
        @retval     ResultMediaAccessError      fBAANZXG[
        @retval     ResultNotFormatted          tH[}bg
        @retval     ResultInvalidEncoder        GR[_[G[
    */
    Result EndStream();

    /*!
        @brief      ݂s܂

                    StartStreamĂяoĂKv܂B<br/>
                    p[^̐ݒK؂ɍsĂKv܂B<br/>

        @param[out] pOut        ߂Tv
        @param[in]  pBufPCM16   ݃obt@(PCM16/TvO[g 16KHz)
        @param[in]  samples     ރTv(1024̔{łȂ΂ȂȂ)
        @param[in]  flush       ݌ɃtbVs(ftHgtrue)

        @retval     ResultSuccess               ݐ
        @retval     ResultError                 sȃG[
        @retval     ResultMediaNotFound         fBAĂȂ
        @retval     ResultMediaInvalid          fBAs
        @retval     ResultAlreadyExists         t@Cɑ
        @retval     ResultNotEnoughSpace        eʕsG[
        @retval     ResultBadFormat             stH[}bg
        @retval     ResultArchiveInvalidated    A[JCus
        @retval     ResultOperationDenied       ANZX
        @retval     ResultWriteProtected        ݋֎~G[
        @retval     ResultMediaAccessError      fBAANZXG[
        @retval     ResultNotFormatted          tH[}bg
        @retval     ResultOutOfResource         \[Xs
        @retval     ResultInvalidEncoder        GR[_[G[
    */
    Result TryWrite( s32* pOut, s16* pBufPCM16, s32 samples, bool flush = true );

    /*!
        @brief      ܂ꂽTv擾܂

                    TryWriteŏ݂sTv擾ł܂<br/>

        @return     ݃Tv
    */
    s32 GetWriteSamples() const;

    /*!
        @brief      p[^ݒ

                    쐬p[^\̂gĐݒ肵܂<br/>

        @param[in]  param       쐬p[^
    */
    void SetParameter( const VoiceParameter& param );

    /*!
        @brief      쐬w肵܂

        @param[in]  date        쐬̃p[^
    */
    void SetCreateDate( const nn::fnd::DateTimeParameters& date );

    /*!
        @brief      ^Cgݒ肵܂

        @param[in]  titleName   ^Cg
    */
    void SetTitleName( const wchar_t* titleName );

    /*!
        @brief      Aoݒ肵܂

        @param[in]  albumName   Ao
    */
    void SetAlbumName( const wchar_t* albumName );

    /*!
        @brief      A[eBXgݒ肵܂

        @param[in]  artistName  A[eBXg
    */
    void SetArtistName( const wchar_t* artistName );

    /*!
        @brief      ۑ\tgEFAIDݒ肵܂

        @param[in]  \tgEFAID
    */
    void SetSFTW( u8 sftw[5] );

    /*!
        @brief      ^{IDݒ肵܂

        @param[in]  ^{ID
    */
    void SetRCID( u8 rcid[8] );

    /*!
        @brief      CTRAvŕҏW{IDݒ肵܂

        @param[in]  CTRAvŕҏW{ID
    */
    void SetEDID( u8 edid[8] );

    /*!
        @brief      Abv[hĂ邩ݒ肵܂

        @param[in]  true:C^[lbgAbv[h false:
    */
    void SetPermitUpload( bool permitUpload );

    /*!
        @brief      ό̉zz邩ݒ肵܂

        @param[in]  true:zz false:zzs
    */
    void SetPermitEditDistribute( bool permitEditDistribute );

    /*!
        @brief      \ݒ肵܂

        @param[in]  true: false:s
    */
    void SetPermitFirstDistribute( bool permitFirstDistribute );

    /*!
        @brief      񎟔zz\ݒ肵܂

        @param[in]  true:񎟔zz false:񎟔zzs
    */
    void SetPermitSecondDistribute( bool permitSecondDistribute );

private:

    /*!
        @brief      t@CI܂
    */
    Result FlushStream();

    void*                       m_kernel;
    nn::fnd::IAllocator&        m_allocator;            //!< AP[^
    VoiceParameter              m_parameter;            //!< ۑp[^
    nn::fs::IOutputStream*      m_outputStream;         //!< o̓Xg[ւ̃|C^
    u8*                         m_streamBuffer;         //!< o̓obt@
    s32                         m_writeSamples;         //!< ݃Tv
    char                        m_prmt[ 128 ];          //!< PRMTobt@(UTF8)
};


} // namespace sndio


#endif /* SNDIO_VOICE_WRITER_H */
